import bpy
import os
import bmesh
import re
import ctypes
import functools
from mathutils import (
    Vector,
    Quaternion,
    Euler,
    Matrix,
)
from shutil import copyfile
from bpy_extras.image_utils import load_image
from bpy_extras.wm_utils.progress_report import (
    ProgressReport,
    ProgressReportSubstep,
)
from bpy.props import (
        CollectionProperty,
        BoolProperty,
        FloatProperty,
        StringProperty,
        EnumProperty,
        )
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper,
        axis_conversion,
        )
from . import (
    utils,
    file_utils,
    paths,
    kmaterial,
    kdefs,
    panel_bone,
    panel_object,
)

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_import_anim(bpy.types.Operator, ImportHelper):
    """Import a Turok 2 Animation File"""
    bl_idname = "turok2_object.import_anim"
    bl_label = "Import Anim"
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

	# Properties used by the file browser
    filepath: StringProperty(name="File Path", description="File path used for importing the anim file", default="", options={'HIDDEN'})
    files: CollectionProperty(type=bpy.types.OperatorFileListElement, options={'HIDDEN'})
    directory: StringProperty(subtype='FILE_PATH', options={'HIDDEN'})
    filter_folder: BoolProperty(name="Filter Folders", description="", default=True, options={'HIDDEN'})
    filename_ext = ".anim"
    filter_glob: StringProperty(default="*.anim", options={'HIDDEN'})

    # Custom properties
    animBakeYawOffsets: BoolProperty(name="Bake Yaw Offsets", description="For each animation, rotates the root bone by it's Yaw Offset. (Do not turn on if you intend to export this model back into Turok 2)", default=False)

    @classmethod
    def poll(cls, context):
        return context.object is not None
    #end def
    
    def execute(self, context):
        obj = context.object
        global_matrix = (Matrix.Scale(1.0/utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward, to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        successful = do_import(context, obj, self.filepath, global_matrix, self.animBakeYawOffsets)
        if not successful:
            return {'CANCELLED'}
        #end if
        
        return {'FINISHED'}
    #end def
    
    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Animation Options:", icon="ANIM")
        box = layout.box()
        row = box.row()
        row.prop(self, "animBakeYawOffsets")
    #end def
#end class

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_export_anim(bpy.types.Operator, ExportHelper):
    """Export a Turok 2 Animation File"""
    bl_idname = "turok2_object.export_anim"
    bl_label = 'Export Anim'
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

    filename_ext = ".anim"
    filter_glob: StringProperty(default="*.anim", options={'HIDDEN'})
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.t2.objType == "SKINNEDMESH"
    #end def
    
    def execute(self, context):
        armObj = utils.get_root_turok_object2(context.object, "SKINNEDMESH", "ARMATURE")
        global_matrix = (Matrix.Scale(utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward,to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        do_export(context, self.filepath, armObj, global_matrix)
        
        return {'FINISHED'}
    #end def
        
    def invoke(self, context, event):
        obj = context.object
        self.filepath = obj.name
        
        utils.mode_set('OBJECT')
        validateResults = validate(context, obj)
        if len(validateResults["errors"]) > 0 and len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Errors and Warnings - See System Console for full details", toConsole=False)
        elif len(validateResults["errors"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Error - See System Console for full details", toConsole=False)
        #end if
        if len(validateResults["errors"]) > 0:
            for message in validateResults["errors"]:
                print("[Export Anim Error] %s" % (message))
            #end for
            return {'CANCELLED'}
        #end if
        if len(validateResults["warnings"]) > 0:
            for message in validateResults["warnings"]:
                print("[Export Anim Warning] %s" % (message))
            #end for
        #end if

        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    #end def
        
    def draw(self, context):
        layout = self.layout
        #layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Animation Options:", icon="ANIM")
        # box = layout.box()
        # row = box.row()
        # row.prop(self, "")
    #end def
#end class

# -----------------------------------------------------------------------------
# Returns dictionary of anim structure. Returns None if there was an error.
def load(filepath, global_matrix):
    anim = {}
    with open(filepath, 'rb') as data:
        data.seek(0)
        anim["version"] = file_utils.read32(data, False) #version (2)
        if anim["version"] != 2:
            utils.show_error("[Import Anim] %s has incorrect version number of %d. Version must be 2." % (filepath, anim["version"]))
            return None
        #end if
        blockCount = file_utils.read32(data, False) #number of animation blocks
        #check if theres too many animations
        if blockCount > 65535:
            utils.show_error("[Import Anim] %s has more than 65535 animations (%d)" % (filepath, blockCount))
            return None
        elif blockCount == 0:
            utils.show_error("[Import Anim] %s has no animations" % (filepath))
            return None
        #end if
        anim["blocks"] = []
        for i in range(blockCount):
            block = {}
            anim["blocks"].append(block)
            block["id"] = file_utils.read32(data, True) #animID
            block["frameCount"] = file_utils.read32(data, False)
            #check if theres too many frames
            if block["frameCount"] > 65535:
                utils.show_error("[Import Anim] %s animation %d has more than 65535 frames (%d)" % (filepath, i, block["frameCount"]))
                return None
            elif block["frameCount"] == 0:
                utils.show_error("[Import Anim] %s animation %d has no frames" % (filepath, i))
                return None
            #end if
            boneCount = file_utils.read32(data, False)
            intialBoneCount = file_utils.read32(data, False)
            if boneCount != intialBoneCount:
                print("[Import Anim] %s animation %d boneCount (%d) is not the same as intialBoneCount (%d)" % (filepath, i, boneCount, intialBoneCount))
            #end if
            realBoneCount = max(boneCount, intialBoneCount)
            if realBoneCount == 0:
                utils.show_error("[Import Anim] %s animation %d has no bones" % (filepath, i))
                return None
            #end if
            
            translationCount = file_utils.read32(data, False)
            rotationCount = file_utils.read32(data, False)
            kfaCount = file_utils.read32(data, False) #key frame action count
            if translationCount > 65535 or rotationCount > 65535 or kfaCount > 65535:
                utils.show_error("[Import Anim] %s animation %d has more than 65535 trans(%d), rots(%d), kfa(%d)" % (filepath, i, translationCount, rotationCount, kfaCount))
                return None
            #end if
            block["frameScale"] = file_utils.readFloat(data)
            block["blendLength"] = file_utils.read16(data, False)
            block["blendStart"] = file_utils.read16(data, False)
            block["blendEnd"] = file_utils.read16(data, False)
            block["speed"] = file_utils.readFloat(data)
            block["loopFrame"] = file_utils.read16(data, False)

            block["bones"] = []
            for j in range(realBoneCount):
                bone = {
                    "translationIndex": 0xFFFF,             #0xFFFF means it is not used
                    "rotationIndex": 0xFFFF,                #0xFFFF means it is not used
                    "translation": Vector((0.0, 0.0, 0.0)), #used if no translationIndex
                    "rotation": Quaternion(),               #used if no rotationIndex
                    "trans": [], #the local location of the bone for every frame
                    "rot": [], #the local rotation of the bone for every frame
                }
                bone["trans"] = [Vector((0.0, 0.0, 0.0))] * block["frameCount"]
                bone["rot"] = [Quaternion((1.0, 0.0, 0.0, 0.0))] * block["frameCount"]
                block["bones"].append(bone)
            #end for
            
            for j in range(boneCount):
                block["bones"][j]["translationIndex"] = file_utils.read16(data, False)
                block["bones"][j]["rotationIndex"] = file_utils.read16(data, False)
            #end for
            
            for j in range(intialBoneCount):
                block["bones"][j]["translation"] = file_utils.readVector(data)
                block["bones"][j]["rotation"] = file_utils.readQuat(data)
            #end for
            
            block["yawOffsets"] = [0.0] * block["frameCount"]
            for j in range(block["frameCount"]):
                block["yawOffsets"][j] = file_utils.readFloat(data)
            #end for
            
            block["translations"] = []
            for j in range(translationCount):
                transFrames = []
                block["translations"].append(transFrames)
                for n in range(block["frameCount"]):
                    transFrames.append(file_utils.readVector(data))
                #end for
            #end for
            
            block["rotations"] = []
            for j in range(rotationCount):
                rotFrames = []
                block["rotations"].append(rotFrames)
                for n in range(block["frameCount"]):
                    rotFrames.append(file_utils.readQuat(data))
                #end for
            #end for
            
            for j in range(boneCount):
                translationID = block["bones"][j]["translationIndex"]
                rotationID = block["bones"][j]["rotationIndex"]
                for n in range(block["frameCount"]):
                    pos = block["bones"][j]["translation"]
                    rot = block["bones"][j]["rotation"]

                    if translationID != 0xFFFF:
                        pos = block["translations"][translationID][n]
                    #end if
                    if rotationID != 0xFFFF:
                        rot = block["rotations"][rotationID][n]
                    #end if
                    block["bones"][j]["trans"][n] = pos
                    block["bones"][j]["rot"][n] = rot
                #end for
            #end for
            
            block["actions"] = []
            for j in range(kfaCount):
                action = {}
                block["actions"].append(action)
                action["frame"] = file_utils.read16(data, False)
                action["event"] = file_utils.read16(data, False)
                action["bone"] = file_utils.read16(data, False)
                action["offset"] = file_utils.readVector(data)
                action["args"] = [0.0] * 8
                for n in range(8):
                    action["args"][n] = file_utils.readFloat(data)
                #end for
            #end for
        #end for
    #end with
    
    return anim
#end def

# -----------------------------------------------------------------------------
# Returns True if successful
def do_import(context, obj, filepath, global_matrix, bakeYawOffsets):
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        print("Importing Turok 2 Animation %r ..." % filepath)

        # obj = context.object
        if obj is None or obj.type != "ARMATURE":
            utils.show_error("[Import Anim] No Armature Object Selected")
            return False
        #end if
        
        animFileData = load(filepath, global_matrix)
        if animFileData is None:
            return False
        #end if
        
        animBoneCount = 0
        for i in range(len(animFileData["blocks"])):
            if len(animFileData["blocks"][i]["bones"]) > animBoneCount:
                animBoneCount = len(animFileData["blocks"][i]["bones"])
            #end if
        #end for

        armBoneCount = len(obj.data.bones)
        if armBoneCount < animBoneCount:
            utils.show_error("[Import Anim] Armature (%s) requires %d Bones for this Animation (%s)" % (obj.name, animBoneCount, filepath))
            return False
        #end if
        
        #Create bone names
        boneNames = {} #boneindex, name
        for bone in obj.data.bones:
            boneNames[bone.t2.boneIndex] = bone.name
        #end for
        
        fileName = os.path.split(filepath)[1]
        fileNameNoExt = os.path.splitext(fileName)[0]
        scene = bpy.context.scene
        bpy.context.view_layer.objects.active = obj
        
        utils.mode_set('POSE')
        animData = obj.animation_data_create()
        firstAction = None
        totalKeyFrameInserts = 0
        progress.enter_substeps(len(animFileData["blocks"]))
        fpsScale = 1.0
        for animIndex in range(len(animFileData["blocks"])):
            anim = animFileData["blocks"][animIndex]
            animName = kdefs.defs["anims"].get(anim["id"], "Anim %i" % (anim["id"]))

            animAction = bpy.data.actions.new(animName)
            animData.action = animAction
            animAction.use_fake_user = True
            tAction = animData.action.t2
            tAction.animID = anim["id"]
            
            #Set key Frame Actions
            for i in range(len(anim["actions"])):
                keyFrame = tAction.keyFrames.add()
                animKFAction = anim["actions"][i]
                keyFrame.frame = animKFAction["frame"]
                keyFrame.event = animKFAction["event"]
                keyFrame.bone = animKFAction["bone"]
                keyFrame.offset = animKFAction["offset"]
                keyFrame.arg1 = animKFAction["args"][0]
                keyFrame.arg2 = animKFAction["args"][1]
                keyFrame.arg3 = animKFAction["args"][2]
                keyFrame.arg4 = animKFAction["args"][3]
                keyFrame.arg5 = animKFAction["args"][4]
                keyFrame.arg6 = animKFAction["args"][5]
                keyFrame.arg7 = animKFAction["args"][6]
                keyFrame.arg8 = animKFAction["args"][7]
            #end
            
            #set animation action properties
            tAction.frameScale = anim["frameScale"]
            tAction.blendLength = anim["blendLength"]
            tAction.blendStart = anim["blendStart"]
            tAction.blendEnd = anim["blendEnd"]
            tAction.speed = anim["speed"]
            tAction.loopFrame = anim["loopFrame"]
            
            if animIndex == 0:
                firstAction = animAction
                bpy.context.scene.frame_start = 0
                bpy.context.scene.frame_set(0)
            #end if
            
            hasYawOffset = False
            for frameIndex in range(0, anim["frameCount"]):
                if anim["yawOffsets"][frameIndex] != 0.0:
                    hasYawOffset = True
                    break
                #end if
            #end for

            if hasYawOffset:
                usedFrames = utils.get_used_frames(anim["yawOffsets"])
                usedFramesLength = len(usedFrames)
                yawCurve = animAction.fcurves.new(data_path="t2.yawOffset")
                yawCurve.keyframe_points.add(count=usedFramesLength)
                for i in range(usedFramesLength):
                    frameIndex = usedFrames[i]
                    yawCurve.keyframe_points[i].co = (frameIndex * fpsScale, anim["yawOffsets"][frameIndex])
                    yawCurve.keyframe_points[i].interpolation = 'LINEAR'
                #end for
                yawCurve.update()
                totalKeyFrameInserts += usedFramesLength
            #end if
            
            for i in range(animBoneCount):
                boneName = boneNames.get(i, '')
                if boneName not in obj.pose.bones:
                    utils.show_error("[Import Anim] There is no pose bone named %s in the Armature" % (boneName))
                    return False
                #end if

                #Get the new bone Translation and Rotation for all the frames
                boneTrans = []
                if i >= len(anim["bones"]):
                    boneTrans = [Vector((0.0, 0.0, 0.0))] * anim["frameCount"]
                else:
                    boneTrans = anim["bones"][i]["trans"]
                #end if
                
                boneRots = []
                if i >= len(anim["bones"]):
                    boneRots = [Quaternion()] * anim["frameCount"]
                else:
                    if bakeYawOffsets and hasYawOffset and i == 0:
                        boneRots = anim["bones"][i]["rot"].copy()
                    else:
                        boneRots = anim["bones"][i]["rot"]
                    #end if
                #end if
                if bakeYawOffsets and hasYawOffset and i == 0:
                    for j in range(anim["frameCount"]):
                        boneRots[j] = anim["bones"][i]["rot"][j].rotate(Quaternion((0.0, 0.0, 1.0), anim["yawOffsets"][j]))
                    #end for
                #end if

                #set initial poseBone position and rotation to first frame
                poseBone = obj.pose.bones[boneName]
                poseBone.location = boneTrans[0]
                poseBone.rotation_mode = 'QUATERNION'
                poseBone.rotation_quaternion = boneRots[0]

                #Create the curves for only the keyframes used
                usedLocFrames = [utils.get_used_frames([v[j] for v in boneTrans]) for j in range(3)]
                usedRotFrames = [utils.get_used_frames([q[j] for q in boneRots]) for j in range(4)]
                
                
                trans_curves = []
                for j in range(3):
                    trans_curves.append(animAction.fcurves.new(poseBone.path_from_id("location"), index=j))
                    trans_curves[j].keyframe_points.add(count=len(usedLocFrames[j]))
                #end for
                rot_curves = []
                for j in range(4):
                    rot_curves.append(animAction.fcurves.new(poseBone.path_from_id("rotation_quaternion"), index=j))
                    rot_curves[j].keyframe_points.add(count=len(usedRotFrames[j]))
                #end for
                for channelIndex in range(3):
                    totalKeyFrameInserts += len(usedLocFrames[channelIndex])
                    for j in range(len(usedLocFrames[channelIndex])):
                        frameIndex = usedLocFrames[channelIndex][j]
                        trans_curves[channelIndex].keyframe_points[j].co = (frameIndex * fpsScale, boneTrans[frameIndex][channelIndex])
                        trans_curves[channelIndex].keyframe_points[j].interpolation = 'LINEAR'
                    #end for
                #end for
                for channelIndex in range(4):
                    totalKeyFrameInserts += len(usedRotFrames[channelIndex])
                    for j in range(len(usedRotFrames[channelIndex])):
                        frameIndex = usedRotFrames[channelIndex][j]
                        rot_curves[channelIndex].keyframe_points[j].co = (frameIndex * fpsScale, boneRots[frameIndex][channelIndex])
                        rot_curves[channelIndex].keyframe_points[j].interpolation = 'LINEAR'
                    #end for
                #end for
                
                for transCurve in trans_curves:
                    transCurve.update()
                #end for
                
                for rotCurve in rot_curves:
                    rotCurve.update()
                #end for
            #end for
                        
            progress.step()
        #end for
        
        #Exit Armature pose mode
        utils.mode_set('OBJECT')

        if firstAction is not None:
            animData.action = firstAction
        #end if
        
        #select only the armature
        for ob in bpy.context.selected_objects:
            ob.select_set(False)
        #end for
        obj.select_set(True)
        
        progress.leave_substeps("Finished! (actions:%d bones:%d keyframes:%d)" % (len(animFileData["blocks"]), animBoneCount, totalKeyFrameInserts))
    #end with
    
    return True
#end def

# -----------------------------------------------------------------------------
# 
def do_export(context, filepath, armObj, global_matrix):
    validateResults = validate(context, armObj)
    if len(validateResults["errors"]) > 0:
        return {'CANCELLED'}
    #end if

    actions = validateResults["actions"]
    actions.sort(key=lambda x: x.t2.animID)
    scene = context.scene
    armature = armObj.data
    bones = [bone for bone in armature.bones]
    bones.sort(key=lambda bone: bone.t2.boneIndex)
    boneCount = len(bones)
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        print("Exporting Turok 2 Animationn %r ..." % filepath)

        # fileDir, fileName = os.path.split(filepath)
        # fileNameNoExt = os.path.splitext(fileName)[0]

        with open(filepath, 'wb') as data:
            data.seek(0)
            file_utils.write32(data, 2, False) #version (2)
            file_utils.write32(data, len(actions), False) #number of animation blocks
            progress.enter_substeps(len(actions))
            for action in actions:
                frameStart, frameEnd = action.frame_range
                frameStart = int(frameStart)
                frameEnd = int(frameEnd)
                frameCount = (frameEnd - frameStart) + 1
                boneLocs = [] * boneCount
                boneRots = [] * boneCount
                boneRotsEuler = [] * boneCount
                boneRotsAA = [] * boneCount
                boneRotsMode = [] * boneCount
                boneTransIndex = [-1] * boneCount
                boneRotsIndex = [-1] * boneCount
                for i in range(boneCount):
                    boneLocs.append([])
                    boneRots.append([])
                    boneRotsEuler.append([])
                    boneRotsAA.append([])
                    boneRotsMode.append([])
                    for i2 in range(frameCount):
                        boneLocs[i].append(Vector((0.0, 0.0, 0.0)))
                        boneRots[i].append(Quaternion((1.0, 0.0, 0.0, 0.0)))
                        boneRotsEuler[i].append(Euler((0.0, 0.0, 0.0), 'XYZ'))
                        boneRotsAA[i].append(Quaternion((1.0, 0.0, 0.0, 0.0)))
                        boneRotsMode[i].append(-1)
                    #end for
                #end for
                yawOffsets = [0.0] * frameCount
                for frame in range(frameCount):
                    realFrame = frameStart + frame
                    hasRotQuat = [False] * 4
                    hasRotEuler = [False] * 3
                    hasRotAA = [False] * 4
                    for f in action.fcurves:
                        propObjPath, propDot, propType = f.data_path.rpartition('.')
                        propObj = armObj.path_resolve(propObjPath)
                        if type(propObj) is bpy.types.PoseBone:
                            bone = armObj.data.bones[propObj.name]
                            boneIndex = utils.index_of(bones, bone)
                            if propType == "location":
                                boneLocs[boneIndex][frame][f.array_index] = f.evaluate(realFrame)
                            elif propType == "rotation_quaternion":
                                boneRots[boneIndex][frame][f.array_index] = f.evaluate(realFrame)
                                hasRotQuat[f.array_index] = True
                                if hasRotQuat[0] and hasRotQuat[1] and hasRotQuat[2] and hasRotQuat[3]:
                                    boneRotsMode[boneIndex][frame] = 0
                            elif propType == "rotation_euler":
                                boneRotsEuler[boneIndex][frame][f.array_index] = f.evaluate(realFrame)
                                if boneRotsMode[boneIndex][frame] == -1 or boneRotsMode[boneIndex][frame] > 1:
                                    hasRotEuler[f.array_index] = True
                                    if hasRotEuler[0] and hasRotEuler[1] and hasRotEuler[2]:
                                        boneRotsMode[boneIndex][frame] = 1
                            elif propType == "rotation_axis_angle":
                                boneRotsAA[boneIndex][frame][f.array_index] = f.evaluate(realFrame)
                                if boneRotsMode[boneIndex][frame] == -1:
                                    hasRotAA[f.array_index] = True
                                    if hasRotAA[0] and hasRotAA[1] and hasRotAA[2] and hasRotAA[3]:
                                        boneRotsMode[boneIndex][frame] = 2
                            #end if
                        elif type(propObj) is turok_object_panel.TurokObjectSettings:
                            if propType == "yawOffset":
                                yawOffsets[frame] = f.evaluate(realFrame)
                            #end if
                        #end if
                    #end for
                #end for

                #check if need to use euler or axis angle over Quaternion
                for i in range(boneCount):
                    for frame in range(frameCount):
                        if boneRotsMode[i][frame] == 1: #Euler
                            boneRots[i][frame] = boneRotsEuler[i][frame].to_quaternion()
                        elif boneRotsMode[i][frame] == 2: #Axis Angle
                            q = boneRotsAA[i][frame]
                            boneRots[i][frame] = Quaternion(Vector((q[1], q[2], q[3])), q[0])
                        #end if
                    #end for
                #end for
                
                def editbone_matrix(bone):
                    parent = bone.parent
                    if parent == None:
                        return bone.matrix_local
                    else:
                        return parent.matrix_local.inverted() @ bone.matrix_local
                    #end if
                #end def
        
                #Apply Edit Bone position Offset since it was applied in the skinnedmesh
                for i in range(len(boneLocs)): #bone
                    bone = armObj.data.bones[utils.find_bone_name(armObj, i)]
                    editBoneMatrix = editbone_matrix(bone)
                    for i2 in range(len(boneLocs[i])): #frame
                        old = boneLocs[i][i2].copy()
                        boneLocs[i][i2] = Matrix.Translation(editBoneMatrix.translation) @ old
                    #end for
                #end for

                transIndexs = [] #values are boneID
                rotsIndexs = [] #values are boneID
                #check each bone loc/rot and if they are ever different in any frame then use the next transIndex/rotIndex
                for i in range(boneCount):
                    #check if the bones Locs are exactly the same as another transIndex
                    for i2 in range(len(transIndexs)):
                        transBoneIndex = transIndexs[i2]
                        if boneLocs[i] == boneLocs[transBoneIndex]:
                            boneTransIndex[i] = i2
                            break
                        #end if
                    #end for
                    if boneTransIndex[i] == -1:
                        for frame in range(1, frameCount):
                            if boneLocs[i][frame] != boneLocs[i][0]:
                                boneTransIndex[i] = len(transIndexs)
                                transIndexs.append(i)
                                break
                            #end if
                        #end for
                    #end if
                    
                    #check if the bones Rots are exactly the same as another rotsIndex
                    for i2 in range(len(rotsIndexs)):
                        rotsBoneIndex = rotsIndexs[i2]
                        if boneRots[i] == boneRots[rotsBoneIndex]:
                            boneRotsIndex[i] = i2
                            break
                        #end if
                    #end for
                    if boneRotsIndex[i] == -1:
                        for frame in range(1, frameCount):
                            if boneRots[i][frame] != boneRots[i][0]:
                                boneRotsIndex[i] = len(rotsIndexs)
                                rotsIndexs.append(i)
                                break
                            #end if
                        #end for
                    #end if
                #end for
                
                transCount = len(transIndexs) #check if key frames are different for each bone if it's the same from start to finish then use that as initial position
                rotCount = len(rotsIndexs) #check if key frames are different for each bone if it's the same from start to finish then use that as initial position

                file_utils.write32(data, action.t2.animID, True) #animID
                file_utils.write32(data, frameCount, False)
                file_utils.write32(data, boneCount, False) #boneIndexes (trans/rot indexes)
                file_utils.write32(data, boneCount, False) #initalBones (inital translation/rotation)
                file_utils.write32(data, transCount, False) #translations for each frame
                file_utils.write32(data, rotCount, False) #rotations for each frame
                file_utils.write32(data, len(action.t2.keyFrames), False)
                
                file_utils.writeFloat(data, action.t2.frameScale)
                file_utils.write16(data, action.t2.blendLength, False)
                file_utils.write16(data, action.t2.blendStart, False)
                file_utils.write16(data, action.t2.blendEnd, False)
                file_utils.writeFloat(data, action.t2.speed)
                file_utils.write16(data, action.t2.loopFrame, False)
                
                for i in range(boneCount):
                    file_utils.write16(data, boneTransIndex[i], True) #translation index
                    file_utils.write16(data, boneRotsIndex[i], True) #rotation index
                #end for
                for i in range(boneCount):
                    file_utils.writeVector(data, boneLocs[i][0]) #initial translation
                    file_utils.writeQuat(data, boneRots[i][0]) #initial rotation
                #end for
                for i in range(frameCount):
                    file_utils.writeFloat(data, yawOffsets[i]) #yaw offset
                #end for
                for i in range(transCount):
                    for i2 in range(frameCount):
                        file_utils.writeVector(data, boneLocs[transIndexs[i]][i2])
                    #end for
                #end for
                for i in range(rotCount):
                    for i2 in range(frameCount):
                        file_utils.writeQuat(data, boneRots[rotsIndexs[i]][i2])
                    #end for
                #end for
                for kfa in action.t2.keyFrames:
                    file_utils.write16(data, kfa.frame, False)
                    file_utils.write16(data, kfa.event, False)
                    file_utils.write16(data, kfa.bone, False)
                    file_utils.writeVector(data, kfa.offset)
                    file_utils.writeFloat(data, kfa.arg1)
                    file_utils.writeFloat(data, kfa.arg2)
                    file_utils.writeFloat(data, kfa.arg3)
                    file_utils.writeFloat(data, kfa.arg4)
                    file_utils.writeFloat(data, kfa.arg5)
                    file_utils.writeFloat(data, kfa.arg6)
                    file_utils.writeFloat(data, kfa.arg7)
                    file_utils.writeFloat(data, kfa.arg8)
                #end for                
            #end for
        #end with
        
        progress.step("Animation Successfully Exported!")
    #end with
        
    return {'FINISHED'}
#end def

# -----------------------------------------------------------------------------
# Return dictionary also validates animation actions.
# "errors"      List of string
# "warnings"    List of string
# "info"        List of string
# "actions"     List of valid animation actions
def validate(context, armObj):
    result = { "errors": [], "warnings": [], "info": [], "actions": []}
    
    #Error: Check if object is not None
    if armObj is None:
        result["errors"].append("Object is None")
        return result
    #end if
    
    # Find all the animation actions this object could be using
    # Must be using fake user, and must have all properties that all the fcurves use
    nonFakeUserActionCount = 0
    for action in bpy.data.actions:
        if not action.use_fake_user:
            nonFakeUserActionCount += 1
            continue
        #end if
        if len(action.fcurves) == 0:
            result["warnings"].append("Animation Action \"%s\", has no fcurves" % (action.name))
            continue
        #end if
            
        isUsed = True
        for fcurve in action.fcurves:
            prop_path = fcurve.data_path.rpartition('.')[0]
            try:
                armObj.path_resolve(prop_path)
            except ValueError:
                result["warnings"].append("Animation Action \"%s\", fcurve \"%s\" is not valid. This action will not be exported." % (action.name, fcurve.data_path))
                isUsed = False
                break
            #end try except
        #end for
        if isUsed:
            result["actions"].append(action)
        #end if
    #end for
    
    #Error: Must have at least 1 action to export
    if len(result["actions"]) == 0:
        if nonFakeUserActionCount == len(bpy.data.actions):
            result["errors"].append("No Animation Actions to export for this Armature. Make sure you set actions with a fake user.")
        else:
            result["errors"].append("You haven't made any Animation Actions to export for this Armature.")
        #end if
    #end if

    #Check each animation action
    for action in result["actions"]:
        frameStart, frameEnd = action.frame_range
        frameStart = int(frameStart)
        frameEnd = int(frameEnd)
        frameCount = (frameEnd - frameStart) + 1
        #Error: Check if Loop Frame is greater than end of frame
        if action.t2.loopFrame >= frameCount:
            result["warnings"].append("Animation Action \"%s\" loopFrame %i is greater than the last frame of %i" % (action.name, action.t2.loopFrame, frameCount))
        #end if
        
        #Error: check each key frame action's frame if it went beyond the last frame
        for keyFrame in action.t2.keyFrames:
            if keyFrame.frame >= frameCount:
                result["warnings"].append("Animation Action \"%s\", a keyframe action has a frame of %i which is greater than the last frame of %i" % (action.name, keyFrame.frame, frameCount))
            #end if
        #end for
        
        #Error: check each key frame action's bone are set to a bone that exists
        for keyFrame in action.t2.keyFrames:
            if keyFrame.bone != 0xFFFF:
                boneName = utils.find_bone_name(armObj, keyFrame.bone)
                if not boneName:
                    result["warnings"].append("Animation Action \"%s\", a keyframe action has a bone of %i which does not exist in the armature" % (action.name, keyFrame.bone))
                #end if
            #end if
        #end for
    #end for
    
    return result
#end def

# -----------------------------------------------------------------------------
#
def menu_func_import_anim(self, context):
    self.layout.operator(TUROK2_OT_import_anim.bl_idname, text="Turok 2 Animation (.anim)")
#end def

# -----------------------------------------------------------------------------
#
def menu_func_export_anim(self, context):
    self.layout.operator(TUROK2_OT_export_anim.bl_idname, text="Turok 2 Animation (.anim)")
#end def

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(TUROK2_OT_import_anim)
    bpy.utils.register_class(TUROK2_OT_export_anim)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_anim)
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export_anim)
#end def

# -----------------------------------------------------------------------------
#
def unregister():
    bpy.utils.unregister_class(TUROK2_OT_import_anim)
    bpy.utils.unregister_class(TUROK2_OT_export_anim)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_anim)
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export_anim)
#end def
