import os
import aud
import random

from . import (
    utils,
    paths,
    file_utils,
    kdefs,
)

playingSounds = {} #key is ksnd absolute path + sound index, value is array of AudioSound

# -----------------------------------------------------------------------------
# Plays the voice sound
def play_voice(id):
    voice = kdefs.defs["voices"].get(id)
    if voice:
        path = paths.find_t2_file_path(voice["name"] + ".wav")
        if path is not None:
            aud.Device().play(aud.Sound.file(path))
        #end if
    #end if
#end def

# -----------------------------------------------------------------------------
# Returns the ksnd or None
def load_play_ksnd(ksndPath, useLocalGamePath=True):
    path = ksndPath
    if useLocalGamePath:
        path = paths.find_t2_file_path(ksndPath)
    #end if
    if path is not None:
        sound = kdefs.load_sound(path)
        play_ksnd(sound)
        return sound
    #end if
    
    return None
#end def

# -----------------------------------------------------------------------------
# 
def play_ksnd(ksnd):
    if ksnd:
        key = ""
        for ksndIndex, sfx in enumerate(ksnd["sfx"]):
            key = "%s%i" % (ksnd["path"], ksndIndex)
            
            #Random chance failed
            if random.random() > sfx['random']:
                continue
            #end if
            
            #Pitch Envelope max and start value too low
            pitchRandStart = 1.0
            if sfx['bEnablePitchEnvelope'] != 0:
                if sfx['pitch_envelope']['start'] <= 0.0:
                    continue
                #end if
            else:
                if sfx['pitch_min'] == sfx['pitch_max']:
                    pitchRandStart = sfx['pitch_min']
                else:
                    pitchRandStart = random.uniform(sfx['pitch_min'], sfx['pitch_max'])
                #end if
                if pitchRandStart <= 0.0:
                    continue
                #end if
            #end if

            if key in playingSounds:
                #Can't play until sound is finished playing
                if sfx['bOneInstancePerSource']:
                    continue
                #end if
            else:
                playingSounds[key] = []
            #end if
            playingSounds[key].append(AudioSound(sfx, ksndIndex, pitchRandStart))
        #end for
    #end if
#end def

# -----------------------------------------------------------------------------
# Stops all sounds
def stop_sound():
    global playingSounds
    
    playingSounds.clear()
    aud.Device().stopAll()
#end def

# -----------------------------------------------------------------------------
# 
class AudioSound:
    def __init__(self, newKsnd, index, pitchRandomStart):
        self.ksnd = newKsnd
        self.soundIndex = index
        self.pitchRandStart = pitchRandomStart
        self.handle = None
        self.lastPos = 0.0
        self.soundObj = None
        self.ticks = 0
        self.isDelaying = True
        self.doRemove = False
        self.LoopEndReached = False
        self.delayTime = 0.0
        self.length = 1.0 #length of audio clip in seconds
        self.volShiftStartTime = self.ksnd['volume_envelope']['envStartTime'] / 60.0
        self.volShiftEndTime = self.volShiftStartTime + max(self.ksnd['volume_envelope']['duration'] / 60.0, 0.0)
        if self.ksnd['bEnableVolumeEnvelope'] != 0:
            self.volumeTarget = utils.clamp(self.ksnd['volume_envelope']['start'] * self.ksnd['volume_envelope']['max'], 0.0, 1.0)
        else:
            self.volumeTarget = utils.clamp(self.ksnd['volume_envelope']['max'], 0.0, 1.0)
        #end if
        self.pitchShiftStartTime = self.ksnd['pitch_envelope']['envStartTime'] / 60.0
        self.pitchShiftEndTime = self.pitchShiftStartTime + max(self.ksnd['pitch_envelope']['duration'] / 60.0, 0.0)
        if self.ksnd['bEnablePitchEnvelope'] != 0:
            if self.ksnd['volume_envelope']['envStartTime'] == 0 and self.ksnd['pitch_envelope']['duration'] <= 0:
                self.pitchTarget = max(self.ksnd['pitch_envelope']['end'], 0.01)
            else:
                self.pitchTarget = max(self.ksnd['pitch_envelope']['start'], 0.01)
            #end if
        #end if
        else:
            self.pitchTarget = max(self.pitchRandStart, 0.01)
        #end if
        
        wavPath = paths.find_t2_file_path(self.ksnd['wavefile'])
        if wavPath is not None:
            self.soundObj = aud.Sound.file(wavPath)
            self.length = self.soundObj.length / self.soundObj.specs[0] #soundLength / The sampling rate in Hz
        else:
            self.doRemove = True
        #end if
    #end def
    
    # -----------------------------------------------------------------------------
    # 
    def update(self):
        if self.ticks == 0:
            self.delayTime = utils.time_from() + max(self.ksnd['delay'] / 60.0, 0.0)
        #end if
        if self.handle is not None and not self.isDelaying:
            if self.handle.status:
                soundPos = self.handle.position
                newVolume = self.volumeTarget
                newPitch = self.pitchTarget
                if not self.LoopEndReached:
                    if soundPos > self.lastPos:
                        self.lastPos = soundPos
                        if self.ksnd['bEnableVolumeEnvelope'] != 0:
                            lerpTime = utils.time_normalize(soundPos, self.volShiftStartTime, self.volShiftEndTime)
                            newVolume = utils.clamp(utils.lerp(self.ksnd['volume_envelope']['start'], self.ksnd['volume_envelope']['end'], lerpTime) * self.ksnd['volume_envelope']['max'], 0.0, 1.0)
                        else:
                            newVolume = utils.clamp(self.ksnd['volume_envelope']['max'], 0.0, 1.0)
                        #end if
                        
                        if self.ksnd['bEnablePitchEnvelope'] != 0:
                            lerpTime = utils.time_normalize(soundPos, self.pitchShiftStartTime, self.pitchShiftEndTime)
                            newPitch = max(utils.lerp(self.ksnd['pitch_envelope']['start'], self.ksnd['pitch_envelope']['end'], lerpTime), 0.01)
                        else:
                            newPitch = max(self.pitchRandStart, 0.01)
                        #end if
                    elif self.ksnd['bLooping'] and soundPos < 0.0: #loop sound reached end
                        self.LoopEndReached = True
                        if self.ksnd['bEnableVolumeEnvelope'] != 0:
                            newVolume = utils.clamp(self.ksnd['volume_envelope']['end'] * self.ksnd['volume_envelope']['max'], 0.0, 1.0)
                        #end if
                        if self.ksnd['bEnablePitchEnvelope'] != 0:
                            newPitch = max(self.ksnd['pitch_envelope']['end'], 0.01)
                        #end if
                        #if ends with no volume while looping then stop playing
                        if newVolume == 0.0:
                            self.handle.stop()
                            self.doRemove = True
                            self.handle = None
                        #end if
                    #end if
                #end if
                
                if self.handle is not None:
                    if newVolume != self.volumeTarget:
                        self.volumeTarget = newVolume
                        try:
                            self.handle.volume = self.volumeTarget
                        except:
                            pass
                        #end try except
                    #end if

                    if newPitch != self.pitchTarget:
                        self.pitchTarget = newPitch
                        try:
                            self.handle.pitch = self.pitchTarget
                        except:
                            pass
                        #end try except
                    #end if
                #end if
            #end if

            #Check if sound is finished playing
            if not self.doRemove and not self.handle.status:
                self.doRemove = True
                self.handle = None
            #end if
        #end if
        
        self.ticks += 1
    #end def
    
    # -----------------------------------------------------------------------------
    # 
    def play_sound(self):
        self.handle = aud.Device().play(self.soundObj)
        try:
            self.handle.volume = self.volumeTarget
            self.handle.pitch = self.pitchTarget
        except:
            pass
        #end try except
        if self.ksnd['bLooping'] != 0:
            self.handle.loop_count = -1
        #end if
    #end def
    
    # -----------------------------------------------------------------------------
    # 
    def delay_finished(self):
        self.isDelaying = False
        self.play_sound()
    #end def
        
    # -----------------------------------------------------------------------------
    # 
    def __str__(self):
        return "delayTime:%f, pitchTarget:%f, volumeTarget:%f, length:%f" % (self.delayTime,
        self.pitchTarget, self.volumeTarget, self.length)
    #end def

    # -----------------------------------------------------------------------------
    # 
    def __repr__(self):
        return str(self)
    #end def
#end class
