import os
import json
import struct
import sys
from mathutils import Vector, Quaternion
from .pcpp import Preprocessor, OutputDirective, Action
from . import (
    utils,
)

# -----------------------------------------------------------------------------
#
def read32(data, is_signed):
    return int.from_bytes(data.read(4), 'little', signed=is_signed)
#end def

# -----------------------------------------------------------------------------
#
def read16(data, is_signed):
    return int.from_bytes(data.read(2), 'little', signed=is_signed)
#end def
    
# -----------------------------------------------------------------------------
#
def read8(data, is_signed):
    return int.from_bytes(data.read(1), 'little', signed=is_signed)
#end def

# -----------------------------------------------------------------------------
# Returns bytes object
def readBytes(data, length):
    return bytes(struct.unpack("<%iB" % (length), data.read(length)))
#end def

# -----------------------------------------------------------------------------
#
def readFloat(data):
    return struct.unpack('<f', data.read(4))[0]
#end def
    
# -----------------------------------------------------------------------------
#
def readVector(data):
    return Vector(struct.unpack('<3f', data.read(12)))
#end def
    
# -----------------------------------------------------------------------------
#
def readVector2(data):
    return struct.unpack('<2f', data.read(8))
#end def
    
# -----------------------------------------------------------------------------
#
def readVector4(data):
    return struct.unpack('<4f', data.read(16))
#end def
    
# -----------------------------------------------------------------------------
#
def readQuat(data):
    v = struct.unpack('<4f', data.read(16))
    qRot = Quaternion((v[3], v[0], v[1], v[2]))
    return qRot
#end def

# -----------------------------------------------------------------------------
#
def readColorByteRGBA(data):
    return struct.unpack('<4B', data.read(4))
#end def

# -----------------------------------------------------------------------------
#
def readString(data):
    string = ""
    while True:
        c = read8(data, False)
        if c == 0:
            break
            
        string += str(chr(c))
    #end while
        
    return string
#end def

# -----------------------------------------------------------------------------
#
def readStringCount(data, length):
    return struct.unpack("<%is" % (length), data.read(length))[0].decode("utf-8")
#end def

# -----------------------------------------------------------------------------
#
def write8(data, val, is_signed):
    data.write(val.to_bytes(1, 'little', signed=is_signed))
#end def

# -----------------------------------------------------------------------------
#
def write16(data, val, is_signed):
    data.write(val.to_bytes(2, 'little', signed=is_signed))
#end def
    
# -----------------------------------------------------------------------------
#
def write32(data, val, is_signed):
    data.write(val.to_bytes(4, 'little', signed=is_signed))
#end def

# -----------------------------------------------------------------------------
#  
def writeFloat(data, val):
    data.write(struct.pack('<f', val))
#end def

# -----------------------------------------------------------------------------
#  
def writeColorByteRGBA(data, color):
    data.write(struct.pack('<4B', color[0], color[1], color[2], color[3]))
#end def

# -----------------------------------------------------------------------------
#  
def writeString(data, str):
    data.write(bytes(str, 'utf-8')+b'\0')
#end def
    
# -----------------------------------------------------------------------------
#  
def writeBytes(data, bytesObj):
    data.write(bytesObj)
#end def
   
# -----------------------------------------------------------------------------
#   
def writeVector(data, vec):
    data.write(struct.pack('<3f', vec.x, vec.y, vec.z))
#end def

# -----------------------------------------------------------------------------
#   
def writeVector2(data, vec):
    data.write(struct.pack('<2f', vec[0], vec[1]))
#end def
    
# -----------------------------------------------------------------------------
#   
def writeVector4(data, vec):
    data.write(struct.pack('<4f', vec[0], vec[1], vec[2], vec[3]))
#end def
    
# -----------------------------------------------------------------------------
#   
def writeQuat(data, quat):
    data.write(struct.pack('<4f', quat.x, quat.y, quat.z, quat.w))
#end def

# -----------------------------------------------------------------------------
# Returns a set of all file paths found in the directory and its sub directories
# dirName: directory to start search for files and folders
# ext: str or tuple of str eg. ".txt" or (".txt", ".json")
def get_files_in_dir(dirName, ext = ''):
    if not ext:
        return {os.path.realpath(os.path.join(root, name)) for root, dirs, files in os.walk(dirName) for name in files}
    #end if
    
    return {os.path.realpath(os.path.join(root, name)) for root, dirs, files in os.walk(dirName) for name in files if name.endswith(ext)}
#end def

# -----------------------------------------------------------------------------
# Returns a set of all relative file paths found in the directory and its sub directories
# dirName: directory to start search for files and folders
# relDir: directory to make file path relative to
# ext: str or tuple of str eg. ".txt" or (".txt", ".json")
def get_files_in_dir_relpath(dirName, relDir, ext = ''):
    if not ext:
        return {os.path.relpath(os.path.join(root, name), relDir) for root, dirs, files in os.walk(dirName) for name in files}
    #end if
    
    return {os.path.relpath(os.path.join(root, name), relDir) for root, dirs, files in os.walk(dirName) for name in files if name.endswith(ext)}
#end def

# -----------------------------------------------------------------------------
# save class annotation key/values to json dictionary file
def save_class_to_json(path, classObj, ignore_keys=[]):
    keys = classObj.__annotations__.keys() - ignore_keys
    data = {k:classObj.get(k, classObj.__annotations__[k][1].get("default")) for k in keys}
    save_json(path, data)
#end def

# -----------------------------------------------------------------------------
# Sets classObj keys to the loaded json dictionary objects keys if classObj has those keys
# Returns true if successfully loaded the file. 
def load_json_to_class(path, classObj, ignore_keys=[]):
    keys = classObj.__annotations__.keys() - ignore_keys
    data = load_json(path)
    if not data:
        return False
    #end if
    for key in data:
        if key in keys:
            classObj[key] = data.get(key)
        #end if
    #end for
    return True
#end def

# -----------------------------------------------------------------------------
# save json dictionary
def save_json(path, data):
    try:
        with create_save_file(path) as file:
            file.write(json.dumps(data, indent=4))
        #end with
    except:
        pass
    #end try except
#end def

# -----------------------------------------------------------------------------
# loads json to dictionary and returns that dictionary or an empty dictionary if error
def load_json(path):
    try:
        with open(path, 'r') as file:
            return json.load(file)
        #end with
    except:
        e = sys.exc_info()
        print("[Turok 2 EX Tools] Failed to load json file %s: %s %s" % (path, e[0], e[1]))
        return {}
    #end try except
#end def

# -----------------------------------------------------------------------------
# Returns the file object. Makes dirs to path if they don't exist.
def create_save_file(path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return open(path, 'w', encoding='utf-8')
#end def

# -----------------------------------------------------------------------------
# Open windows explorer (if on windows)
def open_folder(path):
    try:
        if os.path.isdir(path) and os.name == "nt":
            os.startfile(path)
    except:
        pass
    #end try except
#end def

# -----------------------------------------------------------------------------
# Returns dictionary object of the kex file contents using user defined props
#
# Special Arg Type: 'ANY' - Checks for 'CPP_ID', 'CPP_INTEGER', 'CPP_FLOAT', 'CPP_STRING'
#
# Prop Types:
# 'OBJ_DEF'    - Requires a callback function, return type is (object, object key name)
# 'OBJ_END'    - Calls callback func (if any) then pops the last object off the chain
# 'PROP'       - If theres no callback func then sets obj[tok[0]] = tok[1]. if only has 1 arg then assumes bool obj[tok[0]] = True
# 'PROP_SVEC'  - If theres no callback func then if tok[1] type is string will extract the vector from it
# 'PROP_SVEC4' - If theres no callback func then if tok[1] type is string will extract the vector from it
# 'PROP_LIST'  - Requires a callback function(adds listIndex to third param) and arg 0 type must be ('CPP_ID', "startMatch"). Assumes after the match in the string is an integer value.
def parse_kex_file(filepath, props, rootDict):
    p = utils.create_preprocessor()
    with open(filepath, 'r', encoding='utf-8') as kexFile:
        p.parse(kexFile)
    #end with
    
    anyArgTypes = {'CPP_ID', 'CPP_INTEGER', 'CPP_FLOAT', 'CPP_STRING'}
    objChain = [(rootDict, 'ROOT')]
    argIndex = 0
    propTokens = [] #list of prop tokens gone through from current prop
    propIndex = -1
    listIndex = -1
    searchContinuePropIndex = -1
    searchContinuePropTokens = []
    while True:
        if searchContinuePropTokens:
            tok = searchContinuePropTokens.pop(0)
            # print('RESEARCH LexToken(%s,%r,%d,%d)' % (tok.type, tok.value, tok.lineno, tok.lexpos))
        else:
            tok = p.token()
            if not tok:
                break
            elif tok.type == 'CPP_WS':
                continue
            #end if
            # print('LexToken(%s,%r,%d,%d)' % (tok.type, tok.value, tok.lineno, tok.lexpos))
        #end if
        # print('LexToken(%s,%r,%d,%d)' % (tok.type, tok.value, tok.lineno, tok.lexpos))

        lastObj = objChain[-1][0]
        lastObjKey = objChain[-1][1]
        lastObjProps = props[lastObjKey]
        searchProps = lastObjProps
        startPropIndex = 0
        if propIndex >= 0:
            searchProps = [lastObjProps[propIndex]]
            startPropIndex = propIndex
        elif searchContinuePropIndex >= 0:
            startPropIndex = searchContinuePropIndex
            searchProps = lastObjProps[searchContinuePropIndex:]
            searchContinuePropIndex = -1
        #end if
        matchedArg = False
        
        for i, prop in enumerate(searchProps):
            propType = prop[0]
            propFunc = prop[1]
            propArgs = prop[2]
            argType = ''
            argValidValue = ''
            
            #Check if theres a match for this prop and set argType and argValidValue
            if isinstance(propArgs[argIndex], tuple):
                argType = propArgs[argIndex][0]
                argValidValue = propArgs[argIndex][1]
                if tok.type == argType or (argType == 'ANY' and tok.type in anyArgTypes):
                    if propType == 'PROP_LIST' and argIndex == 0 and argType == 'CPP_ID':
                        #check if it starts with argValidValue
                        index = tok.value.find(argValidValue)
                        if index >= 0:
                            listIndex = max(int(tok.value[index+len(argValidValue):]), 0)
                            matchedArg = True
                        #end if
                    elif tok.value == argValidValue:
                        matchedArg = True
                    #end if
                #end if
            else:
                argType = propArgs[argIndex]
                if tok.type == argType or (argType == 'ANY' and tok.type in anyArgTypes):
                    matchedArg = True
                #end if
            #end if
            
            #if match was found
            if matchedArg:
                if propIndex < 0:
                    propIndex = startPropIndex + i
                #end if
                propTokens.append(tok)
                argIndex += 1
                if argIndex >= len(propArgs):
                    if propFunc:
                        propFuncResult = None
                        if propType == 'PROP_LIST':
                            propFuncResult = propFunc(objChain, lastObj, listIndex, propTokens)
                        else:
                            propFuncResult = propFunc(objChain, lastObj, propTokens)
                        #end if
                        if propType == 'OBJ_DEF' and propFuncResult is not None:
                            objChain.append(propFuncResult)
                        #end if
                    else:
                        if propType == 'PROP' or (propType in ['PROP_SVEC', 'PROP_SVEC4'] and argType != 'CPP_STRING'):
                            if len(propArgs) == 1:
                                lastObj[propTokens[0].value] = True
                            elif len(propArgs) == 2:
                                if argType == 'CPP_STRING' or (argType == 'ANY' and isinstance(propTokens[1].value, str)):
                                    lastObj[propTokens[0].value] = utils.string_remove_quotes(propTokens[1].value)
                                else:
                                    lastObj[propTokens[0].value] = propTokens[1].value
                                #end if
                            #end if
                        elif propType == 'PROP_SVEC':
                            lastObj[propTokens[0].value] = utils.get_vector_from_string(propTokens[1].value)
                        elif propType == 'PROP_SVEC4':
                            lastObj[propTokens[0].value] = utils.get_vector4_from_string(propTokens[1].value)
                        #end if
                    #end if
                    if propType == 'OBJ_END':
                        if len(objChain) > 1:
                            objChain.pop()
                        #end if
                    #end if
                    propTokens.clear()
                    argIndex = 0
                    propIndex = -1
                    listIndex = -1
                #end if
                
                break
            else: #no match found with this prop
                pass
                # if propIndex >= 0:
                    # propTokens.append(tok)
                    # searchContinuePropIndex = propIndex + 1
                    # searchContinuePropTokens[:] = propTokens #add propTokens to the beginning of the list
                    # #searchContinuePropTokens.extend(propTokens)
                    # propTokens.clear()
                    # argIndex = 0
                    # propIndex = -1
                    # listIndex = -1
                    # break
                # #end if
            #end if
        #end for
        if not matchedArg:
            if propIndex >= 0:
                propTokens.append(tok)
                searchContinuePropIndex = propIndex + 1
                searchContinuePropTokens[:] = propTokens #add propTokens to the beginning of the list
            #end if
            propTokens.clear()
            argIndex = 0
            propIndex = -1
            listIndex = -1
        #end if
    #end while
    
    # print("")
    # print(objChain[0][0])
    # print("")
    
    return objChain[0][0]
#end def
