import bpy
import os
import blf
import bgl
import gpu
import bmesh
import aud
import bpy_extras.view3d_utils
import time
from bpy.app.handlers import persistent
from gpu_extras.batch import batch_for_shader

from . import utils
from . import panel_preferences
from . import aud_utils

# -----------------------------------------------------------------------------
# Update event
@persistent
def OnTimerTick():
    #Calculate time and delta time
    millis = time.monotonic()
    utils.deltaTime = millis - utils.curTime
    utils.curTime = millis

    #Update Sounds
    playingSoundsDeleteKeys = []
    deviceLocked = False
    for key, soundList in aud_utils.playingSounds.items():
        needRemoval = False
        for sound in soundList:
            sound.update()
            #Check if sound is finished playing
            if sound.doRemove:
                needRemoval = True
            #end if
            
            #update sound delaying
            if sound.isDelaying and utils.time_reached(sound.delayTime):
                if not deviceLocked:
                    aud.Device().lock()
                    deviceLocked = True
                #end if
                sound.delay_finished()
            #end if
        #end for
        
        #Remove sounds that are finished playing
        if needRemoval:
            soundList[:] = [sound for sound in soundList if not sound.doRemove]
            if len(aud_utils.playingSounds[key]) == 0:
                playingSoundsDeleteKeys.append(key)
            #end if
        #end if
    #end for
    
    #Delete empty keys
    for key in playingSoundsDeleteKeys:
        del aud_utils.playingSounds[key]
    #end for
    
    if deviceLocked:
        aud.Device().unlock()
    #end if
    
    return 0.0166666
#end def

# -----------------------------------------------------------------------------
# 
def update_frame_fps():
    if not utils.prefs().fps60playback:
        return 
    #end if
    
    obj = bpy.context.object
    if obj is None or obj.t2.objType != "SKINNEDMESH" or obj.type != "ARMATURE" or obj.animation_data is None or obj.animation_data.action is None:
        return
    #end if
    
    screen = bpy.context.screen
    animAction = obj.animation_data.action
    frameStart, frameEnd = animAction.frame_range
    if screen.is_animation_playing:
        if animAction.t2.speed <= 60:
            new_frame_start = int(frameStart)
            new_frame_end = int(frameEnd * (60.0 / animAction.t2.speed))
            new_fps = 60
            new_fps_base = 1.0
            new_frame_map_old = int((animAction.t2.speed / 60.0) * 600)
            new_frame_map_new = 600
        else:
            new_frame_start = int(frameStart)
            new_frame_end = int(frameEnd)
            new_fps = int(animAction.t2.speed)
            new_fps_base = 1.0
            new_frame_map_old = 100
            new_frame_map_new = 100
        #end if
    else:
        new_frame_start = int(frameStart)
        new_frame_end = int(frameEnd)
        new_fps = int(animAction.t2.speed)
        new_fps_base = 1.0
        new_frame_map_old = 100
        new_frame_map_new = 100
    #end if
    
    scene = bpy.context.scene
    #Update scene values only if they changed
    if scene.frame_preview_start != new_frame_start:
        scene.frame_preview_start = new_frame_start
    #end if
    if scene.frame_preview_end != new_frame_end:
        scene.frame_preview_end = new_frame_end
    #end if
    if scene.frame_start != new_frame_start:
        scene.frame_start = new_frame_start
    #end if
    if scene.frame_end != new_frame_end:
        scene.frame_end = new_frame_end
    #end if
    if scene.render.fps != new_fps:
        scene.render.fps = new_fps
    #end if
    if scene.render.fps_base != new_fps_base:
        scene.render.fps_base = new_fps_base
    #end if
    if scene.render.frame_map_old != new_frame_map_old:
        scene.render.frame_map_old = new_frame_map_old
    #end if
    if scene.render.frame_map_new != new_frame_map_new:
        scene.render.frame_map_new = new_frame_map_new
    #end if
#end def

# -----------------------------------------------------------------------------
# 
def handler_main_pre_update(isUndo=False, isRedo=False):
    pass
#end def

# -----------------------------------------------------------------------------
# 
def handler_main_post_update(isUndo=False, isRedo=False):
    update_frame_fps()
#end def

# -----------------------------------------------------------------------------
# Called after whenever a new file is loaded (basically OnStart)
@persistent
def LoadPostHandler(dummy):
    pass
#end def

# -----------------------------------------------------------------------------
# Called before whenever something changes
@persistent
def DepsgraphPreHandler(scene):
    handler_main_pre_update(False, False)
#end def

# -----------------------------------------------------------------------------
# Called After whenever something changes
@persistent
def DepsgraphPostHandler(scene):
    handler_main_post_update(False, False)
#end def

# -----------------------------------------------------------------------------
# Called After Undo
@persistent
def UndoPostHandler(scene):
    handler_main_post_update(True, False)
#end def

# -----------------------------------------------------------------------------
# Called After Redo
@persistent
def RedoPostHandler(scene):
    handler_main_post_update(False, True)
#end def

# -----------------------------------------------------------------------------
# Called on frame change for playback and rendering (before)
@persistent
def FrameChangePreHandler(scene):
    update_frame_fps()
#end def

# -----------------------------------------------------------------------------
# 
def add_handler(handler, func):
    remove_handler(handler, func)
    handler.append(func)
#end def    
# -----------------------------------------------------------------------------
# 
def remove_handler(handler, func):
    for i in range(len(handler)-1, -1, -1):
        if handler[i].__name__.startswith(func.__name__):
            del handler[i]
        #end if
    #end for
#end def    

# -----------------------------------------------------------------------------
#
def register():
    add_handler(bpy.app.handlers.load_post, LoadPostHandler)
    add_handler(bpy.app.handlers.depsgraph_update_pre, DepsgraphPreHandler)
    add_handler(bpy.app.handlers.depsgraph_update_post, DepsgraphPostHandler)
    add_handler(bpy.app.handlers.undo_post, UndoPostHandler)
    add_handler(bpy.app.handlers.redo_post, RedoPostHandler)
    add_handler(bpy.app.handlers.frame_change_pre, FrameChangePreHandler)
    if not bpy.app.timers.is_registered(OnTimerTick):
        utils.curTime = time.monotonic()
        bpy.app.timers.register(OnTimerTick, first_interval=1.0, persistent=True)
    #end if
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    remove_handler(bpy.app.handlers.load_post, LoadPostHandler)
    remove_handler(bpy.app.handlers.depsgraph_update_pre, DepsgraphPreHandler)
    remove_handler(bpy.app.handlers.depsgraph_update_post, DepsgraphPostHandler)
    remove_handler(bpy.app.handlers.undo_post, UndoPostHandler)
    remove_handler(bpy.app.handlers.redo_post, RedoPostHandler)
    remove_handler(bpy.app.handlers.frame_change_pre, FrameChangePreHandler)
    if bpy.app.timers.is_registered(OnTimerTick):
        bpy.app.timers.unregister(OnTimerTick)
    #end if
#end def
