import bpy
import os
import random
import pickle
import sys

from bpy_extras.wm_utils.progress_report import (
    ProgressReport,
    ProgressReportSubstep,
)
from . import (
    utils,
    file_utils,
    paths,
    kmaterial,
)

defs = {}

# -----------------------------------------------------------------------------
# 
def OnStart():
    global defs
    
    defs["anims"] = {}
    defs["anim_events"] = {}
    defs["user_events"] = { "events": {}, "lists": {} }
    defs["impacts"] = {}
    defs["hotpoints"] = {}
    defs["builtin_kmats"] = {}
    defs["texsets"] = {}
    defs["voices"] = {}
    defs["actors"] = {}
    defs["particles"] = {}
    defs["sounds"] = {}
    
    load_cached_defs()
#end def

# -----------------------------------------------------------------------------
# Loads all the custom and quick loading defs
def load_defs_base(cacheDefs=True):
    global defs
    
    with ProgressReport(bpy.context.window_manager) as progress:
        progress.enter_substeps(1)

        #Load custom def files with keys of type int
        defs["anims"] = {int(k):v for k,v in file_utils.load_json(paths.def_anims()).items()}
        defs["anim_events"] = {int(k):v for k,v in file_utils.load_json(paths.def_animevents()).items()}
        defs["impacts"] = {int(k):v for k,v in file_utils.load_json(paths.def_impacts()).items()}
        defs["hotpoints"] = {int(k):v for k,v in file_utils.load_json(paths.def_hotpoints()).items()}
        defs["builtin_kmats"] = kmaterial.load(paths.def_builtin_kmats())
        
        defs["texsets"] = {}
        textureSetInfoPath = paths.find_t2_file_path(os.path.join("defs", "textureSetInfo.txt"))
        if textureSetInfoPath is not None:
            print("Loading Turok 2 TextureSets %s ..." % (textureSetInfoPath))
            defs["texsets"] = load_texturesetinfo(textureSetInfoPath)
        #end if
        
        defs["voices"] = {}
        voiceInfoPath = paths.find_t2_file_path(os.path.join("defs", "voiceInfo.txt"))
        if voiceInfoPath is not None:
            print("Loading Turok 2 VoiceInfo %s ..." % (voiceInfoPath))
            defs["voices"] = load_voiceinfo(voiceInfoPath)
        #end if
        
        defs["user_events"] = { "events": {}, "lists": {} }
        editorUserEventsPath = paths.find_t2_file_path(os.path.join("blender", "userEvents.txt"))
        if editorUserEventsPath is not None:
            print("Loading Turok 2 Editor User Events %s ..." % (editorUserEventsPath))
            defs["user_events"] = load_editor_userevents(editorUserEventsPath)
        #end if

        editorHotpointsPath = paths.find_t2_file_path(os.path.join("blender", "hotpoints.txt"))
        if editorHotpointsPath is not None:
            print("Loading Turok 2 Editor Hot Points %s ..." % (editorHotpointsPath))
            defs["hotpoints"].update(load_editor_hotpoints(editorHotpointsPath))
        #end if
        
        progress.leave_substeps("Finished! (TextureSets:%s Voices:%s AnimNames:%s AnimEvents:%s Impacts:%s Hotpoints:%s BuiltinKmats:%s)" % (len(defs["texsets"]), len(defs["voices"]), len(defs["anims"]), len(defs["anim_events"]), len(defs["impacts"]), len(defs["hotpoints"]), len(defs["builtin_kmats"])))
    #end with
    
    if cacheDefs:
        save_defs_to_cache()
    #end if
#end def

# -----------------------------------------------------------------------------
# 
def load_defs_actors(cacheDefs=True):
    global defs
    
    with ProgressReport(bpy.context.window_manager) as progress:
        progress.enter_substeps(1)

        print("Loading Turok 2 Actor Defs ...")
        defs["actors"] = {}
        actorDefPaths = paths.find_t2_files_in_dir(os.path.join("defs", "actors"), '.txt')
        progress.enter_substeps(len(actorDefPaths))
        for actorDefPath in actorDefPaths:
            actors = load_actorinfo(actorDefPath)
            defs["actors"].update(actors)
            progress.step()
        #end for
        progress.leave_substeps("Finished! (Files:%s Actors:%s)" % (len(actorDefPaths), len(defs["actors"])))
    #end with
    if cacheDefs:
        save_defs_to_cache()
    #end if
#end def

# -----------------------------------------------------------------------------
# 
def load_defs_particles(cacheDefs=True):
    global defs
    
    with ProgressReport(bpy.context.window_manager) as progress:
        progress.enter_substeps(1)

        print("Loading Turok 2 Particles ...")
        defs["particles"] = {}
        particlePaths = paths.find_t2_files_in_dir("particles", '.particle')
        progress.enter_substeps(len(particlePaths))
        for path in particlePaths:
            particle = load_particle(path)
            if particle:
                defs["particles"][particle["id"]] = particle
            #end if
            progress.step()
        #end for
        progress.leave_substeps("Finished! (Particles:%s)" % (len(defs["particles"])))
    #end with
    if cacheDefs:
        save_defs_to_cache()
    #end if
#end def

# -----------------------------------------------------------------------------
# 
def load_defs_sounds(cacheDefs=True):
    global defs
    
    with ProgressReport(bpy.context.window_manager) as progress:
        progress.enter_substeps(1)

        print("Loading Turok 2 Sound Shaders ...")
        defs["sounds"] = {}
        soundPaths = list(paths.find_t2_files_in_dir(os.path.join("sounds", "shaders"), '.ksnd'))
        soundPaths.sort()
        progress.enter_substeps(len(soundPaths))
        for path in soundPaths:
            sound = load_sound(path)
            if sound:
                #same ids get added to the sounds list
                if sound["id"] not in defs["sounds"]:
                    defs["sounds"][sound["id"]] = []
                #end if
                sound["index"] = len(defs["sounds"][sound["id"]])
                defs["sounds"][sound["id"]].append(sound)
            #end if
            progress.step()
        #end for
        progress.leave_substeps("Finished! (Files:%s SoundIDs:%s)" % (len(soundPaths), len(defs["sounds"])))
    #end with
    if cacheDefs:
        save_defs_to_cache()
    #end if
#end def

# -----------------------------------------------------------------------------
# Load all def files
def load_all(cacheDefs=True):
    load_defs_base(False)
    load_defs_actors(False)
    load_defs_particles(False)
    load_defs_sounds(False)
    if cacheDefs:
        save_defs_to_cache()
    #end if
#end def

# -----------------------------------------------------------------------------
def load_cached_defs():
    global defs
    
    try:
        with open(paths.custom_defcache(), 'rb') as pickleFile:
            defs = pickle.load(pickleFile)
        #end with
    except FileNotFoundError as e:
        pass
    except:
        e = sys.exc_info()
        print("[Turok 2 EX Tools] Failed to load cached defs file %s: %s %s" % (paths.custom_defcache(), e[0], e[1]))
    #end try except
#end def

# -----------------------------------------------------------------------------
def save_defs_to_cache():
    global defs
    
    try:
        with open(paths.custom_defcache(), 'wb') as pickleFile:
            pickle.dump(defs, pickleFile, pickle.HIGHEST_PROTOCOL)
        #end with
    except:
        e = sys.exc_info()
        print("[Turok 2 EX Tools] Failed to save defs to cache file %s: %s %s" % (paths.custom_defcache(), e[0], e[1]))
    #end try except
#end def

# -----------------------------------------------------------------------------
# Return all texturesets in the file as a dictionary of "texture Name", textureSetInfo.
def load_texturesetinfo(filepath):
    def on_create_textureset(objChain, texturesets, tokens):
        texPath = utils.string_remove_quotes(tokens[0].value)
        texturesets[texPath] = {
            "speed": 1,
            "animated": 0,
            "selectable": 0,
            "textures": [texPath],
        }
        return (texturesets[texPath], 'TEXTURESET')
    #end def
    
    def on_add_texturepath(objChain, texset, index, tokens):
        while (len(texset["textures"]) < index + 1):
            texset["textures"].append('')
        #end while
        texset["textures"][index] = utils.string_remove_quotes(tokens[1].value)
    #end def
    
    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_textureset, ['CPP_STRING', 'CPP_LCURLY']), # "textures/footy.png" {
        ],
        'TEXTURESET': [
            ('PROP',        None,               [('CPP_ID', 'speed'), 'CPP_INTEGER']), # speed 1
            ('PROP',        None,               [('CPP_ID', 'animated'), 'CPP_INTEGER']), # animated 1
            ('PROP',        None,               [('CPP_ID', 'selectable'), 'CPP_INTEGER']), # selectable 1
            ('PROP_LIST',   on_add_texturepath, [('CPP_ID', 'texture_'), 'CPP_STRING']), # texture_1 "textures/blood/BLAVB02.PNG"
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
        ],
    }
    
    return file_utils.parse_kex_file(filepath, props, {})
#end def

# -----------------------------------------------------------------------------
# Return all texturesets in the file as a dictionary of voiceID, voiceInfo.
def load_voiceinfo(filepath):
    def on_create_voiceinfo(objChain, voices, tokens):
        voiceID = tokens[1].value
        voices[voiceID] = { "name": '' }
        return (voices[voiceID], 'VOICE')
    #end def
    
    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_voiceinfo, ['CPP_ID', 'CPP_INTEGER', 'CPP_LCURLY']), # Voice 401 {
        ],
        'VOICE': [
            ('PROP',        None,               [('CPP_ID', 'name'), 'CPP_STRING']), # name "sounds/voices/PORTOFADIA"
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
        ],
    }
    
    return file_utils.parse_kex_file(filepath, props, {})
#end def

# -----------------------------------------------------------------------------
# Return the particle in the file as a dictionary object.
def load_particle(filepath):
    def on_create_fxlist(objChain, particle, tokens):
        particle["id"] = 0
        particle["fx_length"] = tokens[2].value
        particle["fx"] = []
        
        return (particle["fx"], 'FXLIST')
    #end def
    
    def on_set_particle_prop(objChain, fxList, tokens):
        particle = objChain[0][0]
        key = tokens[0].value
        if key == "typeID":
            key = "id"
        #end if
        if key not in ["fx_length", "fx"]:
            if isinstance(tokens[2].value, str):
                particle[key] = utils.string_remove_quotes(tokens[2].value)
            else:
                particle[key] = tokens[2].value
            #end if
        #end if
    #end def
    
    def on_add_fx(objChain, fxList, tokens):
        fx = {
            "local": { "impact": {} },
            "network": {},
        }
        fxList.append(fx)
        
        return (fx, 'FX')
    #end def
    
    def on_create_fx_local(objChain, fx, tokens):
        return (fx["local"], 'FX_LOCAL')
    #end def
    
    def on_create_fx_network(objChain, fx, tokens):
        fx["network"]["impact"] = {}
        return (fx["network"], 'FX_NETWORK')
    #end def
    
    def on_set_fx_prop_both(objChain, fx, tokens):
        on_set_fx_prop_local(objChain, fx["local"], tokens)
    #end def

    def on_set_fx_prop_local(objChain, fxLocal, tokens):
        set_obj_prop(fxLocal, tokens)
    #end def
    
    def on_set_fx_prop_network(objChain, fxNetwork, tokens):
        set_obj_prop(fxNetwork, tokens)
    #end def
    
    def set_obj_prop(obj, tokens):
        if isinstance(tokens[2].value, str):
            obj[tokens[0].value] = utils.string_remove_quotes(tokens[2].value)
        else:
            obj[tokens[0].value] = tokens[2].value
        #end if
    #end def
    
    def on_set_fx_propvec_both(objChain, fx, tokens):
        on_set_fx_propvec_local(objChain, fx["local"], tokens)
    #end def
    
    def on_set_fx_propvec_local(objChain, fxLocal, tokens):
        set_obj_propvec(fxLocal, tokens)
    #end def
    
    def on_set_fx_propvec_network(objChain, fxNetwork, tokens):
        set_obj_propvec(fxNetwork, tokens)
    #end def

    def set_obj_propvec(obj, tokens):
        obj[tokens[0].value] = (tokens[3].value, tokens[4].value, tokens[5].value)
    #end def
    
    def on_create_impactlist_both(objChain, fx, tokens):
        return (fx["local"]["impact"], 'IMPACTLIST')
    #end def
    
    def on_create_impactlist_local(objChain, fxLocal, tokens):
        return (fxLocal["impact"], 'IMPACTLIST')
    #end def
    
    def on_create_impactlist_network(objChain, fxNetwork, tokens):
        return (fxNetwork["impact"], 'IMPACTLIST')
    #end def
    
    def on_create_impact(objChain, impacts, tokens):
        impactID = tokens[1].value
        impacts[impactID] = {}
        
        return (impacts[impactID], 'IMPACT')
    #end def
    
    def on_set_impact_params(objChain, impact, tokens):
        impact["params"] = (tokens[3].value, tokens[4].value, tokens[5].value, tokens[6].value, tokens[7].value, tokens[8].value, tokens[9].value, tokens[10].value)
    #end def
    
    def on_set_impact_prop(objChain, impact, tokens):
        if isinstance(tokens[2].value, str):
            impact[tokens[0].value] = utils.string_remove_quotes(tokens[2].value)
        else:
            impact[tokens[0].value] = tokens[2].value
        #end if
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_fxlist, ['CPP_ID', 'CPP_LBRACKET', 'CPP_INTEGER', 'CPP_RBRACKET', 'CPP_EQUAL', 'CPP_LCURLY']), # fx[3] = {
        ],
        'FXLIST': [
            ('OBJ_END',     None,                   ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_add_fx,              ['CPP_LCURLY']), # {
            ('PROP',        on_set_particle_prop,   ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any particle prop): typeID = 444
        ],
        'FX': [
            ('OBJ_END',     None,                       ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_create_fx_local,         ['CPP_LBRACKET', ('CPP_ID', 'local'), 'CPP_RBRACKET', 'CPP_LCURLY']), # [local] {
            ('OBJ_DEF',     on_create_fx_network,       ['CPP_LBRACKET', ('CPP_ID', 'network'), 'CPP_RBRACKET', 'CPP_LCURLY']), # [network] {
            ('OBJ_DEF',     on_create_impactlist_both,  [('CPP_ID', 'onImpact'), 'CPP_LCURLY']), # onImpact {
            ('PROP',        on_set_fx_prop_both,        ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any fx prop): light_color = { 1.000000 1.000000 0.501961 }
            ('PROP',        on_set_fx_propvec_both,     ['CPP_ID', 'CPP_EQUAL', 'CPP_LCURLY', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_RCURLY']), # translation = { 5.093750 20.375000 -2.046874 }
        ],
        'FX_LOCAL': [
            ('OBJ_END',     None,                       ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_create_impactlist_local, [('CPP_ID', 'onImpact'), 'CPP_LCURLY']), # onImpact {
            ('PROP',        on_set_fx_propvec_local,    ['CPP_ID', 'CPP_EQUAL', 'CPP_LCURLY', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_RCURLY']), # translation = { 5.093750 20.375000 -2.046874 }
            ('PROP',        on_set_fx_prop_local,       ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any fx prop): light_color = { 1.000000 1.000000 0.501961 }
        ],
        'FX_NETWORK': [
            ('OBJ_END',     None,                           ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_create_impactlist_network,   [('CPP_ID', 'onImpact'), 'CPP_LCURLY']), # onImpact {
            ('PROP',        on_set_fx_propvec_network,      ['CPP_ID', 'CPP_EQUAL', 'CPP_LCURLY', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_RCURLY']), # translation = { 5.093750 20.375000 -2.046874 }
            ('PROP',        on_set_fx_prop_network,         ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any fx prop): light_color = { 1.000000 1.000000 0.501961 }
        ],
        'IMPACTLIST': [
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_create_impact,   ['CPP_LBRACKET', 'CPP_INTEGER', 'CPP_RBRACKET', 'CPP_LCURLY']), # [0] {
        ],
        'IMPACT': [
            ('OBJ_END',     None,                   ['CPP_RCURLY']), # }
            ('PROP',        on_set_impact_params,   [('CPP_ID', 'params'), 'CPP_EQUAL', 'CPP_LCURLY', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_FLOAT', 'CPP_RCURLY']), # params = { 1280.000000 1024.000000 0.000000 0.000000 0.000000 0.000000 0.000000 0.000000 }
            ('PROP',        on_set_impact_prop,     ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any impact prop besides params): soundID = 502
        ],
    }
    
    data = file_utils.parse_kex_file(filepath, props, {})
    if data:
        data["path"] = filepath
    #end if
    
    return data
#end def

# -----------------------------------------------------------------------------
# Returns a sound with the soundID. Since multiple sounds can have the same id
# you must specify the index of the sound in the list of sounds with that id.
# Returns None if soundID or subSoundIndex doesn't exist
#
# subSoundIndex: if set to negative value, will use a random sound from the list
def get_sound(soundID, subSoundIndex=0):
    if soundID in defs["sounds"]:
        if len(defs["sounds"][soundID]) > 0:
            if subSoundIndex < 0:
                subSoundIndex = random.randint(0, len(defs["sounds"][soundID]) - 1)
            #end if
            if subSoundIndex < len(defs["sounds"][soundID]):
                return defs["sounds"][soundID][subSoundIndex]
            #end if
        #end if
    #end if
    
    return None
#end def
# -----------------------------------------------------------------------------
# Return the sound in the file as a dictionary object.
def load_sound(filepath):
    def on_create_soundlist(objChain, sound, tokens):
        sound["id"] = 0
        sound["index"] = 0 #index in soundID list (set after loading all the sounds)
        sound["sfx_length"] = tokens[2].value
        sound["sfx"] = []
        sound["volume_min"] = 100
        sound["volume_max"] = 100
        sound["weight"] = 100
        
        return (sound["sfx"], 'SFXLIST')
    #end def
    
    def on_set_sound_prop(objChain, sfxList, tokens):
        sound = objChain[0][0]
        key = tokens[0].value
        if key == "type":
            key = "id"
        #end if
        if key not in ["sfx_length", "sfx"]:
            if isinstance(tokens[2].value, str):
                sound[key] = utils.string_remove_quotes(tokens[2].value)
            else:
                sound[key] = tokens[2].value
            #end if
        #end if
    #end def
    
    def on_add_sfx(objChain, sfxList, tokens):
        sfx = {
            'wavefile': '',
            'bOneInstancePerSource': 0,
            'bEnableVolumeEnvelope': 0,
            'bEnablePitchEnvelope': 0,
            'bVoice': 0,
            'bUpdateNoEnvelope': 0,
            'bFullVolume': 0,
            'bLimitToSource': 0,
            'bLooping': 0,
            'delay': 0,
            'priority': 100,
            'random': 1.0,
            'latentVolume': 1.0,
            'latentPitch': 1.0,
            'pitch_min': 1.0,
            'pitch_max': 1.0,
            'volume_envelope': {
                'max': 1.0,
                'start': 1.0,
                'end': 1.0,
                'duration': 0,
                'envStartTime': 0,
            },
            'pitch_envelope': {
                'max': 1.0,
                'start': 1.0,
                'end': 1.0,
                'duration': 0,
                'envStartTime': 0,
            },
        }
        sfxList.append(sfx)
        
        return (sfx, 'SFX')
    #end def
    
    def on_set_obj_prop(objChain, obj, tokens):
        if isinstance(tokens[2].value, str):
            obj[tokens[0].value] = utils.string_remove_quotes(tokens[2].value)
        else:
            obj[tokens[0].value] = tokens[2].value
        #end if
    #end def

    def on_set_volume_env(objChain, sfx, tokens):
        return (sfx['volume_envelope'], 'VOL_ENV')
    #end def
    
    def on_set_pitch_env(objChain, sfx, tokens):
        return (sfx['pitch_envelope'], 'PITCH_ENV')
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_soundlist, ['CPP_ID', 'CPP_LBRACKET', 'CPP_INTEGER', 'CPP_RBRACKET', 'CPP_EQUAL', 'CPP_LCURLY']), # sounds[5] =
        ],
        'SFXLIST': [
            ('OBJ_END',     None,                   ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_add_sfx,             ['CPP_LCURLY']), # {
            ('PROP',        on_set_sound_prop,      ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any sound prop): volume_min = 40
        ],
        'SFX': [
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_set_volume_env,  [('CPP_ID', 'volume_envelope'), 'CPP_EQUAL', 'CPP_LCURLY']), # volume_envelope = {
            ('OBJ_DEF',     on_set_pitch_env,   [('CPP_ID', 'pitch_envelope'), 'CPP_EQUAL', 'CPP_LCURLY']), # pitch_envelope = {
            ('PROP',        on_set_obj_prop,    ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any sfx prop): bOneInstancePerSource = 0
        ],
        'VOL_ENV': [
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
            ('PROP',        on_set_obj_prop,    ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any volume_envelope prop): max = 0.359375
        ],
        'PITCH_ENV': [
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
            ('PROP',        on_set_obj_prop,    ['CPP_ID', 'CPP_EQUAL', 'ANY']), # (Any pitch_envelope prop): max = 1.000000
        ],
    }
    
    data = file_utils.parse_kex_file(filepath, props, {})
    if data:
        data["path"] = filepath
    #end if
    
    return data
#end def

# -----------------------------------------------------------------------------
# Return all actors in the file as a dictionary of actorID, actorInfo.
def load_actorinfo(filepath):
    def on_create_actorinfo(objChain, actors, tokens):
        actorID = tokens[1].value
        actors[actorID] = {
            "name": tokens[0].value,
            "info": {
                "components": {}
            },
        }
        
        return (actors[actorID]["info"], 'ACTORINFO')
    #end def
    
    def on_create_component(objChain, actorInfo, tokens):
        componentName = utils.string_remove_quotes(tokens[1].value)
        actorInfo["components"][componentName] = {}
        
        return (actorInfo["components"][componentName], 'COMPONENT')
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_actorinfo, ['CPP_ID', 'CPP_INTEGER', 'CPP_LCURLY']), # Actorthing 401
        ],
        'ACTORINFO': [
            ('OBJ_END',     None,                   ['CPP_RCURLY']), # }
            ('OBJ_DEF',     on_create_component,    [('CPP_ID', 'Begin_Component'), 'CPP_STRING']), # Begin_Component "kexScriptComponent"
            ('PROP',        None,                   ['CPP_ID', 'ANY']), # AnyPropName AnyValue('CPP_ID', 'CPP_INTEGER', 'CPP_FLOAT', 'CPP_STRING')
        ],
        'COMPONENT': [
            ('OBJ_END',     None,                   [('CPP_ID', 'End_Component')]), # End_Component
            ('PROP',        None,                   ['CPP_ID', 'ANY']), # AnyProp AnyValue('CPP_ID', 'CPP_INTEGER', 'CPP_FLOAT', 'CPP_STRING')
        ],
    }
    
    data = file_utils.parse_kex_file(filepath, props, {})
    for key in data:
        data[key]["path"] = filepath
    #end for

    return data
#end def

# -----------------------------------------------------------------------------
# Return all Editor Hot Point Names
def load_editor_hotpoints(filepath):
    def on_set_hp_name(objChain, hotpointNames, tokens):
        bitIndex = tokens[0].value
        if bitIndex >= 0 and bitIndex < 32:
            hpName = utils.string_remove_quotes(tokens[1].value)
            if hpName:
                hotpointNames[bitIndex] = hpName
            #end if
        #end if
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_set_hp_name, ['CPP_INTEGER', 'CPP_STRING']), # 1 "On Fire"
        ],
    }
    
    data = file_utils.parse_kex_file(filepath, props, {})
    
    return data
#end def

# -----------------------------------------------------------------------------
# Return all Editor User Event Types and lists
def load_editor_userevents(filepath):
    def on_create_event(objChain, user_events, tokens):
        eventID = tokens[1].value
        
        user_events["events"][eventID] = {
            "name": utils.string_remove_quotes(tokens[2].value),
            "args": [],
        }
        
        return (user_events["events"][eventID], 'EVENTINFO')
    #end def
    
    def on_add_arg(objChain, eventInfo, index, tokens):
        while (len(eventInfo["args"]) < index):
            eventInfo["args"].append('')
        #end while
        eventInfo["args"][index-1] = utils.string_remove_quotes(tokens[1].value)
    #end def
    
    def on_create_list(objChain, user_events, tokens):
        listName = utils.string_remove_quotes(tokens[1].value)
        user_events["lists"][listName] = {}
        return (user_events["lists"][listName], 'LISTINFO')
    #end def
    
    def on_add_listitem(objChain, listInfo, tokens):
        listInfo[tokens[1].value] = utils.string_remove_quotes(tokens[2].value)
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_event,    [('CPP_ID', 'event'), 'CPP_INTEGER', 'CPP_STRING', 'CPP_LCURLY']), # event 100 "Name" {
            ('OBJ_DEF', on_create_list,     [('CPP_ID', 'list'), 'CPP_STRING', 'CPP_LCURLY']), # list "CharacterNames" {
        ],
        'EVENTINFO': [
            ('OBJ_END',     None,           ['CPP_RCURLY']), # }
            ('PROP_LIST',   on_add_arg,     [('CPP_ID', 'arg_'), 'CPP_STRING']), # arg_1 "Sound ID"
        ],
        'LISTINFO': [
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
            ('PROP',        on_add_listitem,    [('CPP_ID', 'item'), 'CPP_INTEGER', 'CPP_STRING']), # item 0 "Turok"
        ],
    }
    
    rootDict = { "events": {}, "lists": {} }
    data = file_utils.parse_kex_file(filepath, props, rootDict)
    
    return data
#end def

# -----------------------------------------------------------------------------
# Return the events arg name
def get_anim_kfa_eduserevent_argname(event, argIndex, args):
    #lookup names from User/Script Event
    if event == 22:
        edEventInfo = defs["user_events"]["events"].get(int(args[0]))
        if edEventInfo is not None:
            if argIndex == 0:
                return edEventInfo["name"]
            elif argIndex - 1 < len(edEventInfo["args"]):
                listName = edEventInfo["args"][argIndex - 1]
                if listName:
                    return listName
                #end if
            #end if
        #end if
    #end if
    
    return ""
#end def
# -----------------------------------------------------------------------------
# Return the events arg name
def get_anim_kfa_argname(event, argIndex, args):
    argName = get_anim_kfa_eduserevent_argname(event, argIndex, args)
    if argName:
        return argName
    #end if
    
    #lookup name from anim_events
    eventDef = defs["anim_events"].get(event)
    if eventDef and argIndex < len(eventDef["args"]):
        return eventDef["args"][argIndex]
    #end if
    
    return ""
#end def

# -----------------------------------------------------------------------------
# Return the events name + the first args lookup name for: Sound ID, Voice ID, Particle ID, Actor ID, Hot Point Bit, Impact Type
def get_anim_kfa_fullname(event, args):
    eventDef = defs["anim_events"].get(event)
    if eventDef:
        desc = eventDef["name"]
        if event == 22: #User/Script Event
            argDesc = get_anim_kfa_eduserevent_argname(event, 0, args)
        else:
            argDesc = get_anim_kfa_arg_lookup(event, 0, args)
        #end if
        if argDesc != "":
            return ("%s: %s" % (desc, argDesc))
        #end if
            
        return desc
    #end if
    
    return "Event " + str(event)
#end def

# -----------------------------------------------------------------------------
# Returns the args Sound ID, Voice ID, Particle ID, Actor ID, Hot Point Bit, Impact Type name
def get_anim_kfa_arg_lookup(event, argIndex, args):
    eventDef = defs["anim_events"].get(event)
    if eventDef:
        if argIndex < len(eventDef["args"]):
            argValue = int(args[argIndex])
            #argDesc = eventDef["args"][argIndex]
            argDesc = get_anim_kfa_argname(event, argIndex, args)
            if argDesc == "Sound ID":
                sound = defs["sounds"].get(argValue)
                if sound:
                    return os.path.splitext(os.path.split(sound[0]["path"])[1])[0]
                #end if
            elif argDesc == "Voice ID":
                voice = defs["voices"].get(argValue)
                if voice:
                    return os.path.split(voice["name"])[1]
                #end if
            elif argDesc == "Particle ID":
                particle = defs["particles"].get(argValue)
                if particle:
                    return os.path.splitext(os.path.split(particle["path"])[1])[0]
                #end if
            elif argDesc == "Hot Point Bit":
                return defs["hotpoints"].get(argValue, str(argValue))
            elif argDesc == "Actor ID":
                actor = defs["actors"].get(argValue)
                if actor:
                    return actor["name"]
                #end if
            elif argDesc == "Impact Type":
                return defs["impacts"].get(argValue, "")
            elif event == 22 and argIndex > 0: #User/Script Event
                listInfo = defs["user_events"]["lists"].get(argDesc)
                if listInfo is not None:
                    itemName = listInfo.get(argValue)
                    if itemName is not None:
                        return itemName
                    #end if
                #end if
            #end if
        #end if
    #end if
        
    return ""
#end def
