import os
import math
from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from bpy_extras.image_utils import load_image
from . import (
    utils,
    file_utils,
    paths,
    panel_material,
)

# -----------------------------------------------------------------------------
# Compares the blender material.t2 to the kmat (for floats need to compare with math.isclose)
# and if they are the same then returns True
def bmat_has_kmat(bmat, kmat):
    newkmat = bmat_to_kmat(bmat)
    return (kmat["params"] == newkmat["params"] and
            kmat["samplers"] == newkmat["samplers"] and
            kmat["blend"] == newkmat["blend"] and
            kmat["cull"] == newkmat["cull"] and
            kmat["fullbright"] == newkmat["fullbright"] and
            kmat["sort"] == newkmat["sort"] and
            kmat["surface_type"] == newkmat["surface_type"] and
            kmat["depthtest"] == newkmat["depthtest"] and
            math.isclose(kmat["reflectivePower"], newkmat["reflectivePower"], rel_tol=1e-5) and
            kmat["nodraw"] == newkmat["nodraw"] and
            kmat["nodepthmask"] == newkmat["nodepthmask"] and
            kmat["alphatest"] == newkmat["alphatest"] and
            kmat["alphafunc"] == newkmat["alphafunc"] and
            math.isclose(kmat["alphamask"], newkmat["alphamask"], rel_tol=1e-5) and
            kmat["color_environment"] == newkmat["color_environment"] and
            kmat["color_diffuse"] == newkmat["color_diffuse"])
#end def

# -----------------------------------------------------------------------------
# Converts a blender material to a kmaterial
def bmat_to_kmat(bmat):
    kmat = {
        "blend": bmat.t2.blend,
        "cull": bmat.t2.cull.lower(),
        "fullbright": bmat.t2.fullBright,
        "sort": bmat.t2.sort.lower(),
        "surface_type": bmat.t2.surfaceType.lower(),
        "depthtest": bmat.t2.depthTest,
        "reflectivePower": bmat.t2.reflectivePower,
        "nodraw": bmat.t2.noDraw,
        "nodepthmask": bmat.t2.noDepthMask,
        "alphatest": bmat.t2.alphaTest,
        "alphafunc": bmat.t2.alphaFunc.lower(),
        "alphamask": float(bmat.t2.alphaMask),
        "color_environment": (bmat.t2.colorEnvironment[0], bmat.t2.colorEnvironment[1], bmat.t2.colorEnvironment[2], bmat.t2.colorEnvironment[3]),
        "color_diffuse": (bmat.t2.colorDiffuse[0], bmat.t2.colorDiffuse[1], bmat.t2.colorDiffuse[2], bmat.t2.colorDiffuse[3]),
        "params": {},
        "samplers": [],
    }
    
    if bmat.t2.cull == "AUTO":
        if bmat.use_backface_culling:
            kmat["cull"] = "back"
        else:
            kmat["cull"] = "none"
        #end if
    #end if
    
    for item in bmat.t2.params:
        if item.name:
            pValue = item.value if item.value else "0"
            kmat["params"][item.name] = { "type": item.valueType, "value": pValue }
        #end if
    #end for
    
    #go through shaderTexNodes to get samplers
    for node in bmat.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeTexImage':
            sampler = {}
            kmat["samplers"].append(sampler)
            sampler["type"] = "diffuse"
            sampler["index"] = node.t2.samplerIndex
            if node.interpolation == 'Closest':
                sampler["filter"] = 'nearest'
            else:
                if node.t2.anisotropicFilter:
                    sampler["filter"] = 'linear_anisotropic'
                else:
                    sampler["filter"] = 'linear'
                #end if
            #end if
            if node.t2.mirrored:
                sampler["wrap"] = 'mirrored'
            else:
                if node.extension == 'EXTEND':
                    sampler["wrap"] = 'clamp'
                else:
                    sampler["wrap"] = 'repeat'
                #end if
            #end if
            sampler["path"] = ""
            if node.image is not None:
                sampler["path"] = paths.texture_filepath_to_local_path(node.image.filepath)
            #end if
        #end if
    #end for

    return kmat
#end def

# -----------------------------------------------------------------------------
# Converts kmaterial to a blender material.
def to_material(kmat, bmat):
    #bmat.name = kmat["name"] #utils.get_material_texset_name(kmat["name"])
    
    bmat.preview_render_type = 'FLAT'
    bmat.use_nodes = True
    bmat.specular_intensity = 0.0
    bmat.diffuse_color = kmat["color_diffuse"]
    bmat.t2.showConButtons = False
    bmat.t2.blend = kmat["blend"]
    if kmat["alphatest"]:
        bmat.blend_method = "CLIP"
        bmat.alpha_threshold = 0.3475
    elif kmat["blend"]:
        bmat.blend_method = "BLEND"
    else:
        bmat.blend_method = "OPAQUE"
    #end if
    if kmat["cull"] == "back":
        bmat.use_backface_culling = True
    else:
        bmat.use_backface_culling = False
    #end if
    
    if kmat["cull"] == "front":
        cullType = kmat["cull"].upper()
    else:
        cullType = "AUTO"
    #end if
    
    if utils.has_enum_type(panel_material.EnumMaterialCulls, cullType):
        bmat.t2.cull = cullType
    #end if
    bmat.t2.fullBright = kmat["fullbright"]
    bmat.t2.alphaTest = kmat["alphatest"]
    alphaFuncType = kmat["alphafunc"].upper()
    if utils.has_enum_type(panel_material.EnumMaterialAlphaFuncs, alphaFuncType):
        bmat.t2.alphaFunc = alphaFuncType
    #end if
    bmat.t2.alphaMask = kmat["alphamask"]
    sortType = kmat["sort"].upper()
    if utils.has_enum_type(panel_material.EnumMaterialSorts, sortType):
        bmat.t2.sort = sortType
    #end if
    bmat.t2.noDraw = kmat["nodraw"]
    bmat.t2.depthTest = kmat["depthtest"]
    bmat.t2.noDepthMask = kmat["nodepthmask"]
    bmat.t2.colorEnvironment = kmat["color_environment"]
    bmat.t2.colorDiffuse = kmat["color_diffuse"]
    bmat.t2.reflectivePower = kmat["reflectivePower"]
    bmat.t2.surfaceType = kmat["surface_type"].upper()
    bmat.t2.params.clear()
    for paramName in kmat["params"]:
        item = bmat.t2.params.add()
        item.selfMat = bmat
        item.name = paramName
        item.valueType = kmat["params"][paramName]['type']
        item.value = kmat["params"][paramName]['value']
    #end for
    
    bmat.node_tree.nodes.clear()
    moNode = bmat.node_tree.nodes.new("ShaderNodeOutputMaterial")
    moNode.location = [300, 300]

    bsdf = bmat.node_tree.nodes.new("ShaderNodeBsdfPrincipled")
    bsdf.location = [10, 300]
    bsdf.inputs['Specular'].default_value = 0.0
    
    bmat.node_tree.links.new(moNode.inputs['Surface'], bsdf.outputs['BSDF'])

    #add to unique textures used
    for sampler in kmat["samplers"]:
        #create blender texture shader node
        texName = os.path.split(sampler["path"])[1]
        if texName is None:
            texName = sampler["path"]
        #end if
        
        texShaderNode = bmat.node_tree.nodes.new('ShaderNodeTexImage')
        texShaderNode.t2.samplerIndex = sampler["index"]
        texShaderNode.label = texName
        texNodePosY = 300.0 + (sampler["index"] * 280.0)
        texShaderNode.location = [-280.0, texNodePosY]
        texShaderNode.projection = 'FLAT'
        texFilter = sampler["filter"].lower()
        texShaderNode.t2.anisotropicFilter = True
        if texFilter == "linear_anisotropic":
            texShaderNode.interpolation = 'Linear'
        elif texFilter == "linear":
            texShaderNode.interpolation = 'Linear'
            texShaderNode.t2.anisotropicFilter = False
        else:
            texShaderNode.interpolation = 'Closest'
        #end if
        
        texWrap = sampler["wrap"].lower()
        if texWrap == "clamp":
            texShaderNode.extension = 'EXTEND'
        elif texWrap == "repeat":
            texShaderNode.extension = 'REPEAT'
        else: #mirrored
            texShaderNode.extension = 'REPEAT'
            texShaderNode.t2.mirrored = True
        #end if
        
        #load unique image
        if sampler["path"] in utils.TUROK2_BUILTIN_TEXTURES:
            texImage = load_image(os.path.join(paths.addon_texture_dir(), sampler["path"] + ".png"), check_existing=True)
        else:
            texPath = paths.find_t2_file_path(sampler["path"])
            if texPath is not None:
                texImage = load_image(texPath, check_existing=True)
            else:
                texImage = load_image(os.path.join(paths.addon_texture_dir(), "_default.png"), check_existing=True)
            #end if
        #end if
        texShaderNode.image = texImage;
        
        if sampler["index"] == 0:
            bmat.node_tree.links.new(bsdf.inputs['Base Color'], texShaderNode.outputs['Color'])
            if bmat.t2.fullBright:
                bmat.node_tree.links.new(bsdf.inputs['Emission'], texShaderNode.outputs['Color'])
            #end if
            bmat.node_tree.links.new(bsdf.inputs['Alpha'], texShaderNode.outputs['Alpha'])
        #end if
    #end for
    
    utils.update_shader_groups(bmat)
    
    for shaderNode in bmat.node_tree.nodes:
        shaderNode.select = False
    #end for
#end def

# -----------------------------------------------------------------------------
# Saves a dictionary of kMaterials to a .kmat file
# kMaterials: Dictionary k:matname, v:kmaterial dictionary
def save(filepath, kMaterials):
    with open(filepath, 'w+t', encoding='utf-8') as file:
        for kmatName in kMaterials:
            kmat = kMaterials[kmatName]
            file.write("material %s {\n" % (kmatName))
            if kmat["blend"]:
                file.write("\tblend\n")
            if kmat["cull"] != "none":
                file.write("\tcull %s\n" % (kmat["cull"]))
            for paramName, param in kmat["params"].items():
                if paramName:
                    pValue = param["value"] if param["value"] else "0"
                    file.write('\tparam "%s" %i "%s"\n' % (paramName, param["type"], pValue))
                #end if
            #end for
            if kmat["fullbright"]:
                file.write("\tfullbright\n")
            file.write("\tsort %s\n" % (kmat["sort"]))
            file.write("\tsurface_type %s\n" % (kmat["surface_type"]))
            if kmat["depthtest"]:
                file.write("\tdepthtest\n")
            if kmat["reflectivePower"] > 0.0:
                file.write("\treflectivePower %f\n" % (kmat["reflectivePower"]))
            if kmat["nodraw"]:
                file.write("\tnodraw\n")
            if kmat["nodepthmask"]:
                file.write("\tnodepthmask\n")
            if kmat["alphatest"]:
                file.write("\talphatest\n")
                file.write("\talphafunc %s\n" % (kmat["alphafunc"]))
                file.write("\talphamask %f\n" % (kmat["alphamask"]))
            file.write('\tcolor_environment "%f %f %f %f"\n' % (kmat["color_environment"][0], kmat["color_environment"][1], kmat["color_environment"][2], kmat["color_environment"][3]))
            file.write('\tcolor_diffuse "%f %f %f %f"\n' % (kmat["color_diffuse"][0], kmat["color_diffuse"][1], kmat["color_diffuse"][2], kmat["color_diffuse"][3]))
            file.write('\n')
            kmat["samplers"].sort(key=lambda sampler: sampler["index"])
            for samplerIndex, sampler in enumerate(kmat["samplers"]):
                if samplerIndex > 3: #can only have 4 textures per material
                    break
                file.write('\tsampler "%s" %i "%s" {\n' % (sampler["type"], sampler["index"], sampler["path"]))
                file.write('\t\tfilter\t\t%s\n' % (sampler["filter"]))
                file.write('\t\twrap\t\t%s\n' % (sampler["wrap"]))
                file.write('\t}\n')
            #end for
            file.write('}\n\n')
        #end for
    #end with
#end def

# -----------------------------------------------------------------------------
# Return all materials in the kmat file as a dictionary of "material Name", kmaterial.
def load(filepath):
    def on_create_material(objChain, materials, tokens):
        matName = tokens[1].value
        materials[matName] = {
            #"id": tokens[0].value,
            #"name": matName,
            "blend": False,
            "cull": "none", #none, back, front
            "fullbright": False,
            "sort": "default", #default, masked, translucent, water
            "surface_type": "default", #default, water, metal, stone, flesh_human, slime, forcefield 
            "depthtest": False,
            "reflectivePower": 0.0,
            "nodraw": False,
            "nodepthmask": False,
            "alphatest": False,
            "alphafunc": "gequal",  #gequal, equal, lequal
            "alphamask": 0.6525,
            "color_environment": (0.0, 0.0, 0.0, 1.0),
            "color_diffuse": (1.0, 1.0, 1.0, 1.0),
            "params": {},
            "samplers": [],
        }
        return (materials[matName], 'MATERIAL')
    #end def
    
    def on_add_param(objChain, mat, tokens):
        paramKey = utils.string_remove_quotes(tokens[1].value) #string value of shader uniform variable name
        mat["params"][paramKey] = {
            "type": tokens[2].value, #0=Integer, 1=Float
            "value": utils.string_remove_quotes(tokens[3].value), #value to set the variable (as a string)
        }
    #end def
    
    def on_create_sampler(objChain, mat, tokens):
        sampler = {
            "type": utils.string_remove_quotes(tokens[1].value), #always "diffuse" except if path is _depthbuffer then it's "depthBuffer"
            "index": tokens[2].value, #material texture slot index
            "path": utils.string_remove_quotes(tokens[3].value), #texture path relative to devDirectory/gameDirectory
            "filter": "linear_anisotropic", #'linear_anisotropic', 'linear', 'nearest'
            "wrap": "repeat", #'repeat', 'clamp', 'mirrored'
        }
        mat["samplers"].append(sampler)
        
        return (sampler, 'SAMPLER')
    #end def

    props = {
        'ROOT': [
            ('OBJ_DEF', on_create_material, ['CPP_ID', 'CPP_ID', 'CPP_LCURLY']), # material surf_312_00_00 {
        ],
        'MATERIAL': [
            ('PROP',        None,               [('CPP_ID', 'blend')]), # blend
            ('PROP',        None,               [('CPP_ID', 'cull'), 'CPP_ID']), # cull back
            ('PROP',        None,               [('CPP_ID', 'fullbright')]), # fullbright
            ('PROP',        None,               [('CPP_ID', 'sort'), 'CPP_ID']), # sort default
            ('PROP',        None,               [('CPP_ID', 'surface_type'), 'CPP_ID']), # surface_type default
            ('PROP',        None,               [('CPP_ID', 'depthtest')]), # depthtest
            ('PROP',        None,               [('CPP_ID', 'reflectivePower'), 'CPP_FLOAT']), # reflectivePower 0.700000
            ('PROP',        None,               [('CPP_ID', 'nodraw')]), # nodraw
            ('PROP',        None,               [('CPP_ID', 'nodepthmask')]), # nodepthmask
            ('PROP',        None,               [('CPP_ID', 'alphatest')]), # alphatest
            ('PROP',        None,               [('CPP_ID', 'alphafunc'), 'CPP_ID']), # alphafunc gequal
            ('PROP',        None,               [('CPP_ID', 'alphamask'), 'CPP_FLOAT']), # alphamask 0.6525
            ('PROP_SVEC4',  None,               [('CPP_ID', 'color_environment'), 'CPP_STRING']), # color_environment "0.000000 0.000000 0.000000 1.000000"
            ('PROP_SVEC4',  None,               [('CPP_ID', 'color_diffuse'), 'CPP_STRING']), # color_diffuse "1.000000 1.000000 1.000000 1.000000"
            ('PROP',        on_add_param,       [('CPP_ID', 'param'), 'CPP_STRING', 'CPP_INTEGER', 'CPP_STRING']), # param "uLimitAlpha" 0 "0"
            ('OBJ_DEF',     on_create_sampler,  [('CPP_ID', 'sampler'), 'CPP_STRING', 'CPP_INTEGER', 'CPP_STRING', 'CPP_LCURLY']), # sampler "diffuse" 0 "textures/texff140c00.png" {
            ('OBJ_END',     None,               ['CPP_RCURLY']), # }
        ],
        'SAMPLER': [
            ('PROP',    None, [('CPP_ID', 'filter'), 'CPP_ID']), # filter linear
            ('PROP',    None, [('CPP_ID', 'wrap'), 'CPP_ID']), # wrap clamp
            ('OBJ_END', None, ['CPP_RCURLY']), # }
        ],
    }
    
    return file_utils.parse_kex_file(filepath, props, {})
#end def
