import bpy
from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from . import (
    utils,
    kdefs,
    panel_object,
)

EnumSymbolTypes = [
    ("BODY",  "Body",  "", 0),
    ("HEAD",  "Head",  "", 1),
    ("ARM",  "Arm",  "", 2),
    ("LEG",  "Leg",  "", 3),
]

EnumSymbolSides = [
    ("LEFT",  "Left",  "", 0),
    ("RIGHT",  "Right",  "", 1),
]

EnumSymbolPartsBody = [
    ("ROOT",  "Root",  "", 0),
    ("UPPER TORSO",  "Upper Torso",  "", 1),
    ("LOWER TORSO",  "Lower Torso",  "", 2),
    ("PELVIS",  "Pelvis",  "", 3),
]

EnumSymbolPartsHead = [
    ("SKULL", "Skull",  "", 0),
    ("FACE",  "Face",  "", 1),
    ("CHIN",  "Chin",  "", 2),
    ("NECK",  "Neck",  "", 3),
]

EnumSymbolPartsArm = [
    ("SHOULDER", "Shoulder",  "", 0),
    ("UPPER ARM",  "Upper Arm",  "", 1),
    ("LOWER ARM",  "Lower Arm",  "", 2),
    ("HAND",  "Hand",  "", 3),
]

EnumSymbolPartsLeg = [
    ("HIP", "Hip",  "", 0),
    ("UPPER LEG",  "Upper Leg",  "", 1),
    ("LOWER LEG",  "Lower Leg",  "", 2),
    ("FOOT",  "Foot",  "", 3),
]

EnumBoneFlags = [
    ("0",  "0",  "", 1 << 0),
    ("1",  "1",  "", 1 << 1),
    ("2",  "2",  "", 1 << 2),
    ("3",  "3",  "", 1 << 3),
    ("4",  "4",  "", 1 << 4),
    ("5",  "5",  "", 1 << 5),
    ("6",  "6",  "", 1 << 6),
    ("7",  "7",  "", 1 << 7),
    ("8",  "8",  "", 1 << 8),
    ("9",  "9",  "", 1 << 9),
    ("10", "10", "", 1 << 10),
    ("11", "11", "", 1 << 11),
    ("12", "12", "", 1 << 12),
    ("13", "13", "", 1 << 13),
    ("14", "14", "", 1 << 14),
    ("15", "15", "", 1 << 15),
    ("16", "16", "", 1 << 16),
    ("17", "17", "", 1 << 17),
    ("18", "18", "", 1 << 18),
    ("19", "19", "", 1 << 19),
    ("20", "20", "", 1 << 20),
    ("21", "21", "", 1 << 21),
    ("22", "22", "", 1 << 22),
    ("23", "23", "", 1 << 23),
    ("24", "24", "", 1 << 24),
    ("25", "25", "", 1 << 25),
    ("26", "26", "", 1 << 26),
    ("27", "27", "", 1 << 27),
    ("28", "28", "", 1 << 28),
    ("29", "29", "", 1 << 29),
    ("30", "30", "", 1 << 30),
    ("31", "31", "", -2147483648)
]

# -----------------------------------------------------------------------------
#
class TUROK2_OT_setup_bone(bpy.types.Operator):
    bl_idname = "turok2_bone.setup_bone"
    bl_label = "Setup Bone"
    bl_description = "Sets the bone head/tail position and other settings"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
            
        bone = obj.data.bones.active
        return bone is not None
    #end def
        
    def execute(self, context):
        obj = context.object
        bone = obj.data.bones.active
        boneName = bone.name
        lastMode = context.object.mode
        #Must make obj active and in edit mode to change edit bones
        bpy.context.view_layer.objects.active = obj
        utils.mode_set('EDIT')
        bone = obj.data.edit_bones[boneName]
        oldHead = bone.head
        bone.head = (oldHead[0], oldHead[1], oldHead[2])
        bone.tail = (oldHead[0], oldHead[1] + utils.TUROK2_BONE_LENGTH, oldHead[2])
        bone.roll = 0.0
        utils.mode_set('OBJECT')
        poseBone = obj.pose.bones[boneName]
        poseBone.lock_scale = (True, True, True)
        utils.mode_set(lastMode)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_setup_all_bones(bpy.types.Operator):
    bl_idname = "turok2_bone.setup_all_bones"
    bl_label = "Setup All Bones"
    bl_description = "Sets the bone head/tail position and other settings for all bones in the armature"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
        
        return True
    #end def
        
    def execute(self, context):
        obj = context.object
        lastMode = context.object.mode
        bpy.context.view_layer.objects.active = obj
        utils.mode_set('EDIT')
        for bone in obj.data.edit_bones:
            oldHead = bone.head
            bone.head = (oldHead[0], oldHead[1], oldHead[2])
            bone.tail = (oldHead[0], oldHead[1] + utils.TUROK2_BONE_LENGTH, oldHead[2])
            bone.roll = 0.0
        #end for
        utils.mode_set('OBJECT')
        for poseBone in obj.pose.bones:
            poseBone.lock_scale = (True, True, True)
        #end for
        utils.mode_set(lastMode)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_autoname_bone(bpy.types.Operator):
    bl_idname = "turok2_bone.autoname_bone"
    bl_label = "Auto Name Bone"
    bl_description = "Sets the name of this bone and it's vertex group according to it's bone index value and symbols"
    bl_options = {'REGISTER', 'UNDO'}

    includeIndex: bpy.props.BoolProperty(name="Include Index", description="Includes the bone index value in the name", default=False)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
        bone = obj.data.bones.active
        return bone is not None
    #end def
        
    def execute(self, context):
        obj = context.object
        bone = obj.data.bones.active
        bonePrevName = bone.name
        boneNewName = utils.get_bonename_symbols(utils.get_bone_symbols(bone), bone.t2.boneIndex, self.includeIndex)
        if self.includeIndex and not utils.bone_index_is_valid(obj, bone):
            utils.show_error("Cannot set bone name because the Bone Index is not valid")
            return {'CANCELLED'}
        #end if
        bone.name = "_TEMP__" #so auto naming is always correct
        bone.name = boneNewName
        boneNewName = bone.name
        
        #find vertex group in armature mesh
        meshObj = utils.get_skinnedmesh_mesh(obj)
        if meshObj is None:
            for vg in meshObj.vertex_groups:
                if vg.name == bonePrevName:
                    vg.name = boneNewName
                #end if
            #end for
        #end if
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_autoname_all_bones(bpy.types.Operator):
    bl_idname = "turok2_bone.autoname_all_bones"
    bl_label = "Auto Name All Bones"
    bl_description = "Sets the name of all bones and their vertex groups according to their bone index and symbols"
    bl_options = {'REGISTER', 'UNDO'}

    includeIndex: bpy.props.BoolProperty(name="Include Index", description="Includes the bone index value in the name", default=False)
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
        
        return True
    #end def
        
    def execute(self, context):
        obj = context.object
        boneNames = {}
        invalidBoneIndexCount = 0
        for bone in obj.data.bones:
            bonePrevName = bone.name
            boneNewName = utils.get_bonename_symbols(utils.get_bone_symbols(bone), bone.t2.boneIndex, self.includeIndex)
            if self.includeIndex and not utils.bone_index_is_valid(obj, bone):
                invalidBoneIndexCount += 1
                continue
            #end if
            bone.name = "_TEMP__" #so auto naming is always correct
            boneNames[bonePrevName] = {"name":boneNewName, "bone":bone}
        #end for
        
        for bonePrevName, boneInfo in boneNames.items():
            boneInfo["bone"].name = boneInfo["name"]
        #end for
        
        #find vertex groups in armature mesh
        meshObj = utils.get_skinnedmesh_mesh(obj)
        if meshObj is None:
            for bonePrevName, boneInfo in boneNames.items():
                for vg in meshObj.vertex_groups:
                    if vg.name == bonePrevName:
                        vg.name = boneInfo["bone"].name
                    #end if
                #end for
            #end for
        #end if
        
        if invalidBoneIndexCount > 0:
            utils.show_error("Could not set bone names for %s out of %s bones because their Bone Indexes are not valid" % (invalidBoneIndexCount, len(obj.data.bones)))
        #end if
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_bone_pose_mode(bpy.types.Operator):
    bl_idname = "turok2_bone.enter_pose_mode"
    bl_label = "Enter Pose Mode"
    bl_description = "Enters Pose Mode"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.object is not None and context.object.mode != 'POSE'
    #end def
        
    def execute(self, context):
        utils.mode_set('POSE')
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_bone_impact_list(bpy.types.Operator):
    bl_idname = "turok2_bone.boneimpact_list"
    bl_label = "Impact Type from list"
    bl_description = "Select a Turok 2 Impact from a list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["impacts"]:
            name = kdefs.defs["impacts"][id]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
            
        bone = obj.data.bones.active
        return bone is not None
    #end def
        
    def execute(self, context):
        obj = context.object
        bone = obj.data.bones.active
        bone.t2.impactType = utils.get_enum_type(self.items_cb(context), self.enumprop)
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_bone_hotpoint_add(bpy.types.Operator):
    bl_idname = "turok2_bone.bone_hotpoint_add"
    bl_label = "Add Hot Point"
    bl_description = "Adds a new Hot Point to this bones armature"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return False
        #end if
            
        bone = obj.data.bones.active
        return bone is not None
    #end def
        
    def execute(self, context):
        obj = context.object
        bone = obj.data.bones.active
        utils.create_bone_hotpoint(obj, bone.t2.boneIndex)
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2BoneSettings(bpy.types.PropertyGroup):

    def impact_items_cb(self, context):
        return kdefs.defs["impacts_enums"]
    #end def

    def get_symbol_type(self):
        return self.get("symbolType", 0) #enums are their int values when in get/set funcs
    #end def

    def set_symbol_type(self, value):
        if self.get("symbolType", 0) != value:
            self["symbolType"] = value
            self["symbolPartsBody"] = "ROOT"
            self["symbolPartsHead"] = "SKULL"
            self["symbolPartsArm"] = "SHOULDER"
            self["symbolPartsLeg"] = "HIP"
        #end if
    #end def
    
    boneIndex: bpy.props.IntProperty(name="Bone Index", min=0, max=254, default=0)
    impactType: bpy.props.IntProperty(name="Impact Type", min=0, max=255, default=0)
    symbolType: bpy.props.EnumProperty(name="Symbol Type", items=EnumSymbolTypes, default="BODY", get=get_symbol_type, set=set_symbol_type)
    symbolSide: bpy.props.EnumProperty(name="Symbol Side", items=EnumSymbolSides, default="LEFT")
    symbolPartsBody: bpy.props.EnumProperty(name="Symbol Parts Body", items=EnumSymbolPartsBody, default="ROOT")
    symbolPartsHead: bpy.props.EnumProperty(name="Symbol Parts Head", items=EnumSymbolPartsHead, default="SKULL")
    symbolPartsArm: bpy.props.EnumProperty(name="Symbol Parts Arm", items=EnumSymbolPartsArm, default="SHOULDER")
    symbolPartsLeg: bpy.props.EnumProperty(name="Symbol Parts Leg", items=EnumSymbolPartsLeg, default="HIP")
    damageScalar: bpy.props.FloatProperty(name="Damage Scalar", default=1.0, precision=6, step=10, description="Damage multiplier for hitting this bone")
    flags: bpy.props.EnumProperty(name="Flags", items = EnumBoneFlags, options={'ENUM_FLAG'}, description="Specifies whether or not this bone will have collision depending on attached models, such as don't do collision if head has been blown off, etc. Each number represents the attached model index. If that number is ON then that bone will have collision with that attached model.")
    
    autoNameUseIndex: bpy.props.BoolProperty(name="Include Index In Name", description="Includes the bone index value in the name", default=True)

    def draw(self, context, panel, obj, bone):
        layout = panel.layout
        
        if obj.data.is_editmode:
            row = layout.row()
            row.label(text="Exit edit mode to edit Turok 2 Bone data")
            row = layout.row()
            row.operator(TUROK2_OT_bone_pose_mode.bl_idname)
            return
        #end if
        
        row = layout.row()
        row.prop(obj.t2, "boneTabIndex", expand=True)
        layout.box()
        if obj.t2.boneTabIndex == "DATA":
            isBoneInvalid = not utils.bone_index_is_valid(obj, bone)
            box = layout.box()
            row = box.row()
            row.alert = isBoneInvalid
            row.label(text="Bone Index", icon='BONE_DATA')
            row.prop(bone.t2, "boneIndex", text="")
                    
            row = box.row()
            row = row.split(factor=0.5)
            col = row.column()
            subrow = col.row()
            impactName = kdefs.defs["impacts"].get(bone.t2.impactType, '')
            impactLabelText = "Impact%s%s" % (" - " if impactName else "", impactName)
            subrow.label(text=impactLabelText)
            col = row.column()
            subrow = col.row(align=True)
            subrow.prop(bone.t2, "impactType", text="")
            subrow.operator(TUROK2_OT_bone_impact_list.bl_idname, icon='COLLAPSEMENU', text="")
            row = box.row()
            row.label(text="Damage Scalar")
            row.prop(bone.t2, "damageScalar", text="")

            box = layout.box()
            row = box.row()
            row.alignment = 'LEFT'
            row.label(text="Symbols", icon='OUTLINER_OB_ARMATURE')
            row = box.row()
            row.prop(bone.t2, "symbolType", text="Type")
            row = box.row()
            row.prop(bone.t2, "symbolSide", text="Side")
            row = box.row()
            if bone.t2.symbolType == "BODY":
                row.prop(bone.t2, "symbolPartsBody", text="Body Part")
            elif bone.t2.symbolType == "HEAD":
                row.prop(bone.t2, "symbolPartsHead", text="Head Part")
            elif bone.t2.symbolType == "ARM":
                row.prop(bone.t2, "symbolPartsArm", text="Arm Part")
            else: #"LEG"
                row.prop(bone.t2, "symbolPartsLeg", text="Leg Part")
            #end if

            box = layout.box()
            row = box.row()
            row.alignment = 'LEFT'
            row.label(text="Bone has collision with these model indexes", icon='SHADING_SOLID')
            for i in range(32):
                if i % 11 == 0:
                    row = box.row(align=True)
                #end if
                row.prop_enum(bone.t2, "flags", str(i))
            #end for
            
            #Add Hot Point
            box = layout.box()
            row = box.row()
            row.operator(TUROK2_OT_bone_hotpoint_add.bl_idname, text="Add Hot Point")
        elif obj.t2.boneTabIndex == "BB":
            panel_object.draw_bounding_box(layout, obj, bone.name)
        else: #Setup
            box = layout.box()
            row = box.row()
            row.operator(TUROK2_OT_setup_bone.bl_idname)
            row.operator(TUROK2_OT_setup_all_bones.bl_idname)
            row = box.row()
            row.operator(TUROK2_OT_autoname_bone.bl_idname).includeIndex = bone.t2.autoNameUseIndex
            row.operator(TUROK2_OT_autoname_all_bones.bl_idname).includeIndex = bone.t2.autoNameUseIndex
            row = box.row()
            row.alignment = "CENTER"
            row.prop(bone.t2, "autoNameUseIndex")
        #end if
        
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2BonePanel(bpy.types.Panel):
    bl_label = "Turok 2 Bone"
    bl_idname = "BONE_PT_Turok2BoneProp"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "bone"

    def draw(self, context):
        obj = context.object
        if obj is None or obj.type != "ARMATURE":
            return
        bone = obj.data.bones.active
        if bone is None:
            return
        
        bone.t2.draw(context, self, obj, bone)
    #end def
#end class

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(TUROK2_OT_setup_bone)
    bpy.utils.register_class(TUROK2_OT_setup_all_bones)
    bpy.utils.register_class(TUROK2_OT_autoname_bone)
    bpy.utils.register_class(TUROK2_OT_autoname_all_bones)
    bpy.utils.register_class(TUROK2_OT_bone_pose_mode)
    bpy.utils.register_class(TUROK2_OT_bone_impact_list)
    bpy.utils.register_class(TUROK2_OT_bone_hotpoint_add)
    bpy.utils.register_class(Turok2BoneSettings)
    bpy.utils.register_class(Turok2BonePanel)
    bpy.types.Bone.t2 = bpy.props.PointerProperty(type=Turok2BoneSettings)
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(TUROK2_OT_setup_bone)
    bpy.utils.unregister_class(TUROK2_OT_setup_all_bones)
    bpy.utils.unregister_class(TUROK2_OT_autoname_bone)
    bpy.utils.unregister_class(TUROK2_OT_autoname_all_bones)
    bpy.utils.unregister_class(TUROK2_OT_bone_pose_mode)
    bpy.utils.unregister_class(TUROK2_OT_bone_impact_list)
    bpy.utils.unregister_class(TUROK2_OT_bone_hotpoint_add)
    bpy.utils.unregister_class(Turok2BoneSettings)
    bpy.utils.unregister_class(Turok2BonePanel)
    del bpy.types.Bone.t2
#end def
