import bpy
import os
from . import (
    utils,
    kmaterial,
    panel_shading,
)
from bpy_extras.io_utils import ImportHelper

EnumMaterialSurfaces = [
    ("DEFAULT",     "0: default",       "", 0),
    ("WATER",       "1: water",         "", 1),
    ("METAL",       "2: metal",         "", 2),
    ("STONE",       "3: stone",         "", 3),
    ("FLESH_HUMAN", "4: flesh_human",   "", 4),
    ("SLIME",       "8: slime",         "", 5),
    ("FORCEFIELD",  "9: forcefield",    "", 6),
]

EnumMaterialAlphaFuncs = [
    ("GEQUAL", "gequal", "", 0),
    ("EQUAL", "equal", "", 1),
    ("LEQUAL", "lequal", "", 2),
]

EnumMaterialSorts = [
    ("DEFAULT", "default", "", 0),
    ("MASKED", "masked", "", 1),
    ("TRANSLUCENT", "translucent", "", 2),
    ("WATER", "water", "", 3),
]

EnumMaterialCulls = [
    ("AUTO", "Use Settings", 'Uses the material setting "Backface Culling"', 0),
    ("BACK", "Back", "", 1),
    ("FRONT", "Front", "", 2),
    ("NONE", "None", "", 3),
]

EnumMaterialQuickSetups = [
    ("ONESIDE", "One Sided", "", 0),
    ("TWOSIDE", "Two Sided", "", 1),
    ("WATER", "Water", "", 2),
]

EnumMaterialCommonParams = [
    ("CUSTOM", "Custom", "", 0),
    ("uReflective", "Reflective", "", 1),
    ("uLimitAlpha", "Limit Alpha", "", 2),
    ("uPseduoColor", "Pseduo Color", "", 3),
]

# -----------------------------------------------------------------------------
#
class Turok2MaterialParamsPropGroup(bpy.types.PropertyGroup):
    def get_value_type(self):
        return self.get("valueType", 1)
    #end def
        
    def set_value_type(self, value):
        self["valueType"] = value
        self["valueTypeBool"] = self.valueType != 0
    #end def
        
    def get_value_type_bool(self):
        return self.get("valueTypeBool", True)
    #end def
        
    def set_value_type_bool(self, value):
        self["valueTypeBool"] = value
        self["valueType"] = 1 if self.valueTypeBool else 0
    #end def
    
    def get_onname(self):
        return self.get("name", "")
    #end def
        
    def set_onname(self, value):
        self["name"] = value
        utils.update_shader_groups(self["selfMat"])
    #end def
    
    def get_onvalue(self):
        return self.get("value", "")
    #end def
        
    def set_onvalue(self, value):
        self["value"] = value
        utils.update_shader_groups(self["selfMat"])
    #end def
        
    name: bpy.props.StringProperty(name="Name", default='', get=get_onname, set=set_onname)
    valueType: bpy.props.IntProperty(name="realType", description="Param type to Int or Float", default = 0, get=get_value_type, set=set_value_type)
    value: bpy.props.StringProperty(name = "Value", description="Param value", default='', get=get_onvalue, set=set_onvalue)
    #used for ui only
    valueTypeBool: bpy.props.BoolProperty(name="Type", description="Param type to Int or Float", default = False, get=get_value_type_bool, set=set_value_type_bool)
    selfMat: bpy.props.PointerProperty(name="Self Mat", type = bpy.types.Material)
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_UL_material_param_props(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        if not item.name:
            layout.alert = True
        #end if
        row = layout.row()
        row.prop(item, "name")
        row = layout.row()
        row.prop(item, "valueTypeBool", icon_only=True, icon="FILE_FONT" if item.valueTypeBool else "ITALIC")
        row = layout.row()
        row.prop(item, "value", text="")
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_add_material_param(bpy.types.Operator):
    bl_idname = "turok2_material.param_add"
    bl_label = "Add Turok 2 Material Param"
    bl_description = "Adds a new Param"
    bl_options = {'INTERNAL', 'UNDO'}

    #paramName: bpy.props.StringProperty()
    paramName: bpy.props.EnumProperty(items=EnumMaterialCommonParams)

    @classmethod
    def poll(cls, context):
        obj = context.object
        return obj is not None and obj.active_material is not None
    #end def

    def execute(self, context):
        obj = context.object
        mat = obj.active_material
        mat = utils.get_material_texset_index(mat, 0)
        if mat is not None:
            param = None
            if self.paramName != "CUSTOM":
                for p in mat.t2.params:
                    if p.name == self.paramName:
                        param = p
                        break
                    #end if
                #end for
            #end if
            if param is None:
                param = mat.t2.params.add()
            #end if
            param.selfMat = mat
            if self.paramName != "CUSTOM":
                param.name = self.paramName
                param.value = "1"
                param.valueTypeBool = False
            else:
                param.name = ""
                param.value = "0"
                param.valueTypeBool = False
            #end if
        #end if

        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
# 
class TUROK2_OT_remove_material_param(bpy.types.Operator):
    bl_idname = "turok2_material.param_remove"
    bl_label = "Remove Turok 2 Material Param"
    bl_description = "Removes selected Param"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.active_material is None:
            return False
        mat = utils.get_material_texset_index(obj.active_material , 0)    
        return len(mat.t2.params) > 0
    #end def
    
    def execute(self, context):
        obj = context.object
        mat = obj.active_material
        mat = utils.get_material_texset_index(mat, 0)
        paramsLength = len(mat.t2.params)
        if mat.t2.paramsIndex >= 0 and mat.t2.paramsIndex < paramsLength:
            mat.t2.params.remove(mat.t2.paramsIndex)
            mat.t2.paramsIndex -= 1
        else:
            mat.t2.params.remove(paramsLength - 1)
        #end if
        utils.update_shader_groups(mat)

        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
# 
class TUROK2_OT_remove_material_param_all(bpy.types.Operator):
    bl_idname = "turok2_material.param_remove_all"
    bl_label = "Remove All Turok 2 Material Params"
    bl_description = "Removes All Params"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.active_material is None:
            return False
        mat = utils.get_material_texset_index(obj.active_material , 0)    
        return len(mat.t2.params) > 0
    #end def
    
    def execute(self, context):
        obj = context.object
        mat = obj.active_material
        mat = utils.get_material_texset_index(mat, 0)
        mat.t2.params.clear()
        mat.t2.paramsIndex -= 1
        utils.update_shader_groups(mat)

        return {'FINISHED'}
    #end def
#end class
 
# -----------------------------------------------------------------------------
#
class TUROK2_OT_material_quicktex_import(bpy.types.Operator):
    """Setup the material quickly with a texture"""
    bl_idname = "turok2_material.mat_quicktex_import"
    bl_label = "Quick Material Texture Setup"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.active_material is None:
            return False
        #end if
        mat = utils.get_material_texset_index(obj.active_material , 0)    
        return mat == obj.active_material and mat.t2.quickImage is not None
    #end def
        
    def execute(self, context):
        mat = context.object.active_material
        mat = utils.get_material_texset_index(mat, 0)

        imageFilePath = os.path.normpath(mat.t2.quickImage.filepath).strip('\\')
        imageFileDir, imageFileName = os.path.split(imageFilePath)
        imageFileNameNoExt = os.path.splitext(imageFileName)[0]
        realMatName, textureSetIndex = utils.get_real_material_name(mat.name)
        if textureSetIndex < 0:
            mat.name = utils.get_material_texset_name(imageFileNameNoExt, 0)
        else:
            mat.name = utils.get_material_texset_name(realMatName, textureSetIndex)
        #end if
        
        #setup nodes with image
        mat.node_tree.nodes.clear()
        moNode = mat.node_tree.nodes.new("ShaderNodeOutputMaterial")
        moNode.location = [300, 300]
        bsdf = mat.node_tree.nodes.new("ShaderNodeBsdfPrincipled")
        bsdf.location = [10, 300]
        bsdf.inputs['Specular'].default_value = 0.0
        mat.node_tree.links.new(moNode.inputs['Surface'], bsdf.outputs['BSDF'])
        texShaderNode = mat.node_tree.nodes.new('ShaderNodeTexImage')
        texShaderNode.label = imageFileName
        texShaderNode.location = [-280.0, 300.0]
        texShaderNode.projection = 'FLAT'
        texShaderNode.interpolation = 'Linear'
        texShaderNode.extension = 'REPEAT'
        texShaderNode.image = mat.t2.quickImage;
        
        mat.preview_render_type = 'FLAT'
        mat.use_nodes = True
        mat.specular_intensity = 0.0
        mat.alpha_threshold = 0.3475
        
        mat.node_tree.links.new(bsdf.inputs['Base Color'], texShaderNode.outputs['Color'])
        mat.node_tree.links.new(bsdf.inputs['Alpha'], texShaderNode.outputs['Alpha'])
        for shaderNode in mat.node_tree.nodes:
            shaderNode.select = False
        #end for
            
        #Setup Material Props
        mat.t2.cull = "AUTO"
        if mat.t2.quickImageSetup == "ONESIDE" or mat.t2.quickImageSetup == "TWOSIDE":
            if mat.t2.quickImageSetup == "ONESIDE":
                mat.use_backface_culling = True
            else:
                mat.use_backface_culling = False
            #end if
            if mat.t2.quickImageAlpha:
                mat.t2.blend = True
                mat.t2.alphaTest = True
                mat.t2.sort = "MASKED"
                mat.blend_method = "CLIP"
            else:
                mat.t2.blend = False
                mat.t2.alphaTest = False
                mat.t2.sort = "DEFAULT"
                mat.blend_method = "OPAQUE"
            #end if
            mat.t2.depthTest = True
        else: #"WATER"
            mat.use_backface_culling = False
            mat.blend_method = "BLEND"
            mat.t2.blend = True
            mat.t2.sort = "WATER"
            mat.t2.depthTest = True
            mat.t2.alphaTest = False
        #end if
        
        #Reset convenience function props
        mat.t2.quickImage = None
        mat.t2.quickImageSetup = "ONESIDE"
        mat.t2.quickImageAlpha = False
        return {'FINISHED'}
    #end def

#end class

# -----------------------------------------------------------------------------
#
kmat_import_enum_items = []
kmat_import_materials = {}
kmat_import_isShowingList = False
# -----------------------------------------------------------------------------
#
class TUROK2_OT_material_import(bpy.types.Operator, ImportHelper):
    """Import existing material over current material"""
    bl_idname = "turok2_material.mat_import"
    bl_label = "Import Material"
    bl_options = {'PRESET', 'INTERNAL', 'UNDO'}

	# Properties used by the file browser
    filepath: bpy.props.StringProperty(name="File Path", description="File path used for importing the material from the KMAT file", default="", options={'HIDDEN'})
    files: bpy.props.CollectionProperty(type=bpy.types.OperatorFileListElement, options={'HIDDEN'})
    directory: bpy.props.StringProperty(subtype='FILE_PATH', options={'HIDDEN'})
    filter_folder: bpy.props.BoolProperty(name="Filter Folders", description="", default=True, options={'HIDDEN'})
    filename_ext = ".kmat"
    filter_glob: bpy.props.StringProperty(default="*.kmat", options={'HIDDEN'})
    
    bl_property = "enumprop"
    
    def items_cb(self, context):
        global kmat_import_enum_items
        
        return kmat_import_enum_items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb, options={'HIDDEN'})

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None or obj.active_material is None:
            return False
        #end if
        mat = utils.get_material_texset_index(obj.active_material , 0)    
        return mat == obj.active_material
    #end def

    def execute(self, context):
        global kmat_import_materials
        global kmat_import_enum_items
        global kmat_import_isShowingList
        
        obj = context.object
        mat = obj.active_material
        mat = utils.get_material_texset_index(mat, 0)
        if kmat_import_isShowingList:
            kmat_import_isShowingList = False
            kmaterial.to_material(kmat_import_materials[self.enumprop], mat)
            kmat_import_materials.clear()
            kmat_import_enum_items.clear()
        else:
            #read filepath. if exists then get all the materials and add to a list of materials (global var)
            kmat_import_materials = kmaterial.load(self.filepath)
            if kmat_import_materials and len(kmat_import_materials) > 0:
                kmat_import_enum_items.clear()
                for i, key in enumerate(kmat_import_materials):
                    kmat_import_enum_items.append((key, key, "", i))
                #end for
                kmat_import_isShowingList = True
                context.window_manager.invoke_search_popup(self)
            else:
                utils.show_error("[Import Material] %s contains no valid materials." % (self.filepath))
            #end if
        #end if
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_mat_node_set_mirrored_tex(bpy.types.Operator):
    bl_idname = "turok2_shading.mat_node_set_mirrored"
    bl_label = "Mirror Tex"
    bl_description = 'Overrides texture wrapping to "mirrored" which flips the texture on the x and y when it repeats'
    bl_options = {'INTERNAL', 'UNDO'}

    setValue: bpy.props.BoolProperty(name="New Value", default=False)
    
    @classmethod
    def poll(cls, context):
        mat = utils.get_material_texset_index(context.active_object.active_material, 0)
        texNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']
        return len(texNodes) > 0
    #end def

    def execute(self, context):
        obj = context.object
        mat = utils.get_material_texset_index(context.active_object.active_material, 0)
        texNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']
        if len(texNodes) > 0:
            texNodes.sort(key=lambda node: node.t2.samplerIndex)
            panel_shading.set_mirrored(context, obj, mat, texNodes[0], self.setValue)
        #end if
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_mat_node_set_anisotropic_tex(bpy.types.Operator):
    bl_idname = "turok2_shading.mat_node_set_anisotropic"
    bl_label = "Anisotropic Filter"
    bl_description = 'Overrides texture filtering to "linear_anisotropic" when Linear, Cubic, or Smart is selected'
    bl_options = {'INTERNAL', 'UNDO'}

    setValue: bpy.props.BoolProperty(default=False)
    
    @classmethod
    def poll(cls, context):
        if context.object is None or context.active_object is None or context.active_object.active_material is None:
            return False
        #end if
        mat = utils.get_material_texset_index(context.active_object.active_material, 0)
        texNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']
        return len(texNodes) > 0
    #end def

    def execute(self, context):
        obj = context.object
        mat = utils.get_material_texset_index(context.active_object.active_material, 0)
        texNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']
        texNodes.sort(key=lambda node: node.t2.samplerIndex)
        texNodes[0].t2.anisotropicFilter = self.setValue
        mat.use_backface_culling = mat.use_backface_culling #Forces rendering to refresh
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2MaterialSettings(bpy.types.PropertyGroup):

    def matcull_items_cb(self, context):
        obj = context.active_object
        if obj is None or obj.active_material is None:
            return EnumMaterialCulls
        #end if
        mat = utils.get_material_texset_index(obj.active_material, 0)
        items = []
        for item in EnumMaterialCulls:
            if item[0] == "AUTO":
                itemName = "%s (%s)" % (item[1], "Back" if mat.use_backface_culling else "None")
            else:
                itemName = item[1]
            #end if
            items.append((item[0], itemName, item[2], item[3]))
        #end for
        return items
    #end def
    
    def on_update_cull(self, context):
        obj = context.active_object
        if obj is None or obj.active_material is None:
            return
        #end if
        if self.cull in ["BACK", "NONE"]:
            mat = utils.get_material_texset_index(obj.active_material, 0)
            mat.use_backface_culling = self.cull == "BACK"
        #end if
    #end def

    def on_update_blend_method(self, context):
        obj = context.active_object
        if obj is None or obj.active_material is None:
            return
        #end if
        mat = utils.get_material_texset_index(obj.active_material, 0)
        if self.alphaTest:
            if mat.blend_method != "CLIP":
                mat.blend_method = "CLIP"
                mat.alpha_threshold = 0.3475
            #end if
        elif self.blend:
            if mat.blend_method != "BLEND":
                mat.blend_method = "BLEND"
            #end if
        else:
            if mat.blend_method != "OPAQUE":
                mat.blend_method = "OPAQUE"
            #end if
        #end if
    #end def
    
    def on_update_fullbright(self, context):
        obj = context.active_object
        if obj is None or obj.active_material is None:
            return
        #end if
        mat = utils.get_material_texset_index(obj.active_material, 0)
        texImageNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage' and node.t2.samplerIndex == 0]
        bsdfNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeBsdfPrincipled']
        if len(texImageNodes) > 0 and len(bsdfNodes) > 0:
            texImageNode = texImageNodes[0]
            bsdf = bsdfNodes[0]
            if mat.t2.fullBright:
                mat.node_tree.links.new(bsdf.inputs['Emission'], texImageNode.outputs['Color'])
            else:
                for link in bsdf.inputs['Emission'].links:
                    mat.node_tree.links.remove(link)
                #end for
                #Forces rendering to refresh (There appears to be a bug when removing links. It doesn't update the shading rendering)
                mat.use_backface_culling = mat.use_backface_culling
            #end if
        #end if
    #end def
    
    isMatPathCustom: bpy.props.BoolProperty(name="Use Custom Material Path", description="If is turned off will automatically set the material file path to materials/mat_<ExportFileName>", default=False)
    matPath: bpy.props.StringProperty(name="Material Path", description="The path to the .kmat file where this material should be placed inside. Do not include the .kmat extension.", default="materials/mat_customName", subtype = "FILE_NAME")
    fullBright: bpy.props.BoolProperty(name="Full Bright", description="Do not use any lighting", default=False, update=on_update_fullbright)
    blend: bpy.props.BoolProperty(name="Blend", description="Turn on for material transparency", default=False, update=on_update_blend_method)
    alphaTest: bpy.props.BoolProperty(name="Alpha Test", default=False, update=on_update_blend_method, description="Turn ON if you have a texture that uses alpha and use sort type Masked. Discards pixels depending on the outcome of the Alpha Func comparison between the textures alpha value and the Alpha Mask value.")
    alphaFunc: bpy.props.EnumProperty(name="Alpha Func", items = EnumMaterialAlphaFuncs, default = "GEQUAL", description="The alpha comparison function (default: gequal)")
    alphaMask: bpy.props.FloatProperty(name="Alpha Mask", default = 0.6525, subtype="FACTOR", soft_min=0.0, soft_max=1.0, min=0.0, max=1.0, precision=6, description="The alpha cut off value to discard the pixel. (default: 0.6525)")
    sort: bpy.props.EnumProperty(name="Sort", items = EnumMaterialSorts, default = "DEFAULT", description="The draw layer to use")
    cull: bpy.props.EnumProperty(name="Cull", items=matcull_items_cb, description="Which side to render Triangles", update=on_update_cull)
    noDraw: bpy.props.BoolProperty(name="No Draw", description="Do not render", default=False)
    depthTest: bpy.props.BoolProperty(name="Depth Test", default=True, description="Draws behind objects based on Depth buffer, otherwise will always draw on top of current render.")
    noDepthMask: bpy.props.BoolProperty(name="No Depth Mask", default=False, description="Turns off depth mask. The depth mask masks out areas on the model depending on how far from the camera it is.")
    colorEnvironment: bpy.props.FloatVectorProperty(name = "Environment Color", subtype='COLOR', size = 4, default=[0.0,0.0,0.0,1.0], soft_min=0.0, soft_max=1.0, min=0.0, max=1.0, precision=6)
    colorDiffuse: bpy.props.FloatVectorProperty(name = "Diffuse Color", subtype='COLOR', size = 4, default=[1.0,1.0,1.0,1.0], soft_min=0.0, soft_max=1.0, min=0.0, max=1.0, precision=6)
    reflectivePower: bpy.props.FloatProperty(name="Reflective Power", default = 0.0, subtype="FACTOR", soft_min=0.0, soft_max=1.0, min=0.0, max=1.0, precision=6, description="Screen Space reflection power")
    surfaceType: bpy.props.EnumProperty(name="Surface", items = EnumMaterialSurfaces, default = "DEFAULT", description="Type of surface impact")
    params: bpy.props.CollectionProperty(name="Params", type = Turok2MaterialParamsPropGroup, description="A list of uniform variables the shader will use. Assign the correct names from the shader and specify whether the variable is an Integer or Float. Then assign it's value.")
    
    #blender editor props
    paramsIndex: bpy.props.IntProperty(name="", default = -1)
    quickImage: bpy.props.PointerProperty(name="Quick Setup Image", type = bpy.types.Image)
    quickImageSetup: bpy.props.EnumProperty(name="Setup Type", items = EnumMaterialQuickSetups, default = "ONESIDE")
    quickImageAlpha: bpy.props.BoolProperty(name="Image Has Alpha", default=False)
    
    def draw(self, context, panel):
        layout = panel.layout
        obj = context.active_object
        mat = utils.get_material_texset_index(obj.active_material, 0)
        texNodes = [node for node in mat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']

        # #check if theres already a none texture set suffix material as the first index then do nothing
        # if (textureSetIndex == 0 and baseMat is not None) or (textureSetIndex > 0):
            # row = layout.row()
            # row.label(text="This material is a variant of")
            # box = layout.box()
            # row = box.row()
            # row.label(text=matName)
            # row = layout.row()
            # row.label(text="Select that material instead to edit it's properties.")
            # return
        # #end if
        row = layout.row()
        row.prop(obj.t2, "materialTabIndex", expand=True)
        if obj.t2.materialTabIndex == "SETTINGS":
            if mat == context.active_object.active_material:
                showConUI = len(texNodes) == 0
                if showConUI:
                    box = layout.box()
                    row = box.row()
                    row.label(text="Setup Image Quickly", icon='NODE_TEXTURE')
                    row = box.row()
                    row.template_ID_preview(mat.t2, "quickImage", open="image.open", rows=3, cols=6, hide_buttons=True)
                    # row.template_ID(mat.t2, "quickImage", new="image.new", open="image.open", live_icon=True)
                    row = box.row()
                    row.prop(mat.t2, "quickImageSetup")
                    row = box.row()
                    row.alignment = "RIGHT"
                    row.prop(mat.t2, "quickImageAlpha")
                    row = box.row()
                    row.alignment = "CENTER"
                    row.operator(TUROK2_OT_material_quicktex_import.bl_idname, text="Quick Image Setup")
                #end if
            else:
                row = layout.row()
                row.alignment = "CENTER"
                row.alert = True
                row.label(text="Viewing: %s" % (mat.name))
            #end if
            
            if len(texNodes) > 0:
                texNodes.sort(key=lambda node: node.t2.samplerIndex)
                texNode = texNodes[0]
                box = layout.box()
                row = box.row()
                row = row.split(factor=0.55)
                col = row.column()
                col.scale_y = 0.7
                col.template_ID_preview(texNode, "image", open="image.open", rows=3, cols=6, hide_buttons=True)
                col = row.column()
                col.prop(texNode, "extension", text="Wrap")
                col.prop(texNode, "interpolation", text="Filter")
                col.separator()
                panel_shading.draw_shading_panel(col, context, texNode, TUROK2_OT_mat_node_set_mirrored_tex, TUROK2_OT_mat_node_set_anisotropic_tex, True)
            #end if
            
            row = layout.row()
            row.prop(mat.t2, "colorEnvironment")
            row = layout.row()
            row.prop(mat.t2, "colorDiffuse")
            row = layout.row()
            row.prop(mat.t2, "cull")
            row = layout.row()
            row.prop(mat.t2, "sort")
            row = layout.row()
            row.prop(mat.t2, "surfaceType")
            row = layout.row()
            row.label(text="Reflective Power")
            row.prop(mat.t2, "reflectivePower", text="")
            row = layout.row()
            row.prop(mat.t2, "blend")
            row.prop(mat.t2, "fullBright")
            row = layout.row()
            row.prop(mat.t2, "depthTest")
            row.prop(mat.t2, "noDraw")
            row = layout.row()
            row.prop(mat.t2, "noDepthMask")
            row = layout.row()
            row.prop(mat.t2, "alphaTest")
            col = row.column()
            col.enabled = mat.t2.alphaTest
            col.prop(mat.t2, "alphaFunc", text="Func")
            row = layout.row()
            row.enabled = mat.t2.alphaTest
            row.label(text="Alpha Mask")
            row.prop(mat.t2, "alphaMask", text="")
            box = layout.box()
            row = box.row()
            paramsText = 'Params'
            if len(mat.t2.params) > 0:
                paramsText = 'Params (%d)' % (len(mat.t2.params))
            #end if
            row.label(text=paramsText)
            row = box.row()
            row.template_list('TUROK2_UL_material_param_props', "Turok2MaterialParamsPropGroup", mat.t2, "params", mat.t2, "paramsIndex", rows=1, maxrows=3, type='DEFAULT')
            row = box.row()
            col = row.column()
            col.operator_menu_enum(TUROK2_OT_add_material_param.bl_idname, "paramName", text="Add Param", icon='ADD')
            col = row.column()
            row = col.row(align=True)
            row.operator(TUROK2_OT_remove_material_param.bl_idname, text="Remove Param", icon='REMOVE')
            row.operator(TUROK2_OT_remove_material_param_all.bl_idname, text="", icon='X')
        else:
            if mat == context.active_object.active_material:
                box = layout.box()
                row = box.row()
                row.label(text="Setup Image Quickly", icon='NODE_TEXTURE')
                row = box.row()
                row.template_ID_preview(mat.t2, "quickImage", open="image.open", rows=3, cols=6, hide_buttons=True)
                # row.template_ID(mat.t2, "quickImage", new="image.new", open="image.open", live_icon=True)
                row = box.row()
                row.prop(mat.t2, "quickImageSetup")
                row = box.row()
                row.alignment = "RIGHT"
                row.prop(mat.t2, "quickImageAlpha")
                row = box.row()
                row.alignment = "CENTER"
                row.operator(TUROK2_OT_material_quicktex_import.bl_idname, text="Quick Image Setup")
                box = layout.box()
                row = box.row()
                row.operator(TUROK2_OT_material_import.bl_idname, icon="IMPORT", text="Import Existing Material from .kmat")
            else:
                row = layout.row()
                row.alignment = "CENTER"
                row.alert = True
                row.label(text="Viewing: %s" % (mat.name))
            #end if
            box = layout.box()
            row = box.row()
            row.prop(mat.t2, "isMatPathCustom")
            if mat.t2.isMatPathCustom:
                row = box.row()
                row.prop(mat.t2, "matPath", text="")
            #end if
        #end if
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2MaterialPanel(bpy.types.Panel):
    bl_label = "Turok 2 Material"
    bl_idname = "MATERIAL_PT_Turok2MatProp"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "material"

    def draw(self, context):
        if context.active_object is None or context.active_object.active_material is None:
            return
        #end if

        mat = context.active_object.active_material
        mat = utils.get_material_texset_index(mat, 0)
        mat.t2.draw(context, self)
    #end def
#end class

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2MaterialParamsPropGroup)
    bpy.utils.register_class(TUROK2_UL_material_param_props)
    bpy.utils.register_class(TUROK2_OT_add_material_param)
    bpy.utils.register_class(TUROK2_OT_remove_material_param)
    bpy.utils.register_class(TUROK2_OT_remove_material_param_all)
    bpy.utils.register_class(TUROK2_OT_material_quicktex_import)
    bpy.utils.register_class(TUROK2_OT_material_import)
    bpy.utils.register_class(TUROK2_OT_mat_node_set_mirrored_tex)
    bpy.utils.register_class(TUROK2_OT_mat_node_set_anisotropic_tex)
    bpy.utils.register_class(Turok2MaterialSettings)
    bpy.utils.register_class(Turok2MaterialPanel)
    bpy.types.Material.t2 = bpy.props.PointerProperty(type=Turok2MaterialSettings)
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(Turok2MaterialParamsPropGroup)
    bpy.utils.unregister_class(TUROK2_UL_material_param_props)
    bpy.utils.unregister_class(TUROK2_OT_add_material_param)
    bpy.utils.unregister_class(TUROK2_OT_remove_material_param)
    bpy.utils.unregister_class(TUROK2_OT_remove_material_param_all)
    bpy.utils.unregister_class(TUROK2_OT_material_quicktex_import)
    bpy.utils.unregister_class(TUROK2_OT_material_import)
    bpy.utils.unregister_class(TUROK2_OT_mat_node_set_mirrored_tex)
    bpy.utils.unregister_class(TUROK2_OT_mat_node_set_anisotropic_tex)
    bpy.utils.unregister_class(Turok2MaterialSettings)
    bpy.utils.unregister_class(Turok2MaterialPanel)
    del bpy.types.Material.t2
#end def
