import bpy
import math
import os

from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
import bpy_extras.view3d_utils
from bpy_extras.image_utils import load_image
from bpy_extras.wm_utils.progress_report import (
    ProgressReport,
    ProgressReportSubstep,
)
from . import (
    utils,
    file_utils,
    paths,
    staticmesh,
    skinnedmesh,
    kdefs,
    aud_utils,
    panel_action,
    panel_bone,
)

T2ObjectTypes = [
    ("NONE", "None", "", 0),
    ("STATICMESH", "Staticmesh", "", 1),
    ("SKINNEDMESH", "Skinnedmesh", "", 2),
    ("BOUNDING BOX", "Bounding Box", "", 3),
    ("HOT POINT", "Hot Point", "", 4),
]

EnumHotpointFlags = [
    ("0",  "0",  "", 1 << 0),
    ("1",  "1",  "", 1 << 1),
    ("2",  "2",  "", 1 << 2),
    ("3",  "3",  "", 1 << 3),
    ("4",  "4",  "", 1 << 4),
    ("5",  "5",  "", 1 << 5),
    ("6",  "6",  "", 1 << 6),
    ("7",  "7",  "", 1 << 7),
    ("8",  "8",  "", 1 << 8),
    ("9",  "9",  "", 1 << 9),
    ("10", "10", "", 1 << 10),
    ("11", "11", "", 1 << 11),
    ("12", "12", "", 1 << 12),
    ("13", "13", "", 1 << 13),
    ("14", "14", "", 1 << 14),
    ("15", "15", "", 1 << 15),
    ("16", "16", "", 1 << 16),
    ("17", "17", "", 1 << 17),
    ("18", "18", "", 1 << 18),
    ("19", "19", "", 1 << 19),
    ("20", "20", "", 1 << 20),
    ("21", "21", "", 1 << 21),
    ("22", "22", "", 1 << 22),
    ("23", "23", "", 1 << 23),
    ("24", "24", "", 1 << 24),
    ("25", "25", "", 1 << 25),
    ("26", "26", "", 1 << 26),
    ("27", "27", "", 1 << 27),
    ("28", "28", "", 1 << 28),
    ("29", "29", "", 1 << 29),
    ("30", "30", "", 1 << 30),
    ("31", "31", "", -2147483648)
]

EnumMaterialTabs = [
    ("SETTINGS", "Settings", "", 0),
    ("MISC", "Misc", "", 1),
]

EnumSkinTabs = [
    ("ANIM", "Animation", "", 0),
    ("BB", "BB", "", 1),
    ("HEAD", "Head Track", "", 2),
    ("MISC", "Misc", "", 3),
]

EnumBoneTabs = [
    ("DATA", "Data", "", 0),
    ("BB", "BB", "", 1),
    ("SETUP", "Setup", "", 2),
]

EnumActionTabs = [
    ("DATA", "Data", "", 0),
    ("KFA", "Key Frame Actions", "", 1),
]

EnumBBVisibilities = [
    ("0", "Show Only This Bounding Box", "", 0),
    ("1", "Show Collision Bounding Boxes", "", 1),
    ("2", "Show All Bounding Boxes", "", 2),
    ("3", "Hide All Bounding Boxes", "", 3),
    ("4", "Hide This Bounding Box", "", 4),
]

EnumHPVisibilities = [
    ("0", "Show Hot Points with same Bit", "", 0),
    ("1", "Show All Hot Points", "", 1),
    ("2", "Hide Hot Points with same Bit", "", 2),
    ("3", "Hide All Hot Points", "", 3),
]

# -----------------------------------------------------------------------------
#
class Turok2ObjectProps(bpy.types.PropertyGroup):

    def on_update_textureset_index(self, context):
        if context is None or not hasattr(context, "object"):
            return
        #end if
        
        obj = utils.get_root_turok_object2(context.object, "SKINNEDMESH", "ARMATURE")
        if obj.type == "ARMATURE" and obj.t2.objType == "SKINNEDMESH":
            utils.set_obj_material_tex_index(obj, self.textureSetIndex)
        else:
            utils.set_obj_material_tex_index(context.object, self.textureSetIndex)
        #end if
    #end def
    
    def on_update_kfaPointIsRoot(self, context):
        if context is None or not hasattr(context, "object"):
            return
        #end if
        turok2PointObj = utils.get_object(utils.TUROK2_ACTIONPOINT_NAME)
        if turok2PointObj is None:
            return
        #end if
        obj = context.object
        if obj is None:
            return
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.animation_data is None or armObj.animation_data.action is None:
            return
        #end if
        tAction = armObj.animation_data.action.t2
        if tAction.keyFramesIndex < 0 or tAction.keyFramesIndex >= len(tAction.keyFrames):
            return
        #end if
        kfa = tAction.keyFrames[tAction.keyFramesIndex]
        turok2PointObj = utils.create_kfa_point(context, kfa, armObj)
        turok2PointObj.select_set(True)
    #end def

    def hpflags_items_cb(self, context):
        items = []
        for item in EnumHotpointFlags:
            hpName = kdefs.defs["hotpoints"].get(int(item[0]), '')
            if hpName:
                flagName = "%s: %s" % (item[1], hpName)
            else:
                flagName = item[1]
            #end if
            items.append((item[0], flagName, item[2], item[3]))
        #end for
        return items
    #end def
    
    #Object Properties
    objType: bpy.props.EnumProperty(name="Turok 2 Object Types", description="The type of Turok 2 Object this Blender Object is.", items=T2ObjectTypes, default="NONE")
    
    #Model Properties
    textureSetIndex: bpy.props.IntProperty(name="Preview TextureSet", min=0, max=2147483647, default=0, update=on_update_textureset_index)
    headTrackStart: bpy.props.IntProperty(name="Head Track Start Bone", min=0, max=255, default=255, description="Head Track Start Bone")
    headTrackEnd: bpy.props.IntProperty(name="Head Track End Bone", min=0, max=255, default=255, description="Head Track End Bone")
    headTrackFactor: bpy.props.IntProperty(name="Head Track Factor", min=0, max=255, default=1, description="Head Track Factor")
    hotpointGroupCount: bpy.props.IntProperty(name="Hotpoint Group Count", min=-2147483648, max=2147483647, default=1)
    skinnedLayers: bpy.props.IntProperty(name="Skinnedmesh Layers", min=-2147483648, max=2147483647, default=1)
    kfaPointIsRoot: bpy.props.BoolProperty(name="Show Point On Root Bone", default=False, update=on_update_kfaPointIsRoot, description="If ON the point will be parented to the Root Bone (turn on if the animation will be used with Root Animation in game). Else the point will be parented to the model. model or root bone Z is not used, only XY.")
    
    #Hot Point Properties
    #enum of what bit this hotpoint is using (should only be 1 bit. So when this value is 1024 that's only bit 10)
    flags: bpy.props.EnumProperty(name="Flags", items=hpflags_items_cb, description="What bit this Hotpoint is using")
    particle1: bpy.props.IntProperty(name="Particle 1 ID", default=0, min=0, max=65535, description="First Particle to spawn every tick while this hotpoint is enabled")
    particle2: bpy.props.IntProperty(name="Particle 2 ID", default=0, min=0, max=65535, description="Second Particle to spawn every tick while this hotpoint is enabled")
    
    #No Bounding Box Properties
    
    #Anim Properties
    yawOffset: bpy.props.FloatProperty(name="Yaw Offset", subtype='ANGLE', description="Animate this value in your animation actions.")
    
    #UI only Properties
    materialTabIndex: bpy.props.EnumProperty(name="Tabs", items=EnumMaterialTabs, default="SETTINGS")
    skinTabIndex: bpy.props.EnumProperty(name="Tabs", items=EnumSkinTabs, default="ANIM")
    boneTabIndex: bpy.props.EnumProperty(name="Tabs", items=EnumBoneTabs, default="DATA")
    actionTabIndex: bpy.props.EnumProperty(name="Tabs", items=EnumActionTabs, default="DATA")
    hpAddBoneName: bpy.props.StringProperty(name="HotPointAddBoneName", default='')
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_object_hotpoint_bone_list(bpy.types.Operator):
    bl_idname = "turok2_object.hotpoint_bone_list"
    bl_label = "Add HotPoint Bone List"
    bl_description = "Add HotPoint Bone List"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        obj = context.object
        armObj = utils.get_root_object(obj, "ARMATURE")
        armature = armObj.data
        for bone in armature.bones:
            items.append((bone.name, bone.name, "", len(items)))
        #end for
        return items
    #end def
            
    enumprop: bpy.props.EnumProperty(items=items_cb)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE"
    #end def
        
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        boneName = self.enumprop
        obj.t2.hpAddBoneName = boneName
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_object_htbone_list(bpy.types.Operator):
    bl_idname = "turok2_object.htbone_list"
    bl_label = "Object Head Track Bone"
    bl_description = "Select a Armature Bone from a list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        obj = utils.get_root_object(context.object, "ARMATURE")
        armature = obj.data
        for bone in armature.bones:
            items.append((bone.name, bone.name, "", len(items)))
        #end for
        return items
    #end def
            
    enumprop: bpy.props.EnumProperty(items=items_cb)
    isStart: bpy.props.BoolProperty()

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE"
    #end def
        
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        armature = obj.data
        boneName = self.enumprop
        if boneName in armature.bones:
            if self.isStart:
                obj.t2.headTrackStart = armature.bones[boneName].t2.boneIndex
            else:
                obj.t2.headTrackEnd = armature.bones[boneName].t2.boneIndex
            #end if
        #end if
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_object_hotpoint_particle_list(bpy.types.Operator):
    """Set an Arg in a Turok 2 Animation Action from a particle list"""
    bl_idname = "turok2_object.hotpoint_particle_list"
    bl_label = "Hot Point Particle ID"
    bl_description = "Select a Hot Point Particle from a list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        for id in kdefs.defs["particles"]:
            particle = kdefs.defs["particles"][id]
            name = os.path.splitext(os.path.split(particle["path"])[1])[0]
            items.append((str(id), name, "", id))
        #end for
        return items
    #end def

    enumprop: bpy.props.EnumProperty(items=items_cb)
    particleIndex: bpy.props.IntProperty()

    @classmethod
    def poll(cls, context):
        return context.object is not None
    #end def

    def execute(self, context):
        obj = context.object
        id = utils.get_enum_type(self.items_cb(context), self.enumprop)
        if self.particleIndex == 0:
            obj.t2.particle1 = id
        else:
            obj.t2.particle2 = id
        #end if
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_validate_staticemesh(bpy.types.Operator):
    bl_idname = "turok2_object.validate_staticemesh"
    bl_label = "Validate Turok 2 Staticemesh"
    bl_description = "Validates the staticemesh for exporting to Turok 2 and prints any errors to console"

    @classmethod
    def poll(cls, context):
        obj = context.object
        return obj is not None and obj.type == "MESH"
    #end def
        
    def execute(self, context):
        obj = context.object
        validateResults = staticmesh.validate(context, obj)
        if len(validateResults["errors"]) > 0 and len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Errors and Warnings - See System Console for full details", toConsole=False)
        elif len(validateResults["errors"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Error - See System Console for full details", toConsole=False)
        elif len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["warnings"][0]), "Warning - See System Console for full details", toConsole=False)
        else:
            utils.show_message("Everything's good!", "Staticmesh Validation", toConsole=False)
        #end if
        
        if len(validateResults["errors"]) > 0:
            for message in validateResults["errors"]:
                print("[Staticmesh Validation Error] %s" % (message))
            #end for
        #end if
        if len(validateResults["warnings"]) > 0:
            for message in validateResults["warnings"]:
                print("[Staticmesh Validation Warning] %s" % (message))
            #end for
        #end if
        
        return {'FINISHED'}
    #end def
#end class
# -----------------------------------------------------------------------------
#
class TUROK2_OT_validate_skinnedmesh(bpy.types.Operator):
    bl_idname = "turok2_object.validate_skinnedmesh"
    bl_label = "Validate Turok 2 Skinnedmesh"
    bl_description = "Validates the skinnedmesh for exporting to Turok 2 and prints any errors to console"

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.t2.objType == "SKINNEDMESH"
    #end def
        
    def execute(self, context):
        obj = utils.get_root_turok_object2(context.object, "SKINNEDMESH", "ARMATURE")
        validateResults = skinnedmesh.validate(context, obj)
        if len(validateResults["errors"]) > 0 and len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Errors and Warnings - See System Console for full details", toConsole=False)
        elif len(validateResults["errors"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Error - See System Console for full details", toConsole=False)
        elif len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["warnings"][0]), "Warning - See System Console for full details", toConsole=False)
        else:
            utils.show_message("Everything's good!", "Skinnedmesh Validation", toConsole=False)
        #end if
        
        if len(validateResults["errors"]) > 0:
            for message in validateResults["errors"]:
                print("[Skinnedmesh Validation Error] %s" % (message))
            #end for
        #end if
        if len(validateResults["warnings"]) > 0:
            for message in validateResults["warnings"]:
                print("[Skinnedmesh Validation Warning] %s" % (message))
            #end for
        #end if
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_htbone_none(bpy.types.Operator):
    bl_idname = "turok2_object.htbone_none"
    bl_label = "Set HT Bone to None"
    bl_description = "Sets the Head Track Bone to None"
    bl_options = {'INTERNAL', 'UNDO'}
    
    isStart: bpy.props.BoolProperty()
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE"
    #end def
        
    def execute(self, context):
        obj = utils.get_root_object(context.object, "ARMATURE")
        if self.isStart:
            obj.t2.headTrackStart = 255
        else:
            obj.t2.headTrackEnd = 255
        #end if
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_hotpoint_add(bpy.types.Operator):
    bl_idname = "turok2_object.hotpoint_add"
    bl_label = "Add Hot Point"
    bl_description = "Adds a new Hot Point to the armature"
    bl_options = {'INTERNAL', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_object(obj, "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.t2.hpAddBoneName in armObj.data.bones
    #end def
        
    def execute(self, context):
        obj = context.object
        armObj = utils.get_root_object(obj, "ARMATURE")
        bone = armObj.data.bones[armObj.t2.hpAddBoneName]
        utils.create_bone_hotpoint(armObj, bone.t2.boneIndex)
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_bb_create(bpy.types.Operator):
    bl_idname = "turok2_object.bb_create"
    bl_label = "Create Bounding Box"
    bl_description = "Creates a Bounding Box for the Armature Mesh or Bone"
    bl_options = {'INTERNAL', 'UNDO'}
    
    bbBoneName: bpy.props.StringProperty(name="bbBoneName", default='')
    
    @classmethod
    def poll(cls, context):
        return context.object is not None
    #end def
        
    def execute(self, context):
        obj = context.object
        bbDefaultMin = Vector((-10.0, -10.0, -10.0))
        bbDefaultMax = Vector((10.0, 10.0, 10.0))
        if self.bbBoneName:
            if self.bbBoneName not in obj.data.bones:
                return {'CANCELLED'}
            #end if
            bone = obj.data.bones[self.bbBoneName]
            utils.create_bone_bounding_box("Bounding Box", obj, bone.t2.boneIndex, bbDefaultMin, bbDefaultMax)
        elif obj.type == 'ARMATURE' and obj.t2.objType == "SKINNEDMESH":
            meshObj = utils.get_skinnedmesh_mesh(obj)
            if meshObj is None:
                return {'CANCELLED'}
            #end if
            utils.create_bounding_box(meshObj, bbDefaultMin, bbDefaultMax)
        elif obj.type == 'MESH' and obj.t2.objType == "SKINNEDMESH":
            utils.create_bounding_box(obj, bbDefaultMin, bbDefaultMax)
        #end if
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_hotpoint_bb_visibility(bpy.types.Operator):
    bl_idname = "turok2_object.hotpoint_bb_visibility"
    bl_label = "Hot Point Visibility"
    bl_description = "Show and Hide Hot Points"
    bl_options = {'INTERNAL', 'UNDO'}

    selection: bpy.props.EnumProperty(items=EnumHPVisibilities)

    @classmethod
    def poll(cls, context):
        if context is None or not hasattr(context, "object") or not hasattr(context, "view_layer"):
            return False
        #end if
        obj = context.object
        vl = context.view_layer
        if obj is None or vl is None:
            return False
        #end if
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        if armObj.type != "ARMATURE" or armObj.t2.objType != "SKINNEDMESH":
            return False
        #end if

        return True
    #end def

    def execute(self, context):
        obj = bpy.context.object
        vl = bpy.context.view_layer
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        objs = [o for o in bpy.data.objects if o.type == 'EMPTY' and o.t2.objType == "HOT POINT" and utils.obj_child_of(armObj, o)]
        if self.selection == "0":  #Show all Hot Points with the same Bit
            for hpObj in objs:
                if hpObj.t2.flags == obj.t2.flags:
                    hpObj.hide_set(state=False, view_layer=vl)
                    continue
                #end if
                hpObj.hide_set(state=True, view_layer=vl)
            #end for
        elif self.selection == "1":  #Show all Hot Points
            for hpObj in objs:
                hpObj.hide_set(state=False, view_layer=vl)
            #end for
        elif self.selection == "2":  #Hide all Hot Points with the same Bit
            for hpObj in objs:
                if hpObj.t2.flags == obj.t2.flags:
                    hpObj.hide_set(state=True, view_layer=vl)
                    continue
                #end if
            #end for
        elif self.selection == "3":  #Hide all Hot Points
            for hpObj in objs:
                hpObj.hide_set(state=True, view_layer=vl)
            #end for
        #end if

        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_skin_bb_visibility(bpy.types.Operator):
    bl_idname = "turok2_object.skin_bb_visibility"
    bl_label = "BB Visibility"
    bl_description = "Show and Hide Bounding Boxes"
    bl_options = {'INTERNAL', 'UNDO'}

    selection: bpy.props.EnumProperty(items=EnumBBVisibilities)

    @classmethod
    def poll(cls, context):
        return poll_bb_visibility(cls, context, False)
    #end def

    def execute(self, context):
        execute_bb_visibility(self, context, False)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_bone_bb_visibility(bpy.types.Operator):
    bl_idname = "turok2_object.bone_bb_visibility"
    bl_label = "BB Visibility"
    bl_description = "Show and Hide Bounding Boxes"
    bl_options = {'INTERNAL', 'UNDO'}

    selection: bpy.props.EnumProperty(items=EnumBBVisibilities)

    @classmethod
    def poll(cls, context):
        return poll_bb_visibility(cls, context, True)
    #end def

    def execute(self, context):
        execute_bb_visibility(self, context, True)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
def poll_bb_visibility(cls, context, isBone):
    if context is None or not hasattr(context, "object") or not hasattr(context, "view_layer"):
        return False
    #end if
    obj = context.object
    vl = context.view_layer
    if obj is None or vl is None:
        return False
    #end if
    armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
    if armObj.type != "ARMATURE" or armObj.t2.objType != "SKINNEDMESH":
        return False
    #end if
    if isBone:
        if armObj.data.bones.active is None:
            return False
        #end if
    #end if

    return True
#end def

# -----------------------------------------------------------------------------
#
def execute_bb_visibility(self, context, isBone):
    obj = context.object
    vl = context.view_layer
    armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
    boneName = ''
    if isBone:
        boneName = armObj.data.bones.active.name
    #end if
    if obj.t2.objType == "BOUNDING BOX":
        curObj = obj
    else:
        curObj = armObj
    #end if

    bbObjs = [o for o in bpy.data.objects if o.type == 'MESH' and o.t2.objType == "BOUNDING BOX" and utils.obj_child_of(armObj, o)]
    if self.selection == "0":  #"Show Only This Bounding Box"
        for bbObj in bbObjs:
            bbObj.hide_set(state=True, view_layer=vl)
        #end for
        bbObj = utils.get_bounding_box(curObj, boneName)
        if bbObj:
            bbObj.hide_set(state=False, view_layer=vl)
        #end if
    elif self.selection == "1":  #"Show Collision Bounding Boxes"
        for bbObj in bbObjs:
            if bbObj.parent_type == "BONE" and bbObj.parent_bone in armObj.data.bones:
                bone = armObj.data.bones[bbObj.parent_bone]
                flags = utils.get_enumflags_value(panel_bone.EnumBoneFlags, bone.t2.flags)
                if flags > 0:
                    bbObj.hide_set(state=False, view_layer=vl)
                    continue
                #end if
            #end if
            bbObj.hide_set(state=True, view_layer=vl)
        #end for
    elif self.selection == "2":  #"Show All Bounding Boxes"
        for bbObj in bbObjs:
            bbObj.hide_set(state=False, view_layer=vl)
        #end for
    elif self.selection == "3":  #"Hide All Bounding Boxes"
        for bbObj in bbObjs:
            bbObj.hide_set(state=True, view_layer=vl)
        #end for
    elif self.selection == "4":  #"Hide This Bounding Box"
        bbObj = utils.get_bounding_box(curObj, boneName)
        if bbObj:
            bbObj.hide_set(state=True, view_layer=vl)
        #end if
    #end if
#end def

# -----------------------------------------------------------------------------
#
class TUROK2_PT_object(bpy.types.Panel):
    bl_label = "Turok 2 Object"
    bl_idname = "TUROK2_PT_object"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "object"

    def draw(self, context):
        obj = context.object
        if obj is None:
            return
        #end if
        
        layout = self.layout
        row = layout.row(align=True)
        row.label(text="", icon="MONKEY")
        row.prop(obj.t2, "objType", text="Type")
        
        # row = layout.row(align=True)
        # row.operator(TUROK2_OT_obj_test.bl_idname)
        # row.separator()
        
        if obj.t2.objType == "STATICMESH":
            #Validate Staticmesh
            box = layout.box()
            row = box.row()
            row.operator(TUROK2_OT_validate_staticemesh.bl_idname, icon="QUESTION", text="Validate Staticmesh")
            box.separator()

            #TextureSet Preview
            row = box.row()
            showTexSetWarning = obj.t2.textureSetIndex != 0 and len(kdefs.defs["texsets"]) == 0
            if showTexSetWarning:
                row.alert = True
            #end if
            row.label(text="Preview TextureSet")
            row.prop(obj.t2, "textureSetIndex", text="")
            if showTexSetWarning:
                row = box.row()
                row.alignment = 'CENTER'
                row.alert = True
                row.label(text="textureSetInfo def not loaded", icon="ERROR")
            #end if
        elif obj.t2.objType == "SKINNEDMESH":
            if obj.type == 'ARMATURE':
                draw_skinnedmesh(self, context, layout, obj)
            elif obj.type == 'MESH':
                armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", 'ARMATURE')
                if armObj != obj:
                    draw_skinnedmesh(self, context, layout, armObj)
                #end if
            #end if
        elif obj.t2.objType == "BOUNDING BOX":
            if obj.type == 'MESH':
                draw_bounding_box(layout, obj)
            #end if
        elif obj.t2.objType == "HOT POINT":
            row = layout.row()
            row.operator_menu_enum(TUROK2_OT_hotpoint_bb_visibility.bl_idname, "selection", text="---Show/Hide Hot Points---")

            box = layout.box()
            row = box.row(align=True)
            row.prop(obj.t2, "flags", text="Bit")
            row = box.row()
            row.prop(obj.t2, "particle1")
            particle1Text = ''
            particle1 = kdefs.defs["particles"].get(obj.t2.particle1)
            if particle1:
                particle1Text = os.path.splitext(os.path.split(particle1["path"])[1])[0]
            #end if
            row.label(text=particle1Text)
            row.operator(TUROK2_OT_object_hotpoint_particle_list.bl_idname, icon='COLLAPSEMENU', text="").particleIndex = 0
            row = box.row()
            row.prop(obj.t2, "particle2")
            particle2Text = ''
            particle2 = kdefs.defs["particles"].get(obj.t2.particle2)
            if particle2:
                particle2Text = os.path.splitext(os.path.split(particle2["path"])[1])[0]
            #end if
            row.label(text=particle2Text)
            row.operator(TUROK2_OT_object_hotpoint_particle_list.bl_idname, icon='COLLAPSEMENU', text="").particleIndex = 1
        #end if
    #end def
#end class

# -----------------------------------------------------------------------------
# obj is the armature object
def draw_skinnedmesh(self, context, layout, obj):
    row = layout.row()
    row.prop(obj.t2, "skinTabIndex", expand=True)
    layout.box()
    if obj.t2.skinTabIndex == "ANIM":
        #draw key frame action if has animation action
        if obj.animation_data is not None and obj.animation_data.action is not None:
            panel_action.turok2_draw_action_panel(self, context, obj.animation_data.action)
        #end if
    elif obj.t2.skinTabIndex == "HEAD":
        #Head Track Properties
        box = layout.box()
        draw_headtrack_boneindex(box, obj, True)
        draw_headtrack_boneindex(box, obj, False)
        row = box.row()
        row.prop(obj.t2, "headTrackFactor")
    elif obj.t2.skinTabIndex == "BB":
        draw_bounding_box(layout, obj)
    else: #Misc
        #Validate Skinnedmesh
        box = layout.box()
        row = box.row()
        row.operator(TUROK2_OT_validate_skinnedmesh.bl_idname, icon="QUESTION", text="Validate Skinnedmesh")
        box.separator()

        #TextureSet Preview
        row = box.row()
        showTexSetWarning = obj.t2.textureSetIndex != 0 and len(kdefs.defs["texsets"]) == 0
        if showTexSetWarning:
            row.alert = True
        #end if
        row.label(text="Preview TextureSet")
        row.prop(obj.t2, "textureSetIndex", text="")
        if showTexSetWarning:
            row = box.row()
            row.alignment = 'CENTER'
            row.alert = True
            row.label(text="textureSetInfo def not loaded", icon="ERROR")
        #end if
        
        #Yaw Offset
        row = box.row()
        row.label(text="Yaw Offset (Animate)")
        row.prop(obj.t2, "yawOffset", text="")

        #Add Hot Point
        box.separator()
        row = box.row(align=True)
        row.prop(obj.t2, "hpAddBoneName", text="")
        row.operator(TUROK2_OT_object_hotpoint_bone_list.bl_idname, icon='COLLAPSEMENU', text="")
        row.operator(TUROK2_OT_hotpoint_add.bl_idname, text="Add Hot Point")
    #end if
#end def

# -----------------------------------------------------------------------------
#
def draw_headtrack_boneindex(layout, obj, isStart):
    headBoneIndex = obj.t2.headTrackStart if isStart else obj.t2.headTrackEnd
    headBonePropText = "Start Bone" if isStart else "End Bone"
    headBonePropName = "headTrackStart" if isStart else "headTrackEnd"
    boneName = ""
    if headBoneIndex != 255:
        for bone in obj.data.bones:
            if bone.t2.boneIndex == headBoneIndex:
                boneName = bone.name
                break
            #end if
        #end for
    #end if
    row = layout.row(align=True)
    row = row.split(factor=0.5)
    col = row.column()
    subrow = col.row(align=True)
    subrow.alert = headBoneIndex != 255 and not boneName
    subrow.prop(obj.t2, headBonePropName, text=headBonePropText)
    col = row.column()
    subrow = col.row(align=True)
    if headBoneIndex == 255:
        subrow.label(text="None")
        subrow.operator(TUROK2_OT_object_htbone_list.bl_idname, icon='COLLAPSEMENU', text="").isStart = isStart
    else:
        subrow.label(text=boneName)
        subrow.operator(TUROK2_OT_object_htbone_list.bl_idname, icon='COLLAPSEMENU', text="").isStart = isStart
        subrow.operator(TUROK2_OT_htbone_none.bl_idname, icon='X', text="").isStart = isStart
    #end if
#end def

# -----------------------------------------------------------------------------
#
def draw_bounding_box(layout, obj, boneName=''):
    bbObj = utils.get_bounding_box(obj, boneName)
    
    box = layout.box()
    row = box.row()
    if boneName:
        row.operator_menu_enum(TUROK2_OT_bone_bb_visibility.bl_idname, "selection", text="Show/Hide Bounding Boxes")
    else:
        row.operator_menu_enum(TUROK2_OT_skin_bb_visibility.bl_idname, "selection", text="Show/Hide Bounding Boxes")
    #end if
    
    if obj.t2.objType == "BOUNDING BOX" and obj.parent is not None and obj.parent_type == "BONE" and obj.parent_bone:
        bone = obj.parent.data.bones[obj.parent_bone]
        row = box.row()
        row.alignment = "CENTER"
        row.label(text="Bone %i: %s" % (bone.t2.boneIndex, bone.name))
    #end if
    
    if not bbObj:
        row = box.row()
        row.alignment = "CENTER"
        row.operator(TUROK2_OT_bb_create.bl_idname, icon='ERROR', text="Create Bounding Box").bbBoneName = boneName
        return
    #end if

    bbMin, bbMax = utils.get_boundingbox_minmax(bbObj)
    row = box.row()
    row = row.split(factor=0.13)
    row.label(text="Min:")
    row.label(text="%f" % (bbMin[0]))
    row.label(text="%f" % (bbMin[1]))
    row.label(text="%f" % (bbMin[2]))
    row = box.row()
    row = row.split(factor=0.13)
    row.label(text="Max:")
    row.label(text="%f" % (bbMax[0]))
    row.label(text="%f" % (bbMax[1]))
    row.label(text="%f" % (bbMax[2]))
#end def

# def matrix_world(armature_ob, bone_name):
    # local = armature_ob.data.bones[bone_name].matrix_local
    # basis = armature_ob.pose.bones[bone_name].matrix_basis

    # parent = armature_ob.pose.bones[bone_name].parent
    # if parent == None:
        # return  local @ basis
    # else:
        # parent_local = armature_ob.data.bones[parent.name].matrix_local
        # return matrix_world(armature_ob, parent.name) @ (parent_local.inverted() @ local) @ basis

# def editbone_matrix(bone):
    # parent = bone.parent
    # if parent == None:
        # return bone.matrix_local
    # else:
        # return parent.matrix_local.inverted() @ bone.matrix_local
    # #end if
# #end def

# # -----------------------------------------------------------------------------
# #
# class TUROK2_OT_obj_test(bpy.types.Operator):
    # bl_idname = "turok2_object.test"
    # bl_label = "Test"
    # bl_description = "Just a test"
    # bl_options = {'UNDO'}

    # def execute(self, context):
        # # utils.mode_set('OBJECT')
        # obj = context.object
        # if obj is None: #or obj.type != "ARMATURE":
            # return {'FINISHED'}
        # #end if
    
        # utils.update_shader_groups(obj.active_material, obj)
        
        # return {'FINISHED'}
    
        
        # print("")
        # print("------------%s------------" % (context.object.mode))
        
        # bone = obj.data.bones.active
        # editBone = bpy.context.active_bone
        # poseBone = bpy.context.active_pose_bone
        # if bone is not None:
            # print("bone.name: %s" % (bone.name))
            # print("bone.matrix: %s" % (bone.matrix))
            # print("bone.matrix_local: %s" % (bone.matrix_local))
            # tMatrix = Matrix.Translation(Vector((1.0, 2.3, 0.0)))
            # if bone.parent is not None:
                # bone_matrix_local = bone.matrix_local.copy()
                # bone_matrix_local.translation = Vector((0.0, 0.0, 0.0))
                # # newMatrix = bone_matrix_local @ tMatrix
                # parent_matrix2 = obj.pose.bones[bone.parent.name].matrix
                # parent_matrix_local2 = bone.parent.matrix_local.copy()
                # parent_matrix_local2.translation = Vector((0.0, 0.0, 0.0))
                # #newMatrix = bone_matrix_local @ tMatrix
                # newMatrix = bone.convert_local_to_pose(tMatrix, bone_matrix_local, parent_matrix = parent_matrix2, parent_matrix_local = parent_matrix_local2)
                # print("bone.convert_local_to_pose: %s" % (newMatrix))
                # print(newMatrix[0].xyz)
                # print(newMatrix.to_translation())
                # print(newMatrix.to_quaternion())
            # else:
                # #boneRollMatrix = bone.MatrixFromAxisRoll(axis, editBone.roll)
                
                # bone_matrix_local = bone.matrix_local.copy()
                # bone_matrix_local.translation = Vector((0.0, 0.0, 0.0))
                # newMatrix = bone.convert_local_to_pose(tMatrix, bone_matrix_local)
                # #newMatrix = bone.matrix.to_4x4() @ tMatrix
                # #newMatrix = bone_matrix_local @ tMatrix
                # print("bone.convert_local_to_pose: %s" % (newMatrix))
                # print(newMatrix[0].xyz)
                # print(newMatrix.to_translation())
                # print(newMatrix.to_quaternion())
            # #end if
        # #end if
        
        # if poseBone is not None:
            # print("poseBone.name: %s" % (poseBone.name))
            # print("poseBone.matrix: %s" % (poseBone.matrix))
            # print("poseBone.matrix_basis: %s" % (poseBone.matrix_basis))
            # print("poseBone.matrix_channel: %s" % (poseBone.matrix_channel))
            # worldBoneMatrix = matrix_world(obj, poseBone.name)
            # print("worldBoneMatrix: %s" % (worldBoneMatrix))
            # resultMatrix = bone.matrix_local.inverted() @ poseBone.matrix
            # print("resultMatrix: %s" % (resultMatrix))
        # #end if
        
        # editBone = bpy.context.active_bone
        # if editBone is not None:
            # print("editBone.name: %s" % (editBone.name))
            # print("editBone.matrix: %s" % (editBone.matrix))
        # #end if
        # v1 = Vector((0.0, 0.0, 0.0)) #no parent
        # v2 = Vector((-150.0, 0.0, -100.0))
        # v3 = Matrix.Translation(v1).inverted() @ Matrix.Translation(v2)
        # print(v3.translation) #bone 0 = (-150.0, 0.0, -100.0)

        # v4 = v2 #parent is bone 0 matrix_local
        # v5 = Vector((0.0, 0.0, 300.0))
        # v6 = Matrix.Translation(v4).inverted() @ Matrix.Translation(v5)
        # print(v6.translation) #bone 1 = (150.0, 0.0, 400.0)

        # v7 = v5 #parent is bone 1 matrix_local
        # v8 = Vector((0.0, 175.0, 0.0))
        # v9 = Matrix.Translation(v7).inverted() @ Matrix.Translation(v8)
        # print(v9.translation) #bone 2 = (0.0, 175.0, 300.0)
        
        # for boneIndex, bone in enumerate(obj.data.bones):
            # editBoneMatrix = editbone_matrix(bone)
            # print("Bone: %s Vec:%s" % (boneIndex, editBoneMatrix.translation))
        # #end for
        
        # bone = obj.data.bones.active
        # print(Matrix.Identity(3))
        # print(bone.matrix)
        # print("bone.matrix is identity: %s" % (bone.matrix == Matrix.Identity(3)))
        # print("bone vector: %s" % (bone.vector.normalized()))

        # # meshVert = obj.data.vertices[149]
        # # print(meshVert.co)
        # # print(meshVert.undeformed_co)
        # #print(obj.matrix_world.translation)
        # # boneVertGroups = {vertGroup.index:{"vg":vertGroup, "bone":bone} for bone in obj.parent.data.bones for vertGroup in obj.vertex_groups if vertGroup.name == bone.name}
        # # print(boneVertGroups)
        # # mesh = obj.data
        # # loopIndex = 0
        # # meshVert = mesh.vertices[mesh.loops[loopIndex].vertex_index]
        # # # global_matrix = (Matrix.Scale(utils.IMPORT_EXPORT_SCALE, 4) @ Matrix())
        # # #print("global_matrix: %s" % (global_matrix))
        # # print(obj.matrix_world)
        # # #tCopy = obj.matrix_world.copy()
        # # t = Matrix(obj.matrix_world)
        # # t.translation = Vector((0.0, 0.0, 0.0))
        # # print(t)
        # # print("vert %s: %s, %s, %s" % (mesh.loops[loopIndex].vertex_index, meshVert.co, obj.matrix_world @ meshVert.co, t @ meshVert.co))
        # # print("normal: %s, %s" % (meshVert.normal, obj.matrix_world.to_quaternion() @ meshVert.normal))
                        
        # # # meshVert.select = True
        # # utils.mode_set('EDIT')
        # return {'FINISHED'}
    # #end defs
# #end class

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2ObjectProps)
    bpy.utils.register_class(TUROK2_OT_object_hotpoint_bone_list)
    bpy.utils.register_class(TUROK2_OT_object_htbone_list)
    bpy.utils.register_class(TUROK2_OT_object_hotpoint_particle_list)
    bpy.utils.register_class(TUROK2_PT_object)
    bpy.utils.register_class(TUROK2_OT_validate_staticemesh)
    bpy.utils.register_class(TUROK2_OT_validate_skinnedmesh)
    bpy.utils.register_class(TUROK2_OT_htbone_none)
    bpy.utils.register_class(TUROK2_OT_hotpoint_add)
    bpy.utils.register_class(TUROK2_OT_bb_create)
    bpy.utils.register_class(TUROK2_OT_skin_bb_visibility)
    bpy.utils.register_class(TUROK2_OT_bone_bb_visibility)
    bpy.utils.register_class(TUROK2_OT_hotpoint_bb_visibility)
    # bpy.utils.register_class(TUROK2_OT_obj_test)
    bpy.types.Object.t2 = bpy.props.PointerProperty(type=Turok2ObjectProps)
#end def
    
# -----------------------------------------------------------------------------
#
def unregister():
    bpy.utils.unregister_class(Turok2ObjectProps)
    bpy.utils.unregister_class(TUROK2_OT_object_hotpoint_bone_list)
    bpy.utils.unregister_class(TUROK2_OT_object_htbone_list)
    bpy.utils.unregister_class(TUROK2_OT_object_hotpoint_particle_list)
    bpy.utils.unregister_class(TUROK2_PT_object)
    bpy.utils.unregister_class(TUROK2_OT_validate_staticemesh)
    bpy.utils.unregister_class(TUROK2_OT_validate_skinnedmesh)
    bpy.utils.unregister_class(TUROK2_OT_htbone_none)
    bpy.utils.unregister_class(TUROK2_OT_hotpoint_add)
    bpy.utils.unregister_class(TUROK2_OT_bb_create)
    bpy.utils.unregister_class(TUROK2_OT_skin_bb_visibility)
    bpy.utils.unregister_class(TUROK2_OT_bone_bb_visibility)
    bpy.utils.unregister_class(TUROK2_OT_hotpoint_bb_visibility)
    # bpy.utils.unregister_class(TUROK2_OT_obj_test)
    del bpy.types.Object.t2
#end def
