import bpy
import os
from . import utils
from bpy_extras.image_utils import load_image

# -----------------------------------------------------------------------------
#
class Turok2ShadingSettings(bpy.types.PropertyGroup):
    samplerIndex: bpy.props.IntProperty(name="Texture Sampler Index", min=0, max=3, description='If you only have 1 texture node then keep this at 0', default=0)
    mirrored: bpy.props.BoolProperty(name="Mirrored", description='Overrides texture wrapping to "mirrored" which flips the texture on the x and y when it repeats', default=False)
    anisotropicFilter: bpy.props.BoolProperty(name="Anisotropic Filter", description='Overrides texture filtering to "linear_anisotropic" when Linear, Cubic, or Smart is selected', default=True)
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_set_mirrored_tex(bpy.types.Operator):
    bl_idname = "turok2_shading.set_mirrored"
    bl_label = "Mirror Tex"
    bl_description = 'Overrides texture wrapping to "mirrored" which flips the texture on the x and y when it repeats'
    bl_options = {'INTERNAL', 'UNDO'}

    setValue: bpy.props.BoolProperty(name="New Value", default=False)
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        node = context.active_node
        return obj is not None and obj.active_material is not None and node is not None and node.type == "TEX_IMAGE"
    #end def

    def execute(self, context):
        obj = context.object
        node = context.active_node
        mat = obj.active_material
        set_mirrored(context, obj, mat, node, self.setValue)
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_set_anisotropic_tex(bpy.types.Operator):
    bl_idname = "turok2_shading.set_anisotropic"
    bl_label = "Anisotropic Filter"
    bl_description = 'Overrides texture filtering to "linear_anisotropic" when Linear, Cubic, or Smart is selected'
    bl_options = {'INTERNAL', 'UNDO'}

    setValue: bpy.props.BoolProperty(default=False)
    
    @classmethod
    def poll(cls, context):
        obj = context.object
        node = context.active_node
        return obj is not None and obj.active_material is not None and node is not None and node.type == "TEX_IMAGE"
    #end def

    def execute(self, context):
        obj = context.object
        node = context.active_node
        mat = obj.active_material
        node.t2.anisotropicFilter = self.setValue
        mat.use_backface_culling = mat.use_backface_culling #Forces rendering to refresh
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class Turok2ShadingPanel(bpy.types.Panel):
    bl_label = "Turok 2 Shading Node"
    bl_idname = "SHADING_PT_Turok2ShadingNodeProp"
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_context = "Node"
    bl_category = "Turok2"
    
    @classmethod
    def poll(cls, context):
        return context.active_node is not None
    #end def

    def draw(self, context):
        node = context.active_node
        if node.type != "TEX_IMAGE":
            return
        #end if
        layout = self.layout
        draw_shading_panel(layout, context, node, TUROK2_OT_set_mirrored_tex, TUROK2_OT_set_anisotropic_tex, False)
    #end def
    
#end class

def set_mirrored(context, obj, mat, node, mirrorValue):
    node.t2.mirrored = mirrorValue
    utils.update_shader_groups(mat, obj)
#end def

def draw_shading_panel(layout, context, node, mirrorOT, anisotropicOT, matQuickLook):
    mirrorText = "ON" if node.t2.mirrored else "OFF"
    anisoText = "ON" if node.t2.anisotropicFilter else "OFF"
    if matQuickLook:
        row = layout.row(align=True)
        row = row.split(factor=0.65)
        row.label(text="Mirror Wrap:")
        row.operator(mirrorOT.bl_idname, text=mirrorText).setValue = not node.t2.mirrored
        row = layout.row(align=True)
        row = row.split(factor=0.65)
        row.label(text="Anisotropic:")
        row.operator(anisotropicOT.bl_idname, text=anisoText).setValue = not node.t2.anisotropicFilter
    else:
        row = layout.row(align=True)
        row.label(text="Sampler Index")
        row.prop(node.t2, "samplerIndex", text="")
        row = layout.row(align=True)
        row.label(text="Mirror Wrap")
        row.operator(mirrorOT.bl_idname, text=mirrorText).setValue = not node.t2.mirrored
        row = layout.row(align=True)
        row.label(text="Anisotropic")
        row.operator(anisotropicOT.bl_idname, text=anisoText).setValue = not node.t2.anisotropicFilter
    #end if
#end def

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2ShadingSettings)
    bpy.utils.register_class(TUROK2_OT_set_mirrored_tex)
    bpy.utils.register_class(TUROK2_OT_set_anisotropic_tex)
    bpy.utils.register_class(Turok2ShadingPanel)
    bpy.types.ShaderNodeTexImage.t2 = bpy.props.PointerProperty(type=Turok2ShadingSettings)
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(Turok2ShadingSettings)
    bpy.utils.unregister_class(TUROK2_OT_set_mirrored_tex)
    bpy.utils.unregister_class(TUROK2_OT_set_anisotropic_tex)
    bpy.utils.unregister_class(Turok2ShadingPanel)
    del bpy.types.ShaderNodeTexImage.t2
#end def
