import os
import bpy
from enum import Enum
import math
from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from bpy_extras.image_utils import load_image

from . import (
    kdefs,
    utils,
    paths,
    panel_preferences,
    staticmesh,
    skinnedmesh,
    aud_utils,
)

EnumToolsTabs = [
    ("DEFS", "Reload Def Files", "", 0),
    ("WEAPON", "Setup Weapon Camera", "", 1),
    ("SOUNDS", "Preview Sounds", "", 2),
    ("STATIC", "Preview Staticmeshes", "", 3),
    ("SKIN", "Preview Skinnedmeshes", "", 4),
    ("MISC", "Misc", "", 5),
]

quickStaticmesh = None
quickSkinnedmesh = None
# -----------------------------------------------------------------------------
#
class Turok2SceneFileListPropGroup(bpy.types.PropertyGroup):
    filename: bpy.props.StringProperty(name="", default="", subtype="FILE_NAME")
    path: bpy.props.StringProperty(name="", default="", subtype="FILE_PATH")
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_UL_scene_filelist_props(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        row = layout.row()
        row.label(text=item.filename, translate=False)
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_play_sound_id(bpy.types.Operator):
    bl_idname = "turok2_scene.play_sound_id"
    bl_label = "Play Sound"
    bl_description = "Play sound with SoundID and index"
    bl_options = {'INTERNAL'}

    soundID: bpy.props.IntProperty(default=10)
    soundIDIndex: bpy.props.IntProperty(default=0)

    def execute(self, context):
        aud_utils.stop_sound()
        sound = kdefs.get_sound(self.soundID, self.soundIDIndex)
        aud_utils.play_ksnd(sound)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_play_voice(bpy.types.Operator):
    bl_idname = "turok2_scene.play_voice_id"
    bl_label = "Play Voice"
    bl_description = "Play voice with VoiceID"
    bl_options = {'INTERNAL'}

    voiceID: bpy.props.IntProperty()

    def execute(self, context):
        aud_utils.stop_sound()
        aud_utils.play_voice(self.voiceID)
        
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_stop_all_sounds(bpy.types.Operator):
    bl_idname = "turok2_scene.stop_all_sounds"
    bl_label = "Stop all sounds"
    bl_description = "Stop all Sounds"
    bl_options = {'INTERNAL'}
    
    def execute(self, context):
        aud_utils.stop_sound()
        return {'FINISHED'}
    #end def
    
#end class

# -----------------------------------------------------------------------------
#
class Turok2SceneSettings(bpy.types.PropertyGroup):
    def on_sound_folder_changed(self, context):
        absPath = os.path.abspath(self.soundFolder)
        if os.path.isdir(absPath):
            self.sounds.clear()
            self.soundListLastID = ''
            filenames = [f for f in os.listdir(absPath) if f.endswith('.ksnd')]
            filenames.sort()
            for filename in filenames:
                item = self.sounds.add()
                item.name = filename
                item.filename = filename
                item.path = os.path.join(absPath, filename)
            #end for
        #end if
    #end def
    
    def on_sound_index_changed(self, context):
        if self.soundIndex >= 0 and self.soundIndex < len(self.sounds):
            aud_utils.stop_sound()
            sound = aud_utils.load_play_ksnd(self.sounds[self.soundIndex].path, False)
            if sound is not None:
                self.soundListLastID = "SoundID: %s" % (sound["id"])
            else:
                self.soundListLastID = ''
            #end if
        #end if
    #end def
    
    def on_staticmesh_index_changed(self, context):
        global quickStaticmesh
        
        scene = context.scene
        if scene.t2.staticmeshListIndex >= 0 and scene.t2.staticmeshListIndex < len(scene.t2.staticmeshList):
            utils.mode_set('OBJECT')
            #Remove last loaded by this func
            if quickStaticmesh is not None:
                try:
                    bpy.data.objects.remove(quickStaticmesh)
                except:
                    pass
                #end try except
                utils.clean_all_unused_model_data()
                quickStaticmesh = None
            #end if
            
            #Import staticmesh and cache it to remove it when this func is called again
            item = scene.t2.staticmeshList[scene.t2.staticmeshListIndex]
            global_matrix = (Matrix.Scale(1.0/utils.IMPORT_EXPORT_SCALE, 4) @ Matrix())
            quickStaticmesh = staticmesh.do_import(context, item.path, global_matrix)
            if quickStaticmesh is not None:
                utils.view_selected_object()
            #end if
        #end if
    #end def
    
    def on_get_soundtestid(self):
        return self.get("soundTestID", 10)
    #end if
    def on_set_soundtestid(self, value):
        self["soundTestID"] = value
        
        index = self.get("soundTestIDIndex", 0)
        if value in kdefs.defs["sounds"]:
            index = max(0, min(index, len(kdefs.defs["sounds"][value]) - 1))
            self["soundTestIDIndex"] = index
        #end if
    #end if
    
    def on_get_soundtestidindex(self):
        return self.get("soundTestIDIndex", 0)
    #end if
    def on_set_soundtestidindex(self, value):
        id = self.get("soundTestID", 10)
        if id in kdefs.defs["sounds"]:
            value = max(0, min(value, len(kdefs.defs["sounds"][id]) - 1))
        #end if
        self["soundTestIDIndex"] = value
    #end if
    
    def on_staticmesh_folder_changed(self, context):
        absPath = os.path.abspath(self.staticmeshsFolder)
        if os.path.isdir(absPath):
            self.staticmeshList.clear()
            filenames = [f for f in os.listdir(absPath) if f.endswith('.staticmesh')]
            filenames.sort()
            for filename in filenames:
                item = self.staticmeshList.add()
                item.name = filename
                item.filename = filename
                item.path = os.path.join(absPath, filename)
            #end for
        #end if
    #end def
    
    def on_skinnedmesh_index_changed(self, context):
        global quickSkinnedmesh
        
        scene = context.scene
        if scene.t2.skinnedmeshListIndex >= 0 and scene.t2.skinnedmeshListIndex < len(scene.t2.skinnedmeshList):
            utils.mode_set('OBJECT')
            #Remove last loaded by this func
            if quickSkinnedmesh is not None:
                objs = [o for o in bpy.data.objects if utils.obj_child_of(quickSkinnedmesh, o, True)]
                bpy.ops.object.delete({"selected_objects": objs})
                #Force remove all actions, materials and images
                for action in bpy.data.actions:
                    bpy.data.actions.remove(action)
                #end for
                for mat in bpy.data.materials:
                    bpy.data.materials.remove(mat)
                #end for
                for img in bpy.data.images:
                    bpy.data.images.remove(img)
                #end for
                
                utils.clean_all_unused_model_data()
                quickSkinnedmesh = None
            #end if
            
            #Import skinnedmesh and cache it to remove it when this func is called again
            item = scene.t2.skinnedmeshList[scene.t2.skinnedmeshListIndex]
            global_matrix = (Matrix.Scale(1.0/utils.IMPORT_EXPORT_SCALE, 4) @ Matrix())
            quickSkinnedmesh, meshObj = skinnedmesh.do_import(context, item.path, global_matrix, True, True, False)
            if quickSkinnedmesh is not None:
                if meshObj is not None:
                    meshObj.select_set(True)
                #end if
                utils.view_selected_object()
            #end if
        #end if
    #end def

    def on_skinnedmesh_folder_changed(self, context):
        absPath = os.path.abspath(self.skinnedmeshsFolder)
        if os.path.isdir(absPath):
            self.skinnedmeshList.clear()
            filenames = [f for f in os.listdir(absPath) if f.endswith('.skinnedmesh')]
            filenames.sort()
            for filename in filenames:
                item = self.skinnedmeshList.add()
                item.name = filename
                item.filename = filename
                item.path = os.path.join(absPath, filename)
            #end for
        #end if
    #end def
    
    def on_weapon_offset_changed(self, context):
        camObj = utils.get_object(utils.TUROK2_WEAPON_CAMERA_NAME)
        if camObj is not None:
            offset = self.weaponOffset * utils.TUROK2_WEAPON_OFFSET_SCALE
            camObj.location = (utils.TUROK2_WEAPON_INIT_TRANSLATION * utils.TUROK2_WEAPON_OFFSET_SCALE) + offset
        #end if
    #end def

    sounds: bpy.props.CollectionProperty(name="Sounds", type = Turok2SceneFileListPropGroup)
    soundIndex: bpy.props.IntProperty(name="", default = -1, update=on_sound_index_changed)
    soundTestID: bpy.props.IntProperty(name="Sound ID", default = 10, get=on_get_soundtestid, set=on_set_soundtestid)
    soundTestIDIndex: bpy.props.IntProperty(name="Sound Index", default = 0, min=0, max=2147483647, get=on_get_soundtestidindex, set=on_set_soundtestidindex)
    soundListLastID: bpy.props.StringProperty(name="", default='', description="")
    soundFolder: bpy.props.StringProperty(name="", default='', description="", subtype="DIR_PATH", update=on_sound_folder_changed)
    
    staticmeshList: bpy.props.CollectionProperty(name="Staticmeshes", type = Turok2SceneFileListPropGroup)
    staticmeshListIndex: bpy.props.IntProperty(name="", default = -1, update=on_staticmesh_index_changed)
    staticmeshsFolder: bpy.props.StringProperty(name="", default='', description="", subtype="DIR_PATH", update=on_staticmesh_folder_changed)
    
    skinnedmeshList: bpy.props.CollectionProperty(name="Skinnedmeshes", type = Turok2SceneFileListPropGroup)
    skinnedmeshListIndex: bpy.props.IntProperty(name="", default = -1, update=on_skinnedmesh_index_changed)
    skinnedmeshsFolder: bpy.props.StringProperty(name="", default='', description="", subtype="DIR_PATH", update=on_skinnedmesh_folder_changed)
    
    weaponOffset: bpy.props.FloatVectorProperty(name="Weapon Offset", description="Weapon Offset value as defined in defs/weaponInfo.txt", default=Vector((0.0, 0.0, 0.0)), subtype='XYZ', size = 3, step=10, precision=6, update=on_weapon_offset_changed)
    
    toolsTabIndex: bpy.props.EnumProperty(name="Tabs", items=EnumToolsTabs, default="MISC")

#end class

# -----------------------------------------------------------------------------
#
class TUROK2_PT_tools(bpy.types.Panel):
    bl_label = "Turok 2 Tools"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Turok 2"
    
    def draw(self, context):
        scene = context.scene
        if scene is None:
            return
        #end if
        layout = self.layout
        row = layout.row()
        row.prop(scene.t2, "toolsTabIndex", text="")
        if scene.t2.toolsTabIndex == "DEFS":
            baseDefsLoaded = len(kdefs.defs["anims"]) > 0 and len(kdefs.defs["anim_events"]) > 0 and len(kdefs.defs["impacts"]) > 0 and len(kdefs.defs["hotpoints"]) > 0 and len(kdefs.defs["builtin_kmats"]) > 0 and len(kdefs.defs["texsets"]) > 0 and len(kdefs.defs["voices"]) > 0
            allDefsLoaded = baseDefsLoaded and len(kdefs.defs["actors"]) > 0 and len(kdefs.defs["particles"]) > 0 and len(kdefs.defs["sounds"]) > 0
            row = layout.row()
            row.alert = not allDefsLoaded
            opIcon = 'NONE' if row.alert else 'CHECKMARK'
            row.operator(panel_preferences.TUROK2_OT_pref_load_defs_all.bl_idname, text="All", icon=opIcon)
            row = layout.row()
            row.alert = not baseDefsLoaded
            opIcon = 'NONE' if row.alert else 'CHECKMARK'
            row.operator(panel_preferences.TUROK2_OT_pref_load_defs_base.bl_idname, text="Base", icon=opIcon)
            row = layout.row()
            row.alert = len(kdefs.defs["actors"]) == 0
            opIcon = 'NONE' if row.alert else 'CHECKMARK'
            row.operator(panel_preferences.TUROK2_OT_pref_load_defs_actors.bl_idname, text="Actors", icon=opIcon)
            row = layout.row()
            row.alert = len(kdefs.defs["particles"]) == 0
            opIcon = 'NONE' if row.alert else 'CHECKMARK'
            row.operator(panel_preferences.TUROK2_OT_pref_load_defs_particles.bl_idname, text="Particles", icon=opIcon)
            row = layout.row()
            row.alert = len(kdefs.defs["sounds"]) == 0
            opIcon = 'NONE' if row.alert else 'CHECKMARK'
            row.operator(panel_preferences.TUROK2_OT_pref_load_defs_sounds.bl_idname, text="Sounds", icon=opIcon)
        elif scene.t2.toolsTabIndex == "WEAPON":
            row = layout.row()
            row.alignment = 'CENTER'
            row.prop(scene.t2, "weaponOffset", text="")
            row = layout.row()
            row.operator(TUROK2_OT_scene_setup_weapon_camera.bl_idname, icon='OUTLINER_OB_CAMERA', text="Setup Camera")
        elif scene.t2.toolsTabIndex == "SOUNDS":
            row = layout.row()
            soundTestText = "Sounds (%d)" % (len(scene.t2.sounds))
            row.label(text=soundTestText, icon="PLAY_SOUND", translate=False)
            box = layout.box()
            row = box.row()
            row.prop(scene.t2, "soundFolder")
            row = box.row()
            row.template_list('TUROK2_UL_scene_filelist_props', "Turok2SceneFileListPropGroup", scene.t2, "sounds", scene.t2, "soundIndex")
            row = box.row()
            row.label(text=scene.t2.soundListLastID)
            box = layout.box()
            row = box.row(align=True)
            showSoundDefWarning = len(kdefs.defs["sounds"]) == 0
            if showSoundDefWarning:
                row.alert = True
            #end if
            row.prop(scene.t2, "soundTestID", text="")
            row.prop(scene.t2, "soundTestIDIndex", text="")
            row.separator()
            opSoundProps = row.operator(TUROK2_OT_play_sound_id.bl_idname, icon='PLAY', text="")
            opSoundProps.soundID = scene.t2.soundTestID
            opSoundProps.soundIDIndex = scene.t2.soundTestIDIndex
            row.operator(TUROK2_OT_stop_all_sounds.bl_idname, icon='PAUSE', text="")
            row.separator()
            row.operator(TUROK2_OT_scene_sound_list.bl_idname, icon='COLLAPSEMENU', text="")
            if showSoundDefWarning:
                row = box.row()
                row.alignment = 'CENTER'
                row.alert = True
                row.label(text="ksnds not loaded", icon="ERROR")
            #end if
            row = box.row()
            sound = kdefs.get_sound(scene.t2.soundTestID, scene.t2.soundTestIDIndex)
            if sound is not None:
                fileDir, fileNameWithExt = os.path.split(sound["path"])
                filename, ext = os.path.splitext(fileNameWithExt)
                row.label(text=filename)
            #end if
        elif scene.t2.toolsTabIndex == "STATIC":
            row = layout.row()
            staticmeshTitleText = "Staticmeshes (%d)" % (len(scene.t2.staticmeshList))
            row.label(text=staticmeshTitleText, icon="MESH_CUBE", translate=False)
            box = layout.box()
            row = box.row()
            row.prop(scene.t2, "staticmeshsFolder")
            row = box.row()
            row.template_list('TUROK2_UL_scene_filelist_props', "Turok2SceneFileListPropGroup", scene.t2, "staticmeshList", scene.t2, "staticmeshListIndex")
            row = box.row()
            row.operator(TUROK2_OT_scene_staticmesh_disconnect.bl_idname, icon='UNLINKED', text="Disconnect")
        elif scene.t2.toolsTabIndex == "SKIN":
            row = layout.row()
            skinnedmeshTitleText = "Skinnedmeshes (%d)" % (len(scene.t2.skinnedmeshList))
            row.label(text=skinnedmeshTitleText, icon="ARMATURE_DATA", translate=False)
            box = layout.box()
            row = box.column(align=True)
            row.label(text="Warning:  this will remove", icon='ERROR')
            row.label(text="all Actions, Materials, Images")
            row = box.row()
            row.prop(scene.t2, "skinnedmeshsFolder")
            row = box.row()
            row.template_list('TUROK2_UL_scene_filelist_props', "Turok2SceneFileListPropGroup", scene.t2, "skinnedmeshList", scene.t2, "skinnedmeshListIndex")
            row = box.row()
            row.operator(TUROK2_OT_scene_skinnedmesh_disconnect.bl_idname, icon='UNLINKED', text="Disconnect")
        else: #Misc
            row = layout.row()
            row.alignment = "CENTER"
            row.label(text="%s.%s.%s" % (utils.TUROK2_BL_INFO["version"][0], utils.TUROK2_BL_INFO["version"][1], utils.TUROK2_BL_INFO["version"][2]))
            row = layout.row()
            row.operator(TUROK2_OT_scene_clear_unused_data.bl_idname, icon='FILE_BLEND', text="Clear Unused Blender Data")
            row = layout.row()
            row.operator(panel_preferences.TUROK2_OT_pref_open_addon_folder.bl_idname, icon='FILE_FOLDER', text="Open Add-on Folder")
            row = layout.row()
            if utils.TUROK2_BL_INFO["wiki_url"]:
                row.operator("wm.url_open", text="Online Documentation", icon='HELP').url = utils.TUROK2_BL_INFO["wiki_url"]
            #end if
            row = layout.row()
            row.operator(TUROK2_OT_scene_setup_player_model.bl_idname, icon='OUTLINER_OB_MESH', text="Setup Player Model Scale")
        #end if
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_clear_unused_data(bpy.types.Operator):
    bl_idname = "turok2_scene.clear_unused_data"
    bl_label = "Clear All Unused Data"
    bl_description = "Clears all unused Blender Data in the file that have no users"
    bl_options = {'UNDO'}
    
    def execute(self, context):
        removeCount = utils.clean_all_unused_data()
        utils.show_message("%s Data Blocks Cleared" % (removeCount), "", toConsole=False)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_sound_list(bpy.types.Operator):
    """Set SoundID with sound list"""
    bl_idname = "turok2_scene.test_sound_list"
    bl_label = "Set SoundID with sound list"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_property = "enumprop"
    
    def items_cb(self, context):
        items = []
        
        for id in kdefs.defs["sounds"]:
            for sound in kdefs.defs["sounds"][id]:
                fileDir, fileNameWithExt = os.path.split(sound["path"])
                filename, ext = os.path.splitext(fileNameWithExt)
                enumKey = "%s %s" % (id, sound["index"])
                items.append((enumKey, filename, "", len(items)))
            #end for
        return items
    #end def
    
    enumprop: bpy.props.EnumProperty(items=items_cb)

    @classmethod
    def poll(cls, context):
        return context.scene is not None
    #end def
        
    def execute(self, context):
        props = context.scene.t2
        
        enumData = self.enumprop.split()
        props.soundTestID = int(enumData[0])
        props.soundTestIDIndex = int(enumData[1])
        sound = kdefs.get_sound(props.soundTestID, props.soundTestIDIndex)
        if sound is not None:
            aud_utils.play_ksnd(sound)
        #end if
        
        return {'FINISHED'}
    #end def

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_staticmesh_disconnect(bpy.types.Operator):
    bl_idname = "turok2_scene.staticmesh_disconnect"
    bl_label = "Disconnect Staticmesh"
    bl_description = "Will not remove the last added staticmesh when selecting a new staticmesh"
    bl_options = {'INTERNAL'}
    
    @classmethod
    def poll(cls, context):
        global quickStaticmesh
        return quickStaticmesh is not None
    #end def

    def execute(self, context):
        global quickStaticmesh
        quickStaticmesh = None
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_skinnedmesh_disconnect(bpy.types.Operator):
    bl_idname = "turok2_scene.skinnedmesh_disconnect"
    bl_label = "Disconnect Skinnedmesh"
    bl_description = "Will not remove the last added Skinnedmesh when selecting a new Skinnedmesh"
    bl_options = {'INTERNAL'}
    
    @classmethod
    def poll(cls, context):
        global quickSkinnedmesh
        return quickSkinnedmesh is not None
    #end def

    def execute(self, context):
        global quickSkinnedmesh
        quickSkinnedmesh = None
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_setup_player_model(bpy.types.Operator):
    bl_idname = "turok2_scene.setup_player_model"
    bl_label = "Setup Player Model"
    bl_description = "Creates a box object that is the same scale as Turok"
    bl_options = {'REGISTER', 'UNDO'}
    
    useGameScale: bpy.props.BoolProperty(default=True, description="Will scale the player by the actor of 2.855002")

    def execute(self, context):
        playerObj = utils.get_object(utils.TUROK2_PLAYER_SCALE_NAME)
        if playerObj is None:
            scaleFactor = utils.TUROK2_ACTOR_SCALE if self.useGameScale else 1.0
            playerObj = utils.create_cube(utils.TUROK2_PLAYER_SCALE_NAME, Vector((-40.0, -40.0, 0.0)) * scaleFactor, Vector((40.0, 40.0, 100.0)) * scaleFactor, True)
        else:
            return {'CANCELLED'}
        #end if
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
class TUROK2_OT_scene_setup_weapon_camera(bpy.types.Operator):
    bl_idname = "turok2_scene.setup_weapon_cam"
    bl_label = "Setup Weapon Camera"
    bl_description = "Creates a camera object that has the same settings as the Turok2EX camera"
    bl_options = {'UNDO'}
    
    def execute(self, context):
        camObj = utils.get_object(utils.TUROK2_WEAPON_CAMERA_NAME)
        if camObj is None:
            #create camera data
            cam = bpy.data.cameras.get(utils.TUROK2_WEAPON_CAMERA_NAME)
            if cam is None:
                cam = bpy.data.cameras.new(utils.TUROK2_WEAPON_CAMERA_NAME)
            camObj = utils.create_turok2_object(utils.TUROK2_WEAPON_CAMERA_NAME, cam, "NONE", rotMode="XYZ")
        #end if
        
        if camObj.type == "CAMERA":
            #setup scene
            scene = context.scene
            scene.render.resolution_x = 1920
            scene.render.resolution_y = 1080
            scene.render.pixel_aspect_x = 1.0
            scene.render.pixel_aspect_y = 1.0
            
            #setup camera properties
            cam = camObj.data
            weaponOffset = scene.t2.weaponOffset * utils.TUROK2_WEAPON_OFFSET_SCALE
            camObj.location = (utils.TUROK2_WEAPON_INIT_TRANSLATION * utils.TUROK2_WEAPON_OFFSET_SCALE) + weaponOffset
            camObj.rotation_euler = (math.radians(90.0), 0.0, math.radians(180))
            camObj.scale = (200.0, 200.0, 200.0)
            camObj.lock_location = (True, True, True)
            camObj.lock_rotation = (True, True, True)
            cam.clip_end = 10000
            cam.sensor_width = 35.0
            cam.lens_unit = "FOV"
            cam.angle = math.atan((scene.render.resolution_x * 0.5) / (scene.render.resolution_y * 0.5 / math.tan(math.radians(utils.TUROK2_WEAPON_FOV) * 0.5))) * 2.0
            cam.show_background_images = True
            #setup background image
            if len(cam.background_images) <= 0:
                camImageBack = cam.background_images.new()
            else:
                camImageBack = cam.background_images[0]
            #end if
            camImageBack.image = load_image(os.path.join(paths.addon_texture_dir(), "t2_camera_bg_back.jpg"), check_existing=True)
            camImageBack.alpha = 1.0
            camImageBack.display_depth = 'BACK'
            
            #setup foreground image
            if len(cam.background_images) <= 1:
                camImageFront = cam.background_images.new()
            else:
                camImageFront = cam.background_images[1]
            #end if
            camImageFront.image = load_image(os.path.join(paths.addon_texture_dir(), "t2_camera_bg_front.png"), check_existing=True)
            camImageFront.alpha = 0.7
            camImageFront.display_depth = 'FRONT'
        #end if
            
        return {'FINISHED'}
    #end def
#end class

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(Turok2SceneFileListPropGroup)
    bpy.utils.register_class(TUROK2_UL_scene_filelist_props)
    bpy.utils.register_class(Turok2SceneSettings)
    bpy.utils.register_class(TUROK2_PT_tools)
    bpy.utils.register_class(TUROK2_OT_scene_clear_unused_data)
    bpy.utils.register_class(TUROK2_OT_scene_sound_list)
    bpy.utils.register_class(TUROK2_OT_scene_staticmesh_disconnect)
    bpy.utils.register_class(TUROK2_OT_scene_skinnedmesh_disconnect)
    bpy.utils.register_class(TUROK2_OT_scene_setup_player_model)
    bpy.utils.register_class(TUROK2_OT_scene_setup_weapon_camera)
    bpy.utils.register_class(TUROK2_OT_play_sound_id)
    bpy.utils.register_class(TUROK2_OT_play_voice)
    bpy.utils.register_class(TUROK2_OT_stop_all_sounds)
    bpy.types.Scene.t2 = bpy.props.PointerProperty(type=Turok2SceneSettings)
#end def

# -----------------------------------------------------------------------------
#     
def unregister():
    bpy.utils.unregister_class(Turok2SceneFileListPropGroup)
    bpy.utils.unregister_class(TUROK2_UL_scene_filelist_props)
    bpy.utils.unregister_class(Turok2SceneSettings)
    bpy.utils.unregister_class(TUROK2_PT_tools)
    bpy.utils.unregister_class(TUROK2_OT_scene_clear_unused_data)
    bpy.utils.unregister_class(TUROK2_OT_scene_sound_list)
    bpy.utils.unregister_class(TUROK2_OT_scene_staticmesh_disconnect)
    bpy.utils.unregister_class(TUROK2_OT_scene_skinnedmesh_disconnect)
    bpy.utils.unregister_class(TUROK2_OT_scene_setup_player_model)
    bpy.utils.unregister_class(TUROK2_OT_scene_setup_weapon_camera)
    bpy.utils.unregister_class(TUROK2_OT_play_sound_id)
    bpy.utils.unregister_class(TUROK2_OT_play_voice)
    bpy.utils.unregister_class(TUROK2_OT_stop_all_sounds)
    del bpy.types.Scene.t2
#end def
