import bpy
import os
import bmesh
import re
import ctypes
import functools
from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from shutil import copyfile
from bpy_extras.image_utils import load_image
from bpy_extras.wm_utils.progress_report import (
    ProgressReport,
    ProgressReportSubstep,
)
from bpy.props import (
        CollectionProperty,
        BoolProperty,
        FloatProperty,
        StringProperty,
        EnumProperty,
        )
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper,
        axis_conversion,
        )
from . import (
    utils,
    file_utils,
    paths,
    kmaterial,
    kdefs,
    panel_bone,
    panel_object,
    animation,
)

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_import_skinnedmesh(bpy.types.Operator, ImportHelper):
    """Import a Turok 2 Skinnedmesh File"""
    bl_idname = "turok2_object.import_skinnedmesh"
    bl_label = "Import Skinnedmesh"
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

	# Properties used by the file browser
    filepath: StringProperty(name="File Path", description="File path used for importing the skinnedmesh file", default="", options={'HIDDEN'})
    files: CollectionProperty(type=bpy.types.OperatorFileListElement, options={'HIDDEN'})
    directory: StringProperty(subtype='FILE_PATH', options={'HIDDEN'})
    filter_folder: BoolProperty(name="Filter Folders", description="", default=True, options={'HIDDEN'})
    filename_ext = ".skinnedmesh"
    filter_glob: StringProperty(default="*.skinnedmesh", options={'HIDDEN'})

    # Custom properties
    hideBBHot: BoolProperty(name="Hide BB & Hot Points", description="Hide Bounding Boxes and Hot Points", default=True)
    importAnim: BoolProperty(name="Import Anim", description="Import the .anim file at the same time from the <unzipped game/Turok 2 folder>/anims/ folder. Assumes the Skinnedmesh your loading is in <unzipped game/Turok 2 folder>/char", default=True)
    animBakeYawOffsets: BoolProperty(name="Bake Yaw Offsets", description="For each animation, rotates the root bone by it's Yaw Offset. (Do not turn on if you intend to export this Skinnedmesh back into Turok 2)", default=False)

    def execute(self, context):
        global_matrix = (Matrix.Scale(1.0/utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward, to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        armObj, meshObj = do_import(context, self.filepath, global_matrix, self.hideBBHot, self.importAnim, self.animBakeYawOffsets)
        if armObj is None:
            return {'CANCELLED'}
        #end if

        return {'FINISHED'}
    #end def
    
    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Skinnedmesh Options:", icon="MONKEY")
        box = layout.box()
        row = box.row()
        row.prop(self, "hideBBHot")
        row = box.row()
        row.prop(self, "importAnim")
        if self.importAnim:
            row = layout.row()
            row.label(text="Animation Options:", icon="ANIM")
            box = layout.box()
            row = box.row()
            row.prop(self, "animBakeYawOffsets")
        #end if
    #end def
#end class

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_export_skinnedmesh(bpy.types.Operator, ExportHelper):
    """Export a Turok 2 Skinnedmesh File"""
    bl_idname = "turok2_object.export_skinnedmesh"
    bl_label = 'Export Skinnedmesh'
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

    filename_ext = ".skinnedmesh"
    filter_glob: StringProperty(default="*.skinnedmesh", options={'HIDDEN'})

    exportMaterials: BoolProperty(name="Export Materials", description="Export all used materials to .kmat files relative to the export directory", default=True)
    copyTextures: BoolProperty(name="Copy Textures", description="Copies all used textures to their path relative to the export directory", default=False)
    copyTextureSets: BoolProperty(name="Copy All Textures in Texture Sets", description="Copies all textures in every texture set this skinnedmesh uses", default=False)
    useCharDir: BoolProperty(name="Use char Directory", description="Exports materials, textures, and animations to the parent folder of \"char\" + the sub folders after the char folder if it exists in the export path.", default=True)
    exportAnim: BoolProperty(name="Export Anim", description="Export all animations the Skinnedmesh uses into a .anim file", default=True)

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj is None:
            return False
        #end if
        armObj = utils.get_root_turok_object2(obj, "SKINNEDMESH", "ARMATURE")
        return armObj.type == "ARMATURE" and armObj.t2.objType == "SKINNEDMESH"
    #end def
    
    def execute(self, context):
        obj = utils.get_root_turok_object2(context.object, "SKINNEDMESH", "ARMATURE")
        global_matrix = (Matrix.Scale(utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward,to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        do_export(context, obj, self.filepath, global_matrix, self.exportMaterials, self.copyTextures, self.copyTextureSets, self.useCharDir, self.exportAnim)
        
        return {'FINISHED'}
    #end def
        
    def invoke(self, context, event):
        obj = utils.get_root_turok_object2(context.object, "SKINNEDMESH", "ARMATURE")
        self.filepath = obj.name
        
        utils.mode_set('OBJECT')
        validateResults = validate(context, obj)
        if len(validateResults["errors"]) > 0 and len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Errors and Warnings - See System Console for full details", toConsole=False)
        elif len(validateResults["errors"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Error - See System Console for full details", toConsole=False)
        #end if
        if len(validateResults["errors"]) > 0:
            for message in validateResults["errors"]:
                print("[Export Skinnedmesh Error] %s" % (message))
            #end for
            return {'CANCELLED'}
        #end if
        if len(validateResults["warnings"]) > 0:
            for message in validateResults["warnings"]:
                print("[Export Skinnedmesh Warning] %s" % (message))
            #end for
        #end if

        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    #end def
        
    def draw(self, context):
        layout = self.layout
        #layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Skinnedmesh Options:", icon="MONKEY")
        box = layout.box()
        row = box.row()
        row.prop(self, "exportMaterials")
        row = box.row()
        row.prop(self, "copyTextures")
        if self.copyTextures:
            row = box.row()
            if self.copyTextureSets and len(kdefs.defs["texsets"]) == 0:
                row.alert = True
            #end if
            row.separator()
            row.prop(self, "copyTextureSets")
            if self.copyTextureSets and len(kdefs.defs["texsets"]) == 0:
                row = box.row()
                row.alignment = 'RIGHT'
                row.alert = True
                row.label(text="textureSetInfo def not loaded", icon="ERROR")
            #end if
        #end if
        row = box.row()
        row.prop(self, "useCharDir")
        row = box.row()
        row.prop(self, "exportAnim")
        # if self.exportAnim:
            # row = layout.row()
            # row.label(text="Animation Options:", icon="ANIM")
            # box = layout.box()
            # row = box.row()
            # row.prop(self, "animNLATracks")
        # #end if
    #end def
#end class

# -----------------------------------------------------------------------------
# Returns dictionary of skinnedmesh structure. Returns None if there was an error.
def load(filepath, global_matrix):
    model = {}
    with open(filepath, 'rb') as data:
        data.seek(0)
        model["version"] = file_utils.read32(data, False) #version (3)
        if model["version"] != 3:
            utils.show_error("[Import Skinnedmesh] %s has incorrect version number of %d. Version must be 3." % (filepath, model["version"]))
            return None
        #end if
        model["bbMin"] = file_utils.readVector(data) #bounding box min
        model["bbMax"] = file_utils.readVector(data) #bounding box max
        model["htStartBone"] = file_utils.read8(data, False) #head track start bone
        model["htEndBone"] = file_utils.read8(data, False) #head track end bone
        model["htFactor"] = file_utils.read8(data, False) #head track track factor
        boneCount = file_utils.read32(data, False) #number of bones
        
        #check if theres too many bones
        if boneCount > 255:
            utils.show_error("[Import Skinnedmesh] %s has more than 255 bones (%d)" % (filepath, boneCount))
            return None
        #end if
        model["bones"] = []
        for i in range(boneCount):
            bone = {
                "children": set(),
            }
            model["bones"].append(bone)
            childrenCount = file_utils.read32(data, False) #number of children bones
            #check if theres too many children
            if childrenCount > 255:
                utils.show_error("[Import Skinnedmesh] %s, Bone %s has more than 255 child bones (%d)" % (filepath, i, childrenCount))
                return None
            #end if
            for i in range(childrenCount):
                bone["children"].add(file_utils.read16(data, False))
            #end for
        #end for
        for i in range(boneCount):
            bone = model["bones"][i]
            bone["bbMin"] = file_utils.readVector(data) #bounding box min
            bone["bbMax"] = file_utils.readVector(data) #bounding box max
            bone["impactType"] = file_utils.read8(data, False)
            bone["symbolFlags"] = file_utils.read8(data, False)
            bone["parentBone"] = file_utils.read8(data, False)
            bone["damageScalar"] = file_utils.readFloat(data)
            bone["flags"] = file_utils.read32(data, False)
        #end for
        
        model["hotpointGroupCount"] = file_utils.read32(data, False) #ignored/unused in kex(default value as 1)
        hotpointCount = file_utils.read32(data, False) #number of hotpoints
        #check if theres too many hotpoints
        if hotpointCount > 255:
            utils.show_error("[Import Skinnedmesh] %s has more than 255 hot points (%d)" % (filepath, hotpointCount))
            return None
        #end if
        model["hotpoints"] = []
        for i in range(hotpointCount):
            hotpoint = {}
            model["hotpoints"].append(hotpoint)
            hotpoint["origin"] = file_utils.readVector(data)
            hotpoint["flags"] = file_utils.read32(data, False)
            hotpoint["bone"] = file_utils.read16(data, False)
            hotpoint["particle1"] = file_utils.read16(data, False)
            hotpoint["particle2"] = file_utils.read16(data, False)
        #end for
        
        model["layers"] = file_utils.read32(data, False) #ignored/unused in kex(default value as 1)
        surfaceCount = file_utils.read32(data, False) #number of surfaces
        #check if theres too many surfaces
        if surfaceCount > 255:
            utils.show_error("[Import Skinnedmesh] %s has more than 255 surfaces (%d)" % (filepath, surfaceCount))
            return None
        #end if
        model["surfaces"] = []
        model["triCount"] = 0
        for i in range(surfaceCount):
            surface = { "material": "", "tris": [] }
            model["surfaces"].append(surface)
            surface["material"] = file_utils.readString(data)
            tri = [0] * 3
            indiceCount = 0
            while True:
                tri[indiceCount] = file_utils.read16(data, False)
                if tri[indiceCount] == 0xFFFF:
                    break
                #end if
                if indiceCount == 2:
                    surface["tris"].append(tuple(tri))
                    indiceCount = 0
                else:
                    indiceCount += 1
                #end if
            #end while
            model["triCount"] += len(surface["tris"])
        #end for

        model["verts"] = {"loc": [], "uv": [], "normals": [], "bones": []}
        model["vertCount"] = 0
        #only read vertices if there was any surfaces
        if surfaceCount != 0:
            vertCount = file_utils.read32(data, False) #number of vertices
            #check if theres too many vertices
            if vertCount > 0xFFFF:
                utils.show_error("[Import Skinnedmesh] %s has more than 65535 vertices (%d)" % (filepath, vertCount))
                return None
            #end if
            vertKeys = {}
            usedVertCount = 0 #amount of used vertices
            vertRemap = {}
            for vertIndex in range(vertCount):
                loc = file_utils.readVector(data)
                uv = file_utils.readVector2(data)
                no = file_utils.readVector(data)
                bi = file_utils.read16(data, False)
                key = (loc[0], loc[1], loc[2], uv[0], uv[1], no[0], no[1], no[2], bi)
                if key in vertKeys:
                    vertRemap[vertIndex] = vertKeys[key]
                else:
                    vertRemap[vertIndex] = usedVertCount
                    vertKeys[key] = usedVertCount
                    model["verts"]["loc"].append(loc)
                    model["verts"]["uv"].append(uv)
                    model["verts"]["normals"].append(no)
                    model["verts"]["bones"].append(bi)
                    usedVertCount += 1
                #end if
            #end for
            model["vertCount"] = usedVertCount
            #remap tri indices
            for surface in model["surfaces"]:
                for triIndex, tri in enumerate(surface["tris"]):
                    surface["tris"][triIndex] = (vertRemap[tri[0]], vertRemap[tri[1]], vertRemap[tri[2]])
                #end for
            #end for
        #end if
    #end with
    
    return model
#end def

# -----------------------------------------------------------------------------
# Returns mesh object if successful else None
def do_import(context, filepath, global_matrix, hideBBHot, importAnim, animBakeYawOffsets):
    armObj = None
    meshObj = None
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        print("Importing Turok 2 Skinnedmesh %r ..." % filepath)

        fileName = os.path.split(filepath)[1]
        fileNameNoExt = os.path.splitext(fileName)[0]
        scene = bpy.context.scene
        
        model = load(filepath, global_matrix)
        if model is None:
            return None
        #end if
        
        mesh = bpy.data.meshes.new('mesh')
        
        surfacePaths = {surface["material"]:0 for surface in model["surfaces"]} #unique list of kmat paths
        kmatFilePaths = set() #unique list of absolute paths to the .kmat files
        kmats = {}
        for kmatPath in surfacePaths:
            if kmatPath in kdefs.defs["builtin_kmats"]:
                continue
            #end if
            kmatRelFilePath = os.path.split(kmatPath)[0]
            kmatAbsFilePath = paths.find_t2_file_path(kmatRelFilePath + ".kmat")
            if kmatAbsFilePath is not None and kmatAbsFilePath not in kmatFilePaths:
                kmatFilePaths.add(kmatAbsFilePath)
                kmats.update(kmaterial.load(kmatAbsFilePath))
            #end if
        #end for
        
        for kmatPath in surfacePaths:
            #if is a builtin kmaterial
            if kmatPath in kdefs.defs["builtin_kmats"]:
                kmat = kdefs.defs["builtin_kmats"][kmatPath]
                bmat = utils.get_or_create_material(kmatPath, kmat)
                kmaterial.to_material(kmat, bmat)
                bmatIndex = utils.find_material_index(mesh, bmat)
                if bmatIndex == -1:
                    bmatIndex = len(mesh.materials)
                    mesh.materials.append(bmat)
                #end if
                surfacePaths[kmatPath] = bmatIndex

                continue
            #end if
            
            kmatRelFilePath, kmatName = os.path.split(kmatPath)
            if kmatName not in kmats:
                continue
            #end if
            
            kmat = kmats[kmatName]
            bmat = utils.get_or_create_material(kmatName, kmat)
            kmatFilename = os.path.split(kmatRelFilePath)[1]
            if kmatFilename == ("mat_" + fileNameNoExt):
                bmat.t2.isMatPathCustom = False
            else:
                bmat.t2.isMatPathCustom = True
            #end if
            bmat.t2.matPath = kmatRelFilePath
            kmaterial.to_material(kmat, bmat)
            
            bmatIndex = utils.find_material_index(mesh, bmat)
            if bmatIndex == -1:
                bmatIndex = len(mesh.materials)
                mesh.materials.append(bmat)
            #end if
            
            surfacePaths[kmatPath] = bmatIndex
        #end if
        
        # create armature object
        armature = bpy.data.armatures.new(name='Armature')
        armObj = utils.create_turok2_object(fileNameNoExt, armature, "SKINNEDMESH")
        armObj.data.display_type = 'STICK'
        armObj.show_in_front = True
        
        # create mesh scene object and add armature modifier to it
        meshObj = utils.create_turok2_object("_%s_mesh" % (fileNameNoExt), mesh, "SKINNEDMESH")
        meshObj.parent = armObj
        meshObj.parent_type = 'OBJECT'
        modArm = meshObj.modifiers.new("Armature", 'ARMATURE')
        modArm.show_in_editmode = True
        modArm.show_on_cage = True
        modArm.object = armObj

        #Create bone names
        boneNames = {} #boneindex, name
        for i in range(len(model["bones"])):
            boneNames[i] = utils.get_bonename_symbols(model["bones"][i]["symbolFlags"], i, True)
        #end for

        # create armature bones
        bpy.context.view_layer.objects.active = armObj
        if len(model["bones"]) > 0:
            utils.mode_set('EDIT')
            
            #Create edit bones
            for i in range(len(model["bones"])):
                bone = armature.edit_bones.new(boneNames[i])
                bone.head = (0.0, 0.0, 0.0)
                bone.tail = (0.0, utils.TUROK2_BONE_LENGTH, 0.0)
            #end for

            #Parent edit bones
            for i in range(len(model["bones"])):
                for childBoneIndex in model["bones"][i]["children"]:
                    if childBoneIndex == i:
                        continue
                    #end if
                    childBoneName = boneNames[childBoneIndex]
                    parentBoneName = boneNames[i]
                    armature.edit_bones[childBoneName].parent = armature.edit_bones[parentBoneName]
                #end for
            #end for

            #Exit Armature editing so the edit bone is created internally at this point
            utils.mode_set('OBJECT')
            
            #setup pose bones properties
            for poseBone in armObj.pose.bones:
                poseBone.lock_scale = (True, True, True)
            #end for
            
            #Set Turok 2 Bone Properties
            for i in range(len(model["bones"])):
                armature.bones[i].t2.boneIndex = i
                armature.bones[i].t2.impactType = model["bones"][i]["impactType"]
                utils.set_bone_symbols(armature.bones[i], model["bones"][i]["symbolFlags"])
                armature.bones[i].t2.damageScalar = model["bones"][i]["damageScalar"]
                armature.bones[i].t2.flags = utils.get_enumflags_id(panel_bone.EnumBoneFlags, ctypes.c_int(model["bones"][i]["flags"]).value)
            #end for

            #Create Bones Bounding Box
            for i in range(len(model["bones"])):
                bbName = "Bounding Box"
                boneBBObj = utils.create_bone_bounding_box(bbName, armObj, i, model["bones"][i]["bbMin"], model["bones"][i]["bbMax"])
                if hideBBHot:
                    boneBBObj.hide_set(state=True, view_layer=bpy.context.view_layer)
                #end if
            #end for
        #end if

        if model["triCount"] > 0 and model["vertCount"] > 2:
            blMesh = bmesh.new()
            blMesh.from_mesh(mesh)
            bVerts = []
            for vertIndex, vertLoc in enumerate(model["verts"]["loc"]):
                bVert = blMesh.verts.new(vertLoc)
                bVert.index = vertIndex
                bVert.normal = model["verts"]["normals"][vertIndex]
                bVerts.append(bVert)
            #end for
            
            for surface in model["surfaces"]:
                kmatRelPath = surface["material"]
                surf_mat_index = surfacePaths[kmatRelPath] if kmatRelPath in surfacePaths else 0
                for tri in surface["tris"]:
                    faceVerts = (bVerts[tri[0]], bVerts[tri[1]], bVerts[tri[2]])
                    try:
                        existingFace = blMesh.faces.get(faceVerts)
                    except ValueError: #Tried to get triangle that uses the same vert multiple times
                        continue
                    #end try

                    if existingFace:
                        existingTri = (existingFace.loops[0].vert.index, existingFace.loops[1].vert.index, existingFace.loops[2].vert.index)
                        windingOrderDifferent = existingTri != tri
                        #Only copy the face if the material is different or the tri indice order is different
                        if existingFace.material_index != surf_mat_index or windingOrderDifferent:
                            bFace = existingFace.copy(verts=False, edges=False)
                            if windingOrderDifferent:
                                bFace.normal_flip()
                            #end if
                        else: #exactly the same so skip it
                            continue
                        #end if
                    else:
                        bFace = blMesh.faces.new(faceVerts)
                    #end
                    bFace.smooth = True
                    bFace.material_index = surf_mat_index
                #end for
            #end for
            blMesh.to_mesh(mesh)
            blMesh.free()

            # mesh.validate(clean_customdata=False)
            
            # create uv layer and set uv values
            uv_layer = mesh.uv_layers.new()
            for loop in mesh.loops:
                uv_layer.data[loop.index].uv = model["verts"]["uv"][loop.vertex_index]
            #end for
        #end if
        
        #Setup vertex groups
        vertGroups = {} #key is bone index, value is vert group (with it's name as the bone)
        for i in range(model["vertCount"]):
            boneIndex = model["verts"]["bones"][i]
            if boneIndex not in vertGroups:
                vertGroups[boneIndex] = {"group":meshObj.vertex_groups.new(name=boneNames[boneIndex]), "verts":[]}
            #end if
            vertGroups[boneIndex]["verts"].append(i)
        #end for
        for key in vertGroups:
            vertGroups[key]["group"].add(vertGroups[key]["verts"], 1.0, 'ADD')
        #end for

        #set turok 2 properties for armature object
        bbObj = utils.create_bounding_box(meshObj, model["bbMin"], model["bbMax"])
        if hideBBHot:
            bbObj.hide_set(state=True, view_layer=bpy.context.view_layer)
        #end if
        
        #Set Skinnedmesh properties
        armObj.t2.headTrackStart = model["htStartBone"]
        armObj.t2.headTrackEnd = model["htEndBone"]
        armObj.t2.headTrackFactor = model["htFactor"]
        armObj.t2.hotpointGroupCount = model["hotpointGroupCount"]
        armObj.t2.skinnedLayers = model["layers"]
        
        #Create Hotpoints and parent to bones with the index
        for i in range(len(model["hotpoints"])):
            hpObject = utils.create_bone_hotpoint(armObj, model["hotpoints"][i]["bone"])
            hpObject.location = model["hotpoints"][i]["origin"]
            if hideBBHot:
                hpObject.hide_set(state=True, view_layer=bpy.context.view_layer)
            #end if
            lowestFlagBit = utils.get_lowest_setbit(ctypes.c_int(model["hotpoints"][i]["flags"]).value)
            flagsEnumID = utils.get_enum_id(panel_object.EnumHotpointFlags, lowestFlagBit)
            if flagsEnumID:
                hpObject.t2.flags = flagsEnumID
            #end if
            hpObject.t2.particle1 = model["hotpoints"][i]["particle1"]
            hpObject.t2.particle2 = model["hotpoints"][i]["particle2"]
        #end for
        
        #select only the armature
        for ob in bpy.context.selected_objects:
            ob.select_set(False)
        #end for
        armObj.select_set(True)
        
        progress.leave_substeps("Finished! (verts:%i tris:%i surfaces:%i materials:%i bones:%i hotpoints:%i)" % (model["vertCount"], model["triCount"], len(model["surfaces"]), len(mesh.materials), len(model["bones"]), len(model["hotpoints"])))
    #end with
    
    if importAnim and armObj:
        charFileDir, charFileName = os.path.split(os.path.normpath(filepath))
        charFileNameNoExt = os.path.splitext(charFileName)[0]
        parentCharDir, charSubDir = paths.get_path_from_dir(charFileDir, "char")
        animFileNames = []
        #If name of file ends with _<num> then check for the animation without that suffix first
        matchSuffix = re.search("_([0-9]+)$", charFileNameNoExt)
        if matchSuffix:
            animFileNames.append(charFileNameNoExt[:matchSuffix.span()[0]])
        #end if
        animFileNames.append(charFileNameNoExt)
        for animFileName in animFileNames:
            animPath = paths.find_t2_file_path(os.path.join("anims", charSubDir, animFileName + ".anim"))
            if animPath is not None:
                animation.do_import(context, armObj, animPath, global_matrix, animBakeYawOffsets)
                break
            #end if
        #end for
    #end if
    
    return armObj, meshObj
#end def

# -----------------------------------------------------------------------------
# 
def do_export(context, armObj, filepath, global_matrix, exportMaterials, copyTextures, copyTextureSets, useCharDir, exportAnim):
    armature = armObj.data
    meshObj = utils.get_skinnedmesh_mesh(armObj)
    mesh = meshObj.data
    meshBBObj = utils.get_skinnedmesh_boundingbox(meshObj)
    with ProgressReport(context.window_manager) as progress:
        stepCount = 1
        if exportMaterials: stepCount += 1
        if copyTextures: stepCount += 1
        progress.enter_substeps(stepCount)
        print("Exporting Turok 2 Skinnedmesh %r ..." % filepath)

        fileDir, fileName = os.path.split(filepath)
        fileNameNoExt = os.path.splitext(fileName)[0]
        matTexAnimDir = fileDir
        fileSubDir = ""
        matSubDir = "materials"
        if useCharDir:
            matTexAnimDir, fileSubDir = paths.get_path_from_dir(fileDir, "char")
            #if found char directory
            if matTexAnimDir != fileDir and fileSubDir != "":
                matSubDir = "materials/%s" % paths.GetLocalKexPathFromDir(fileSubDir)
            #end if
        #end if

        scene = bpy.context.scene
        with open(filepath, 'wb') as data:
            data.seek(0)
            file_utils.write32(data, 3, False) #version (3)
            
            meshBBMin = Vector((0.0,0.0,0.0))
            meshBBMax = Vector((0.0,0.0,0.0))
            if meshBBObj:
                meshBBMin, meshBBMax = utils.get_boundingbox_minmax(meshBBObj)
            #end if
            file_utils.writeVector(data, meshBBMin) #bounding box min
            file_utils.writeVector(data, meshBBMax) #bounding box max
            file_utils.write8(data, armObj.t2.headTrackStart, False) #head track start bone
            file_utils.write8(data, armObj.t2.headTrackEnd, False) #head track end bone
            file_utils.write8(data, armObj.t2.headTrackFactor, False) #head track Factor
            boneCount = len(armature.bones)
            file_utils.write32(data, boneCount, False) #number of bones
            boneVertGroups = {vertGroup.index:vertGroup for bone in armature.bones for vertGroup in meshObj.vertex_groups if vertGroup.name == bone.name}
            #get list of armature bones and sort them by Turok boneIndex
            bones = [bone for bone in armature.bones]
            bones.sort(key=lambda bone: bone.t2.boneIndex)
            for bone in bones:
                boneChildrenIndexs = [b.t2.boneIndex for b in bone.children]
                boneChildrenIndexs.sort()
                
                file_utils.write32(data, len(boneChildrenIndexs), False) #number of child bones
                for boneChildIndex in boneChildrenIndexs:
                    file_utils.write16(data, boneChildIndex, False)
                #end for
            #end for

            for bone in bones:
                boneBBObj = utils.get_bone_boundingbox(armObj, bone.name)
                boneBBMin = Vector((0.0,0.0,0.0))
                boneBBMax = Vector((0.0,0.0,0.0))
                if boneBBObj:
                    boneBBMin, boneBBMax = utils.get_boundingbox_minmax(boneBBObj)
                #end if
                file_utils.writeVector(data, boneBBMin) #bounding box min
                file_utils.writeVector(data, boneBBMax) #bounding box max
                file_utils.write8(data, bone.t2.impactType, False) #impact type
                file_utils.write8(data, utils.get_bone_symbols(bone), False) #symbol flags
                parentBoneIndex = 255 if bone.parent is None else bone.parent.t2.boneIndex
                file_utils.write8(data, parentBoneIndex, False) #parent bone index
                file_utils.writeFloat(data, bone.t2.damageScalar) #damage scalar
                boneFlags = utils.get_enumflags_value(panel_bone.EnumBoneFlags, bone.t2.flags)
                file_utils.write32(data, boneFlags, True) #flags
            #end for

            hotPointObjs = [o for o in bpy.data.objects if o.type == 'EMPTY' and o.t2.objType == "HOT POINT" and utils.obj_child_of(armObj, o)]
            file_utils.write32(data, armObj.t2.hotpointGroupCount, False) #hotpointGroupCount (unused)
            file_utils.write32(data, len(hotPointObjs), False) #number of hotpoints
            for hpObj in hotPointObjs:
                flagsIntValue = utils.get_enum_type(panel_object.EnumHotpointFlags, hpObj.t2.flags)
                file_utils.writeVector(data, hpObj.location) #origin
                file_utils.write32(data, flagsIntValue, True) #flags
                file_utils.write16(data, armature.bones[hpObj.parent_bone].t2.boneIndex, False) #bone index
                file_utils.write16(data, hpObj.t2.particle1, False) #particle 1
                file_utils.write16(data, hpObj.t2.particle2, False) #particle 2
            #end for
            
            materials = utils.obj_used_materials(meshObj)
            file_utils.write32(data, armObj.t2.skinnedLayers, False) #number of layers (unused)
            file_utils.write32(data, len(materials), False) #surface count
            verts = [] #for the mesh
            uv_layer = mesh.uv_layers.active.data
            indiceDict = {} #key is tuple of vert data, value is vert index
            kmatPaths = {} #key is local path to .kmat (with the extension), value is dictionary of mats (k: mat name, v: kmaterial dictionary)
            # locMatrix = Matrix(meshObj.matrix_world)
            # locMatrix.translation = Vector((0.0, 0.0, 0.0))
            # locMatrix = global_matrix @ locMatrix
            # normalQuat = (global_matrix @ meshObj.matrix_world).to_quaternion()

            #write surface
            for bmat in materials:
                matIndex = utils.find_material_index(mesh, bmat)
                kMatPath = ""
                matName = utils.get_real_material_name(bmat.name)[0]
                
                #Remove .000 from end of matname
                matchSuffix = re.search("\.([0-9]+)$", matName)
                if matchSuffix:
                    matName = matName[:matchSuffix.span()[0]]
                #end if
                
                #get surface path
                kMatLocalFilePath = ''
                if bmat.t2.isMatPathCustom:
                    kMatLocalFilePath = bmat.t2.matPath
                else:
                    kMatLocalFilePath = "%s/mat_%s" % (matSubDir, fileNameNoExt)
                #end if
                kMatPath = "%s/%s" % (kMatLocalFilePath, matName)
                file_utils.writeString(data, kMatPath) #surface name (e.g. materials/mat_Eye_Of_Truth/surf_587_00_01)
                
                #store the material paths and kmaterials used (for material and texture copying)
                kMatFullLocalFilePath = kMatLocalFilePath + ".kmat"
                if kMatFullLocalFilePath not in kmatPaths:
                    kmatPaths[kMatFullLocalFilePath] = {}
                #end if
                kmatPaths[kMatFullLocalFilePath][matName] = kmaterial.bmat_to_kmat(bmat)
                
                #get the triangle indices, and verts that this material uses
                tris = [] #for this material
                for poly in mesh.polygons:
                    if poly.material_index == matIndex:
                        tri = [] #3 indices for the tri
                        for loopIndex in range(poly.loop_start, poly.loop_start + poly.loop_total):
                            meshVert = mesh.vertices[mesh.loops[loopIndex].vertex_index]
                            bone = [armature.bones[boneVertGroups[vg.group].name] for vg in meshVert.groups if vg.group in boneVertGroups][0]
                            vertKey = tuple([meshVert.co[0], meshVert.co[1], meshVert.co[2], uv_layer[loopIndex].uv[0], uv_layer[loopIndex].uv[1], meshVert.normal[0], meshVert.normal[1], meshVert.normal[2], bone.t2.boneIndex])
                            if vertKey in indiceDict:
                                tri.append(indiceDict[vertKey])
                            else:
                                tri.append(len(verts))
                                indiceDict[vertKey] = len(verts)
                                verts.append((meshVert.co, uv_layer[loopIndex].uv, meshVert.normal, bone))
                            #end if
                        #end for
                        tris.append(tri)
                    #end if
                #end if
                
                for tri in tris:
                    for indice in tri:
                        file_utils.write16(data, indice, False)
                    #end for
                #end for
                file_utils.write16(data, 0xFFFF, False) #end of triangle indices
            #end for
            
            if len(materials) > 0:
                file_utils.write32(data, len(verts), False) #vertice count
                for vert in verts:
                    bone = vert[3]
                    vertLoc = Vector(vert[0]) - bone.matrix_local.translation
                    file_utils.writeVector(data, vertLoc) #xyz
                    file_utils.writeVector2(data, vert[1]) #uv
                    file_utils.writeVector(data, Vector(vert[2])) #normals
                    file_utils.write16(data, bone.t2.boneIndex, False) #bone index
                #end for
            #end if
        #end with
        
        progress.step("Skinnedmesh Successfully Exported!")

        if exportMaterials:
            #read existing files and put all those materials that aren't already in the dictionary in it
            for kmatPath, kmats in kmatPaths.items():
                kmatAbsPath = paths.find_t2_file_path(kmatPath)
                if kmatAbsPath is None:
                    continue
                #end if
                loadedKMaterials = kmaterial.load(kmatAbsPath)
                loadedKMaterials.update(kmats)
                kmats = loadedKMaterials #now contains both
            #end for

            #write kmat files to directory relative to export filepath
            matCount = 0
            matFileCount = 0
            for kmatPath, kMaterials in kmatPaths.items(): 
                matFileCount += 1
                matCount += len(kMaterials)
                kmatFilepath = os.path.normpath(os.path.join(matTexAnimDir, kmatPath))
                kmatDir = os.path.split(kmatFilepath)[0]
                if not os.path.isdir(kmatDir):
                    os.makedirs(kmatDir)
                #end if
                kmaterial.save(kmatFilepath, kMaterials)
            #end for
            progress.step("%i .kmat file(s) Exported (Materials: %i)" % (matFileCount, matCount))
        #end if
        
        if copyTextures:
            #get all used texture images
            images = {} #relative tex path, texture image
            texSetsToCopy = [] #array of string (tex set key)
            for mat in materials: #go through all materials this mesh uses
                if copyTextureSets:
                    #get first material texture set index
                    firstMat = utils.get_material_texset_index(mat, 0, False)

                    #go through nodes of first mat and add them to texSetsToCopy
                    for node in firstMat.node_tree.nodes:
                        if node.bl_idname == 'ShaderNodeTexImage':
                            #save image textures using the filepath if not using a texture set
                            if node.image is not None:
                                localTexPath = paths.texture_filepath_to_local_path(node.image.filepath)
                                if localTexPath not in utils.TUROK2_BUILTIN_TEXTURES:
                                    texSet = kdefs.defs["texsets"].get(localTexPath)
                                    if texSet is not None and localTexPath not in texSetsToCopy:
                                        texSetsToCopy.append(localTexPath)
                                    #end if
                                #end if
                            #end if
                        #end if
                    #end for
                #end if
                
                for node in mat.node_tree.nodes:
                    if node.bl_idname == 'ShaderNodeTexImage':
                        #save image textures using the filepath if not using a texture set
                        if node.image is not None:
                            localTexPath = paths.texture_filepath_to_local_path(node.image.filepath)
                            if localTexPath not in utils.TUROK2_BUILTIN_TEXTURES:
                                images[localTexPath] = node.image
                            #end if
                        #end if
                    #end if
                #end for
            #end for

            #save used images
            for localTexPath, image in images.items():
                texFilepath = os.path.join(matTexAnimDir, localTexPath)
                texDir = os.path.split(texFilepath)[0]
                if not os.path.isdir(texDir):
                    os.makedirs(texDir)
                #end if
                image.save_render(texFilepath)
            #end for
            
            #copy texture set local path texture files
            texsCopied = 0
            for texSetKey in texSetsToCopy:
                texSet = kdefs.defs["texsets"].get(texSetKey)
                for texLocalPath in texSet["textures"]:
                    #don't copy textures that were already saved with images
                    if texLocalPath in images:
                        continue
                    #end if
                    texSourcePath = paths.find_t2_file_path(texLocalPath)
                    if texSourcePath is None:
                        continue
                    #end if
                    texDestPath = os.path.join(matTexAnimDir, texLocalPath)
                    try:
                        os.makedirs(os.path.dirname(texDestPath), exist_ok=True)
                        copyfile(texSourcePath, texDestPath)
                        texsCopied += 1
                    except:
                        import sys
                        e = sys.exc_info()
                        print("[Turok 2 EX Tools] Failed to copy texture from \"%s\" to \"%s\"" % (texSourcePath, texDestPath))
                        print("%s %s" % (e[0], e[1]))
                    #end try except
                #end if
            #end for
            
            progress.step("%i Textures Exported" % (len(images) + texsCopied))
        #end if
        
        if exportAnim:
            animDir = fileDir
            if useCharDir and matTexAnimDir != fileDir:
                animSubDir = fileSubDir
                if animSubDir == "":
                    animSubDir = "anims"
                else:
                    animSubDir = os.path.join("anims", animSubDir)
                #end if
                animDir = os.path.join(matTexAnimDir, animSubDir)
                if not os.path.isdir(animDir):
                    os.makedirs(animDir)
                #end if
            #end if
            
            animation.do_export(context, os.path.join(animDir, fileNameNoExt + ".anim"), armObj, global_matrix)
        #end if
        
    #end with
        
    return {'FINISHED'}
#end def

# -----------------------------------------------------------------------------
# Return dictionary also validates animation actions.
# "errors"      List of string
# "warnings"    List of string
# "info"        List of string
# "actions"     List of valid animation actions
def validate(context, armObj):
    result = { "errors": [], "warnings": [], "info": [], "actions": []}
    
    #Error: Check if object is not None
    if armObj is None:
        result["errors"].append("Object is None")
        return result
    #end if
    
    #Error: Check if armature object has a mesh object
    meshObj = utils.get_skinnedmesh_mesh(armObj)
    if meshObj is None:
        result["errors"].append("Armature has no mesh object")
        return result
    #end if
    
    #Error: Check if material names are valid
    for mat in meshObj.data.materials:
        if mat is None:
            continue
        #end if
        matName = utils.get_real_material_name(mat.name)[0]
        invalid_chars = utils.string_contains_any(matName, "\\/\"{}#$ ")
        if invalid_chars:
            result["errors"].append("Object \"%s\", Material \"%s\", contains invalid characters. Rename the material. (Remove all spaces) The following invalid characters were found: %s" % (meshObj.name, mat.name, invalid_chars))
        #end if
        for param in mat.t2.params:
            if not param.name:
                result["errors"].append("Object \"%s\", Material \"%s\". has a param with no name." % (meshObj.name, mat.name))
            #end if
        #end for
    #end for
    
    #Error: Make sure each mesh has only 3 vertices per poly.
    isMeshTriangulated = True
    for poly in meshObj.data.polygons:
        if poly.loop_total != 3:
            isMeshTriangulated = False
            result["errors"].append("Object \"%s\" has %d verts per poly. Each poly must only have 3 verts. Triangulate the mesh." % (meshObj.name, poly.loop_total))
            break
        #end if
    #end for
    
    #Error: Check if the number of vertices used is more than 65535
    if isMeshTriangulated:
        indiceDict = {} #key is tuple of vert data, value is vert index
        materials = utils.obj_used_materials(meshObj)
        exceededVertLimit = False
        for bmat in materials:
            matIndex = utils.find_material_index(meshObj.data, bmat)
            uv_layer = meshObj.data.uv_layers.active.data
            for poly in meshObj.data.polygons:
                if poly.material_index == matIndex:
                    for loopIndex in range(poly.loop_start, poly.loop_start + poly.loop_total):
                        meshVert = meshObj.data.vertices[meshObj.data.loops[loopIndex].vertex_index]
                        vertKey = tuple([meshVert.co[0], meshVert.co[1], meshVert.co[2], uv_layer[loopIndex].uv[0], uv_layer[loopIndex].uv[1], meshVert.normal[0], meshVert.normal[1], meshVert.normal[2]])
                        if vertKey not in indiceDict:
                            vertIndex = len(indiceDict)
                            indiceDict[vertKey] = vertIndex
                            if vertIndex >= 0xFFFF:
                                exceededVertLimit = True
                                result["errors"].append("Mesh \"%s\", is using over 65535 vertices. Turok 2 only supports up to 65535 vertices for Skinnedmeshes." % (meshObj.name))
                                break
                            #end if
                        #end if
                    #end for
                    if exceededVertLimit:
                        break
                    #end if
                #end if
            #end for
            if exceededVertLimit:
                break
            #end if
        #end for
    #end if
    
    #Error: Too many bones
    if len(armObj.data.bones) >= 255:
        result["errors"].append("This armature is using %i Bones. Turok 2 only supports 255 Bones" % (len(armObj.data.bones)))
    #end if
    
    #Error: Check boneIndex is being used in another bone
    usedBoneIndexes = {}
    boneIndexInUse = False
    rootBones = []
    for bone in armObj.data.bones:
        if bone.parent is None:
            rootBones.append(bone)
        #end if
        if bone.t2.boneIndex in usedBoneIndexes:
            boneIndexInUse = True
            result["errors"].append("Bone \"%s\" is using the same bone index of %i as Bone \"%s\"" % (bone.name, bone.t2.boneIndex, usedBoneIndexes[bone.t2.boneIndex].name))
        else:
            usedBoneIndexes[bone.t2.boneIndex] = bone
        #end if
        if bone.t2.boneIndex >= len(armObj.data.bones):
            result["errors"].append("Bone \"%s\" uses bone index %i. Bone Index can only be from 0 to %i (the number of bones)" % (bone.name, bone.t2.boneIndex, len(armObj.data.bones)-1))
        #end if
    #end for
    
    #Error: Make sure there's only 1 bone with no parent bone
    if len(rootBones) > 1:
        result["errors"].append("You have %i Bones with no parents. Only 1 Bone can have no parent." % (len(rootBones)))
    else: #Blender doesn't allow no bones with a parent so no need to check for that.
        #Error: Make sure that the root bone has a bone index of 0
        if rootBones[0].t2.boneIndex != 0:
            result["errors"].append("Root Bone \"%s\" is using a Bone Index of %i. The Root Bone must be using a Bone index of 0." % (rootBones[0].name, rootBones[0].t2.boneIndex))
        #end if
    #end if
    
    #Error: Make sure each vertice is in only 1 bone vertex group and it's weight is 1.0
    boneVertGroups = {vertGroup.index for bone in armObj.data.bones for vertGroup in meshObj.vertex_groups if vertGroup.name == bone.name}
    for vertIndex, vert in enumerate(meshObj.data.vertices):
        vertGroupsIn = [vg for vg in vert.groups if vg.group in boneVertGroups]
        if len(vertGroupsIn) == 0:
            result["errors"].append("Mesh \"%s\", Vert %d, is not in any Bone Vertex Groups." % (meshObj.name, vert.index))
            break
        elif len(vertGroupsIn) > 1:
            result["errors"].append("Mesh \"%s\", Vert %d, is in more than 1 Bone Vertex Groups. Turok 2 Only Supports 1 Bone per vertice." % (meshObj.name, vert.index))
            break
        elif vertGroupsIn[0].weight < 1.0:
            result["errors"].append("Mesh \"%s\", Vert %d, has a weight of %s in vertex group \"%s\". It must have a weight of 1.0." % (meshObj.name, vert.index, vertGroupsIn[0].weight, meshObj.vertex_groups[vertGroupsIn[0].group].name))
            break
        #end if
    #end for
    
    #Error: Check if Bone is facing forward
    for bone in armObj.data.bones:
        if bone.vector.normalized() != Vector((0.0, 1.0, 0.0)):
        #if bone.matrix != Matrix.Identity(3):
            result["errors"].append("Bone \"%s\", must be facing forward. Press the Setup Turok 2 Bone button in the Bone Panel to set up the Bone correctly." % (bone.name))
        #end if
    #end for
    
    lastMode = context.object.mode
    utils.mode_set('EDIT')
    #Error: Check if Bone has no roll
    for bone in armObj.data.edit_bones:
        if bone.roll != 0.0:
            result["errors"].append("Edit Bone \"%s\" has a roll value of %s. Roll must be set to 0. Press the Setup Turok 2 Bone button in the Bone Panel to set up the Bone correctly." % (bone.name, bone.roll))
        #end if
    #end for
    utils.mode_set(lastMode)
    
    #Error: Check if Head Track Start and End Bones are set to a bone that exists
    if armObj.t2.headTrackStart != 255:
        boneName = utils.find_bone_name(armObj, armObj.t2.headTrackStart)
        if not boneName:
            result["errors"].append("Skinnedmesh \"%s\", has a Head Track Start bone of %i which does not exist in the armature" % (armObj.name, armObj.t2.headTrackStart))
        #end if
    #end if
    if armObj.t2.headTrackEnd != 255:
        boneName = utils.find_bone_name(armObj, armObj.t2.headTrackEnd)
        if not boneName:
            result["errors"].append("Skinnedmesh \"%s\", has a Head Track End bone of %i which does not exist in the armature" % (armObj.name, armObj.t2.headTrackEnd))
        #end if
    #end if
    
    #Get all children of armature
    childArmObjs = [o for o in bpy.data.objects if o.parent == armObj]
    
    #Error: Check for None turok object types in child objects
    for childObj in childArmObjs:
        if childObj.t2.objType == "NONE":
            result["errors"].append("A child object \"%s\" has a Turok Object of None." % (armObj.name, childObj.name))
        #end if
    #end for
    
    #Get all children of armature mesh that are bounding boxes
    childMeshBBObjs = [o for o in bpy.data.objects if o.parent == meshObj and o.t2.objType == "BOUNDING BOX"]
    
    #Warning: Check if there's more than 1 bounding box parented to the armatures mesh
    if len(childMeshBBObjs) > 1:
        result["warnings"].append("There is more than 1 Bounding Box parented to the armatures Mesh \"%s\"" % (meshObj.name))
    #end if
    
    #Error: Check if the bounding box parented to the armatures mesh is a Mesh object type
    for childMeshBBObj in childMeshBBObjs:
        if childMeshBBObj.type != "MESH":
            result["errors"].append("\"%s\" Bounding Box for Mesh \"%s\" must be a Mesh object type." % (childMeshBBObj.name, meshObj.name))
        #end if
    #end for
    
    #bounding boxes must be mesh type and parented to a bone that exists
    boneBBObjs = [o for o in childArmObjs if o.t2.objType == "BOUNDING BOX"]
    for boneBBObj in boneBBObjs:
        if boneBBObj.type != "MESH":
            result["errors"].append("\"%s\" Bounding Box for bone is not a Mesh. Bounding Boxes must be a Mesh object type." % (boneBBObj.name))
        #end if
        if boneBBObj.parent_type != "BONE":
            result["errors"].append("\"%s\" Bounding Box must be parented to a bone on the armature." % (boneBBObj.name))
        elif boneBBObj.parent_bone not in armObj.data.bones:
            result["errors"].append("\"%s\" Bounding Box must be parented to a valid bone name." % (boneBBObj.name))
        #end if
    #end for
    
    #make sure hotpoints are empty type and are parented to a bone that exists
    boneHPObjs = [o for o in childArmObjs if o.t2.objType == "HOT POINT"]
    for boneHPObj in boneHPObjs:
        if boneHPObj.type != "EMPTY":
            result["errors"].append("\"%s\" Hot Point must be a Empty object type." % (boneHPObj.name))
        #end if
        if boneHPObj.parent_type != "BONE":
            result["errors"].append("\"%s\" Hot Point must be parented to a bone on the armature." % (boneHPObj.name))
        elif boneHPObj.parent_bone not in armObj.data.bones:
            result["errors"].append("\"%s\" Hot Point must be parented to a valid bone name." % (boneHPObj.name))
        #end if
    #end for
    
    animResults = animation.validate(context, armObj)
    result["errors"].extend(animResults["errors"])
    result["warnings"].extend(animResults["warnings"])
    result["info"].extend(animResults["info"])
    result["actions"].extend(animResults["actions"])

    return result
#end def

# -----------------------------------------------------------------------------
#
def menu_func_import_skinnedmesh(self, context):
    self.layout.operator(TUROK2_OT_import_skinnedmesh.bl_idname, text="Turok 2 Skinnedmesh (.skinnedmesh)")
#end def

# -----------------------------------------------------------------------------
#
def menu_func_export_skinnedmesh(self, context):
    self.layout.operator(TUROK2_OT_export_skinnedmesh.bl_idname, text="Turok 2 Skinnedmesh (.skinnedmesh)")
#end def

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(TUROK2_OT_import_skinnedmesh)
    bpy.utils.register_class(TUROK2_OT_export_skinnedmesh)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_skinnedmesh)
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export_skinnedmesh)
#end def

# -----------------------------------------------------------------------------
#
def unregister():
    bpy.utils.unregister_class(TUROK2_OT_import_skinnedmesh)
    bpy.utils.unregister_class(TUROK2_OT_export_skinnedmesh)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_skinnedmesh)
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export_skinnedmesh)
#end def
