import bpy
import os
import bmesh
import re

from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from shutil import copyfile
from bpy_extras.image_utils import load_image
from bpy_extras.wm_utils.progress_report import (
    ProgressReport,
    ProgressReportSubstep,
)
from bpy.props import (
        CollectionProperty,
        BoolProperty,
        FloatProperty,
        StringProperty,
        EnumProperty,
        )
from bpy_extras.io_utils import (
        ImportHelper,
        ExportHelper,
        orientation_helper,
        axis_conversion,
        )
from . import (
    utils,
    file_utils,
    paths,
    kmaterial,
    kdefs,
)

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_import_staticmesh(bpy.types.Operator, ImportHelper):
    """Import a Turok 2 Staticmesh File"""
    bl_idname = "turok2_object.import_staticmesh"
    bl_label = "Import Staticmesh"
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

	# Properties used by the file browser
    filepath: StringProperty(name="File Path", description="File path used for importing the staticmesh file", default="", options={'HIDDEN'})
    files: CollectionProperty(type=bpy.types.OperatorFileListElement, options={'HIDDEN'})
    directory: StringProperty(subtype='FILE_PATH', options={'HIDDEN'})
    filter_folder: BoolProperty(name="Filter Folders", description="", default=True, options={'HIDDEN'})
    filename_ext = ".staticmesh"
    filter_glob: StringProperty(default="*.staticmesh", options={'HIDDEN'})

    def execute(self, context):
        global_matrix = (Matrix.Scale(1.0/utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward, to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        modelObj = do_import(context, self.filepath, global_matrix)
        if modelObj is None:
            return {'CANCELLED'}
        #end if

        return {'FINISHED'}
    #end def
    
    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Staticmesh Options:", icon="MONKEY")
        # box = layout.box()
        # row = box.row()
        # row.prop(self, "hideVariants")
    #end def
#end class

# -----------------------------------------------------------------------------
#
@orientation_helper(axis_forward='Y', axis_up='Z')
class TUROK2_OT_export_staticmesh(bpy.types.Operator, ExportHelper):
    """Export a Turok 2 Staticmesh File"""
    bl_idname = "turok2_object.export_staticmesh"
    bl_label = 'Export Staticmesh'
    bl_options = {'PRESET', 'REGISTER', 'UNDO'}

    filename_ext = ".staticmesh"
    filter_glob: StringProperty(default="*.staticmesh", options={'HIDDEN'})

    exportMaterials: BoolProperty(name="Export Materials", description="Export all used materials to .kmat files relative to the export directory", default=True)
    copyTextures: BoolProperty(name="Copy Textures", description="Copies all used textures to their path relative to the export directory", default=False)
    copyTextureSets: BoolProperty(name="Copy All Textures in Texture Sets", description="Copies all textures in every texture set this staticmesh uses", default=False)
    useModelsDir: BoolProperty(name="Use Models Directory", description="Exports materials and textures to the parent folder of \"models\" + the sub folders after the models folder if it exists in the export path.", default=True)

    @classmethod
    def poll(cls, context):
        return context.object is not None and context.object.type == "MESH"
    #end def
    
    def execute(self, context):
        global_matrix = (Matrix.Scale(utils.IMPORT_EXPORT_SCALE, 4) @ axis_conversion(to_forward=self.axis_forward,to_up=self.axis_up).to_4x4())
        utils.mode_set('OBJECT')
        do_export(context, self.filepath, global_matrix, self.exportMaterials, self.copyTextures, self.copyTextureSets, self.useModelsDir)
        
        return {'FINISHED'}
    #end def
        
    def invoke(self, context, event):
        obj = context.object
        self.filepath = obj.name
        
        utils.mode_set('OBJECT')
        validateResults = validate(context, obj)
        if len(validateResults["errors"]) > 0 and len(validateResults["warnings"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Errors and Warnings - See System Console for full details", toConsole=False)
        elif len(validateResults["errors"]) > 0:
            utils.show_error("%s" % (validateResults["errors"][0]), "Error - See System Console for full details", toConsole=False)
        #end if
        if len(validateResults["errors"]) > 0:
            for message in validateResults["errors"]:
                print("[Export Staticmesh Error] %s" % (message))
            #end for
            return {'CANCELLED'}
        #end if
        if len(validateResults["warnings"]) > 0:
            for message in validateResults["warnings"]:
                print("[Export Staticmesh Warning] %s" % (message))
            #end for
        #end if

        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    #end def
        
    def draw(self, context):
        layout = self.layout
        #layout.use_property_split = True
        
        row = layout.row()
        row.label(text="Staticmesh Options:", icon="MONKEY")
        box = layout.box()
        row = box.row()
        row.prop(self, "exportMaterials")
        row = box.row()
        row.prop(self, "copyTextures")
        if self.copyTextures:
            row = box.row()
            if self.copyTextureSets and len(kdefs.defs["texsets"]) == 0:
                row.alert = True
            #end if
            row.separator()
            row.prop(self, "copyTextureSets")
            if self.copyTextureSets and len(kdefs.defs["texsets"]) == 0:
                row = box.row()
                row.alignment = 'RIGHT'
                row.alert = True
                row.label(text="textureSetInfo def not loaded", icon="ERROR")
            #end if
        #end if
        row = box.row()
        row.prop(self, "useModelsDir")
        #end if
        
        # row = layout.row()
        # row = layout.row()
        # row.label(text="Misc Options:", icon="PREFERENCES")
        # box = layout.box()
        # row = box.row()
        # row.prop(self, "axis_forward")
        # row = box.row()
        # row.prop(self, "axis_up")
    #end def
#end class

# -----------------------------------------------------------------------------
# Returns dictionary of staticmesh structure. Returns None if there was an error.
def load(filepath, global_matrix):
    model = {}
    with open(filepath, 'rb') as data:
        data.seek(0)
        model["version"] = file_utils.read32(data, False) #version (2)
        if model["version"] != 2:
            utils.show_error("[Import Staticmesh] %s has incorrect version number of %d. Version must be 2." % (filepath, model["version"]))
            return None
        #end if
        model["bbMin"] = file_utils.readVector(data) #bounding box min
        model["bbMax"] = file_utils.readVector(data) #bounding box max
        model["surfaces"] = []
        model["tris"] = []
        model["verts"] = {"loc": [], "uv": [], "normals": []}
        surfaceCount = file_utils.read32(data, False) #number of surfaces
        #check if theres too many surfaces
        if surfaceCount > 255:
            utils.show_error("[Import Staticmesh] %s has more than 255 surfaces (%d)" % (filepath, surfaceCount))
            return None
        #end if
        for i in range(surfaceCount):
            surface = { "material": "", "tris": [] }
            model["surfaces"].append(surface)
            surface["material"] = file_utils.readString(data)
            triCount = file_utils.read32(data, False) #number of triangles
            #check if theres too many triangles
            if triCount > 1048575:
                utils.show_error("[Import Staticmesh] %s has more than 1048575 triangles (%d) on surface %d: %s" % (filepath, triCount, i, surface["material"]))
                return None
            #end if
            for j in range(triCount):
                v0 = file_utils.read16(data, False)
                v1 = file_utils.read16(data, False)
                v2 = file_utils.read16(data, False)
                tri = (v0, v1, v2)
                surface["tris"].append(tri)
                # print("surface: %s, tris %s: %s" % (i, j, tri))
                model["tris"].append(tri)
            #end for
        #end for
        
        vertCount = file_utils.read32(data, False) #number of vertices
        #check if theres too many vertices
        if vertCount > 1048575:
            utils.show_error("[Import Staticmesh] %s has more than 1048575 vertices (%d)" % (filepath, vertCount))
            return None
        #end if
        vertKeys = {}
        vertRemap = {}
        usedVertCount = 0
        for vertIndex in range(vertCount):
            loc = file_utils.readVector(data)
            uv = file_utils.readVector2(data)
            no = file_utils.readVector(data)
            key = (loc[0], loc[1], loc[2], uv[0], uv[1], no[0], no[1], no[2])
            if key in vertKeys:
                vertRemap[vertIndex] = vertKeys[key]
            else:
                vertRemap[vertIndex] = usedVertCount
                vertKeys[key] = usedVertCount
                model["verts"]["loc"].append(loc)
                model["verts"]["uv"].append(uv)
                model["verts"]["normals"].append(no)
                usedVertCount += 1
            #end if
        #end for
        model["vertCount"] = usedVertCount
        #remap tri indices
        for surface in model["surfaces"]:
            for triIndex, tri in enumerate(surface["tris"]):
                surface["tris"][triIndex] = (vertRemap[tri[0]], vertRemap[tri[1]], vertRemap[tri[2]])
            #end for
        #end for
    #end with
    
    return model
#end def

# -----------------------------------------------------------------------------
# Returns mesh object if successful else None
def do_import(context, filepath, global_matrix):
    obj = None
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        print("Importing Turok 2 Staticmesh %r ..." % filepath)

        fileName = os.path.split(filepath)[1]
        fileNameNoExt = os.path.splitext(fileName)[0]
        scene = bpy.context.scene
        # scene.cursor.location = (0.0, 0.0, 0.0) #need to have the cursor position at center
        
        model = load(filepath, global_matrix)
        if model is None:
            return None
        #end if
        
        mesh = bpy.data.meshes.new('mesh')
        
        surfacePaths = {surface["material"]:0 for surface in model["surfaces"]} #unique list of kmat paths
        kmatFilePaths = set() #unique list of absolute paths to the .kmat files
        kmats = {}
        for kmatPath in surfacePaths:
            if kmatPath in kdefs.defs["builtin_kmats"]:
                continue
            #end if
            kmatRelFilePath = os.path.split(kmatPath)[0]
            kmatAbsFilePath = paths.find_t2_file_path(kmatRelFilePath + ".kmat")
            if kmatAbsFilePath is not None and kmatAbsFilePath not in kmatFilePaths:
                kmatFilePaths.add(kmatAbsFilePath)
                kmats.update(kmaterial.load(kmatAbsFilePath))
            #end if
        #end for
        
        for kmatPath in surfacePaths:
            #if is a builtin kmaterial
            if kmatPath in kdefs.defs["builtin_kmats"]:
                kmat = kdefs.defs["builtin_kmats"][kmatPath]
                bmat = utils.get_or_create_material(kmatPath, kmat)
                kmaterial.to_material(kmat, bmat)
                bmatIndex = utils.find_material_index(mesh, bmat)
                if bmatIndex == -1:
                    bmatIndex = len(mesh.materials)
                    mesh.materials.append(bmat)
                #end if
                surfacePaths[kmatPath] = bmatIndex

                continue
            #end if
            
            kmatRelFilePath, kmatName = os.path.split(kmatPath)
            if kmatName not in kmats:
                continue
            #end if
            
            kmat = kmats[kmatName]
            bmat = utils.get_or_create_material(kmatName, kmat)
            kmatFilename = os.path.split(kmatRelFilePath)[1]
            if kmatFilename == ("mat_" + fileNameNoExt):
                bmat.t2.isMatPathCustom = False
            else:
                bmat.t2.isMatPathCustom = True
            #end if
            bmat.t2.matPath = kmatRelFilePath
            kmaterial.to_material(kmat, bmat)
            
            bmatIndex = utils.find_material_index(mesh, bmat)
            if bmatIndex == -1:
                bmatIndex = len(mesh.materials)
                mesh.materials.append(bmat)
            #end if
            
            surfacePaths[kmatPath] = bmatIndex
        #end if
        
        if len(model["tris"]) > 0 and model["vertCount"] > 2:
            blMesh = bmesh.new()
            blMesh.from_mesh(mesh)
            bVerts = []
            for vertIndex, vertLoc in enumerate(model["verts"]["loc"]):
                bVert = blMesh.verts.new(vertLoc)
                bVert.index = vertIndex
                bVert.normal = model["verts"]["normals"][vertIndex]
                bVerts.append(bVert)
            #end for
            
            for surface in model["surfaces"]:
                kmatRelPath = surface["material"]
                surf_mat_index = surfacePaths[kmatRelPath] if kmatRelPath in surfacePaths else 0
                for tri in surface["tris"]:
                    faceVerts = (bVerts[tri[0]], bVerts[tri[1]], bVerts[tri[2]])
                    try:
                        existingFace = blMesh.faces.get(faceVerts)
                    except ValueError: #Tried to get triangle that uses the same vert multiple times
                        continue
                    #end try
                    if existingFace:
                        existingTri = (existingFace.loops[0].vert.index, existingFace.loops[1].vert.index, existingFace.loops[2].vert.index)
                        windingOrderDifferent = existingTri != tri
                        #Only copy the face if the material is different or the tri indice order is different
                        if existingFace.material_index != surf_mat_index or windingOrderDifferent:
                            bFace = existingFace.copy(verts=False, edges=False)
                            if windingOrderDifferent:
                                bFace.normal_flip()
                            #end if
                        else: #exactly the same so skip it
                            continue
                        #end if
                    else:
                        bFace = blMesh.faces.new(faceVerts)
                    #end try
                    bFace.smooth = True
                    bFace.material_index = surf_mat_index
                #end for
            #end for
            blMesh.to_mesh(mesh)
            blMesh.free()

            # mesh.validate(clean_customdata=False)

            # create uv layer and set uv values
            uv_layer = mesh.uv_layers.new()
            for loop in mesh.loops:
                uv_layer.data[loop.index].uv = model["verts"]["uv"][loop.vertex_index]
            #end for
        #end if
        
        # create mesh scene object
        obj = utils.create_turok2_object(fileNameNoExt, mesh, "STATICMESH")
        for ob in bpy.context.selected_objects:
            ob.select_set(False)
        #end for
        bpy.context.view_layer.objects.active = obj
        obj.select_set(True)
        obj.name = fileNameNoExt

        progress.leave_substeps("Finished! (verts:%i tris:%i surfaces:%i materials:%i)" % (model["vertCount"], len(model["tris"]), len(model["surfaces"]), len(mesh.materials)))
    #end with
    
    return obj
#end def

# -----------------------------------------------------------------------------
# 
def do_export(context, filepath, global_matrix, exportMaterials, copyTextures, copyTextureSets, useModelsDir):
    obj = context.object
    mesh = obj.data
    with ProgressReport(context.window_manager) as progress:
        stepCount = 1
        if exportMaterials: stepCount += 1
        if copyTextures: stepCount += 1
        progress.enter_substeps(stepCount)
        print("Exporting Turok 2 Staticmesh %r ..." % filepath)

        fileDir, fileName = os.path.split(filepath)
        fileNameNoExt = os.path.splitext(fileName)[0]
        matTexAnimDir = fileDir
        fileSubDir = ""
        matSubDir = "materials"
        if useModelsDir:
            matTexAnimDir, fileSubDir = paths.get_path_from_dir(fileDir, "models")
            #if found models directory
            if matTexAnimDir != fileDir and fileSubDir != "":
                matSubDir = "materials/%s" % paths.GetLocalKexPathFromDir(fileSubDir)
            #end if
        #end if

        scene = bpy.context.scene
        with open(filepath, 'wb') as data:
            data.seek(0)
            file_utils.write32(data, 2, False) #version (2)
            file_utils.writeVector(data, Vector((0.0,0.0,0.0))) #bounding box min
            file_utils.writeVector(data, Vector((0.0,0.0,0.0))) #bounding box max
            materials = utils.obj_used_materials(obj)
            file_utils.write32(data, len(materials), False) #surface count
            verts = [] #for the mesh
            uv_layer = mesh.uv_layers.active.data
            indiceDict = {} #key is tuple of vert data, value is vert index
            kmatPaths = {} #key is local path to .kmat (with the extension), value is dictionary of mats (k: mat name, v: kmaterial dictionary)
            locMatrix = Matrix(obj.matrix_world)
            locMatrix.translation = Vector((0.0, 0.0, 0.0))
            locMatrix = global_matrix @ locMatrix
            normalQuat = (global_matrix @ obj.matrix_world).to_quaternion()

            #write surface
            for bmat in materials:
                matIndex = utils.find_material_index(mesh, bmat)
                kMatPath = ""
                matName = utils.get_real_material_name(bmat.name)[0]
                
                #Remove .000 from end of matname
                matchSuffix = re.search("\.([0-9]+)$", matName)
                if matchSuffix:
                    matName = matName[:matchSuffix.span()[0]]
                #end if
                
                #get surface path
                kMatLocalFilePath = ''
                if bmat.t2.isMatPathCustom:
                    kMatLocalFilePath = bmat.t2.matPath
                else:
                    kMatLocalFilePath = "%s/mat_%s" % (matSubDir, fileNameNoExt)
                #end if
                kMatPath = "%s/%s" % (kMatLocalFilePath, matName)
                file_utils.writeString(data, kMatPath) #surface name (e.g. materials/mat_Eye_Of_Truth/surf_587_00_01)
                
                #store the material paths and kmaterials used (for material and texture copying)
                kMatFullLocalFilePath = kMatLocalFilePath + ".kmat"
                if kMatFullLocalFilePath not in kmatPaths:
                    kmatPaths[kMatFullLocalFilePath] = {}
                #end if
                kmatPaths[kMatFullLocalFilePath][matName] = kmaterial.bmat_to_kmat(bmat)
                
                #get the triangle indices, and verts that this material uses
                tris = [] #for this material
                for poly in mesh.polygons:
                    if poly.material_index == matIndex:
                        tri = [] #3 indices for the tri
                        for loopIndex in range(poly.loop_start, poly.loop_start + poly.loop_total):
                            meshVert = mesh.vertices[mesh.loops[loopIndex].vertex_index]
                            vertKey = tuple([meshVert.co[0], meshVert.co[1], meshVert.co[2], uv_layer[loopIndex].uv[0], uv_layer[loopIndex].uv[1], meshVert.normal[0], meshVert.normal[1], meshVert.normal[2]])
                            if vertKey in indiceDict:
                                tri.append(indiceDict[vertKey])
                            else:
                                tri.append(len(verts))
                                indiceDict[vertKey] = len(verts)
                                verts.append((meshVert.co, uv_layer[loopIndex].uv, meshVert.normal))
                            #end if
                        #end for
                        tris.append(tri)
                    #end if
                #end if
                
                file_utils.write32(data, len(tris), False) #tri count
                for tri in tris:
                    for indice in tri:
                        file_utils.write16(data, indice, False)
                    #end for
                #end for
            #end for
            
            file_utils.write32(data, len(verts), False) #vertice count
            for vert in verts:
                file_utils.writeVector(data, locMatrix @ Vector(vert[0]))
                file_utils.writeVector2(data, vert[1])
                file_utils.writeVector(data, normalQuat @ Vector(vert[2]))
            #end for
        #end with
        
        progress.step("Staticmesh Successfully Exported!")

        if exportMaterials:
            #read existing files and put all those materials that aren't already in the dictionary in it
            for kmatPath, kmats in kmatPaths.items():
                kmatAbsPath = paths.find_t2_file_path(kmatPath)
                if kmatAbsPath is None:
                    continue
                #end if
                loadedKMaterials = kmaterial.load(kmatAbsPath)
                loadedKMaterials.update(kmats)
                kmats = loadedKMaterials #now contains both
            #end for

            #write kmat files to directory relative to export filepath
            matCount = 0
            matFileCount = 0
            for kmatPath, kMaterials in kmatPaths.items(): 
                matFileCount += 1
                matCount += len(kMaterials)
                kmatFilepath = os.path.normpath(os.path.join(matTexAnimDir, kmatPath))
                kmatDir = os.path.split(kmatFilepath)[0]
                if not os.path.isdir(kmatDir):
                    os.makedirs(kmatDir)
                #end if
                kmaterial.save(kmatFilepath, kMaterials)
            #end for
            progress.step("%i .kmat file(s) Exported (Materials: %i)" % (matFileCount, matCount))
        #end if
        
        if copyTextures:
            #get all used texture images
            images = {} #relative tex path, texture image
            texSetsToCopy = [] #array of string (tex set key)
            for mat in materials: #go through all materials this mesh uses
                if copyTextureSets:
                    #get first material texture set index
                    firstMat = utils.get_material_texset_index(mat, 0, False)

                    #go through nodes of first mat and add them to texSetsToCopy
                    for node in firstMat.node_tree.nodes:
                        if node.bl_idname == 'ShaderNodeTexImage':
                            #save image textures using the filepath if not using a texture set
                            if node.image is not None:
                                localTexPath = paths.texture_filepath_to_local_path(node.image.filepath)
                                if localTexPath not in utils.TUROK2_BUILTIN_TEXTURES:
                                    texSet = kdefs.defs["texsets"].get(localTexPath)
                                    if texSet is not None and localTexPath not in texSetsToCopy:
                                        texSetsToCopy.append(localTexPath)
                                    #end if
                                #end if
                            #end if
                        #end if
                    #end for
                #end if
                
                for node in mat.node_tree.nodes:
                    if node.bl_idname == 'ShaderNodeTexImage':
                        #save image textures using the filepath if not using a texture set
                        if node.image is not None:
                            localTexPath = paths.texture_filepath_to_local_path(node.image.filepath)
                            if localTexPath not in utils.TUROK2_BUILTIN_TEXTURES:
                                images[localTexPath] = node.image
                            #end if
                        #end if
                    #end if
                #end for
            #end for

            #save used images
            for localTexPath, image in images.items():
                texFilepath = os.path.join(matTexAnimDir, localTexPath)
                texDir = os.path.split(texFilepath)[0]
                if not os.path.isdir(texDir):
                    os.makedirs(texDir)
                #end if
                image.save_render(texFilepath)
            #end for
            
            #copy texture set local path texture files
            texsCopied = 0
            for texSetKey in texSetsToCopy:
                texSet = kdefs.defs["texsets"].get(texSetKey)
                for texLocalPath in texSet["textures"]:
                    #don't copy textures that were already saved with images
                    if texLocalPath in images:
                        continue
                    #end if
                    texSourcePath = paths.find_t2_file_path(texLocalPath)
                    if texSourcePath is None:
                        continue
                    #end if
                    texDestPath = os.path.join(matTexAnimDir, texLocalPath)
                    try:
                        os.makedirs(os.path.dirname(texDestPath), exist_ok=True)
                        copyfile(texSourcePath, texDestPath)
                        texsCopied += 1
                    except:
                        import sys
                        e = sys.exc_info()
                        print("[Turok 2 EX Tools] Failed to copy texture from \"%s\" to \"%s\"" % (texSourcePath, texDestPath))
                        print("%s %s" % (e[0], e[1]))
                    #end try except
                #end if
            #end for
            
            progress.step("%i Textures Exported" % (len(images) + texsCopied))
        #end if
    #end with
        
    return {'FINISHED'}
#end def

# -----------------------------------------------------------------------------
# Return dictionary, errors: array of string, warnings, info
def validate(context, obj):
    result = { "errors": [], "warnings": [], "info": [], "actions": []}
    
    #Error: Check if object is not None
    if obj is None:
        result["errors"].append("Object is None")
        return result
    #end if
    
    #Error: Check if object is not a mesh
    if obj.type != "MESH":
        result["errors"].append("Object is not a Mesh")
        return result
    #end if
    
    #Error: Check if material names are valid
    for mat in obj.data.materials:
        if mat is None:
            continue
        #end if
        matName = utils.get_real_material_name(mat.name)[0]
        invalid_chars = utils.string_contains_any(matName, "\\/\"{}#$ ")
        if invalid_chars:
            result["errors"].append("Object \"%s\", Material \"%s\", contains invalid characters. Rename the material. (Remove all spaces) The following invalid characters were found: %s" % (obj.name, mat.name, invalid_chars))
        #end if
        for param in mat.t2.params:
            if not param.name:
                result["errors"].append("Object \"%s\", Material \"%s\". has a param with no name." % (obj.name, mat.name))
            #end if
        #end for
    #end for
    
    #Error: Make sure each mesh has only 3 vertices per poly.
    for poly in obj.data.polygons:
        if poly.loop_total != 3:
            result["errors"].append("Object \"%s\" has %d verts per poly. Each poly must only have 3 verts. Triangulate the mesh." % (obj.name, poly.loop_total))
            break
        #end if
    #end for
    
    return result
#end def

# -----------------------------------------------------------------------------
#
def menu_func_import_staticmesh(self, context):
    self.layout.operator(TUROK2_OT_import_staticmesh.bl_idname, text="Turok 2 Staticmesh (.staticmesh)")
#end def

# -----------------------------------------------------------------------------
#
def menu_func_export_staticmesh(self, context):
    self.layout.operator(TUROK2_OT_export_staticmesh.bl_idname, text="Turok 2 Staticmesh (.staticmesh)")
#end def

# -----------------------------------------------------------------------------
#
def register():
    bpy.utils.register_class(TUROK2_OT_import_staticmesh)
    bpy.utils.register_class(TUROK2_OT_export_staticmesh)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_staticmesh)
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export_staticmesh)
#end def

# -----------------------------------------------------------------------------
#
def unregister():
    bpy.utils.unregister_class(TUROK2_OT_import_staticmesh)
    bpy.utils.unregister_class(TUROK2_OT_export_staticmesh)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_staticmesh)
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export_staticmesh)
#end def
