import bpy
import urllib.request
import json
import sys
import re
import math
from .pcpp import Preprocessor, OutputDirective, Action
from bpy_extras.image_utils import load_image
from mathutils import (
    Vector,
    Quaternion,
    Matrix,
)
from datetime import (
    datetime,
    timedelta,
)

from . import (
    file_utils,
    kmaterial,
    paths,
    kdefs,
    panel_bone,
)


TUROK2_BL_INFO = {}
IMPORT_EXPORT_SCALE = 1.0
DOUBLE_CLICK_TIME = 0.35
TUROK2_WEBSITE = "https://behemothprogrammer.github.io/Turok2Blender/index.html"
TUROK2_COLLECTION_NAME = "Turok2"
TUROK2_BUILTIN_TEXTURES = ["_default", "_white", "_black", "_simpleShadow", "_wireframeMesh", "_wireframeMeshSelected"]
TUROK2_PLAYER_SCALE_NAME = '__Player_Scale_Turok2_'
TUROK2_WEAPON_CAMERA_NAME = '__Weapon_Camera_Turok2_'
TUROK2_WEAPON_FOV = 47.5
TUROK2_ACTOR_SCALE = 2.855002
TUROK2_WEAPON_OFFSET_SCALE = 105.0
TUROK2_WEAPON_INIT_TRANSLATION = Vector((4, 8, 1))
TUROK2_BONE_LENGTH = 0.001
TUROK2_KFA_POINT_DISPLAY_SIZE = 51.2
TUROK2_KFA_POINT_RADIUS_SCALE = 10.24
TUROK2_NODEGROUP_MIRRORUV = "Turok 2 Mirror UV"
TUROK2_NODEGROUP_REFLECTIVE = "Turok 2 Reflection"
TUROK2_ACTIONPOINT_NAME = '__TEMP_ActionPoint_Turok2_'
MAT_TEXSET_SUFFIX = "__texset__"

# Bone Symbols
BONE_TYPE_MASK = 0x03
BONE_TYPE_OFFSET = 0
BONE_SIDE_MASK = 0x04
BONE_SIDE_OFFSET = 2
BONE_PART_MASK = 0xC0
BONE_PART_OFFSET = 6

curTime = 0.0
deltaTime = 0.0
last_popup_message = ""
removeDataBlockCount = 0
copiedKeyFrame = None

# -----------------------------------------------------------------------------
#
def open_update_link(link):
    import webbrowser

    if link:
        try:
            webbrowser.open_new_tab(link)
        except:
            pass
        #end try except
    #end if
#end def

# -----------------------------------------------------------------------------
# Centers 3DView to view all visible objects
def view_all_objects():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                if region.type == 'WINDOW':
                    override = {'area': area, 'region': region}
                    bpy.ops.view3d.view_all(override)
                #end if
            #end for
        #end if
    #end for
#end def

# -----------------------------------------------------------------------------
# Centers 3DView to view the selected objects
def view_selected_object():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                if region.type == 'WINDOW':
                    override = {'area': area, 'region': region}
                    #bpy.ops.view3d.localview(override)
                    bpy.ops.view3d.view_selected(override)
                #end if
            #end for
        #end if
    #end for
#end def

# -----------------------------------------------------------------------------
# Adds/Removes Shader Groups for rendering the Turok object correctly
def update_shader_groups(bmat, obj=None):
    if bmat is not None and (obj is None or obj.t2.objType == "STATICMESH" or obj.t2.objType == "SKINNEDMESH"):
        usesReflection = False
        for param in bmat.t2.params:
            if param.name == "uReflective" and param.value == "1":
                usesReflection = True
            #end if
        #end for
        
        texImageNodes = [node for node in bmat.node_tree.nodes if node.bl_idname == 'ShaderNodeTexImage']
        
        #remove input links on all teximage nodes
        for texImageNode in texImageNodes:
            for nodeInput in texImageNode.inputs:
                for link in nodeInput.links:
                    bmat.node_tree.links.remove(link)
                #end for
            #end for
        #end for 

        #Add input links and shader group nodes for turok 2 rendering
        nodeGroups = [node for node in bmat.node_tree.nodes if node.bl_idname == 'ShaderNodeGroup' and node.node_tree is not None]
        reflectiveNodeGroups = [node for node in nodeGroups if node.node_tree.name == TUROK2_NODEGROUP_REFLECTIVE]
        mirroredNodeGroups = [node for node in nodeGroups if node.node_tree.name == TUROK2_NODEGROUP_MIRRORUV]
        texImageMirroredNodes = [node for node in texImageNodes if node.t2.mirrored]
        if usesReflection:
            for node in mirroredNodeGroups:
                bmat.node_tree.nodes.remove(node)
            #end for
            if len(reflectiveNodeGroups) == 0:
                reflect_groupnode = bmat.node_tree.nodes.new("ShaderNodeGroup")
                reflect_groupnode.node_tree = create_reflective_nodegroup()
                reflect_groupnode.location = (-480, 100)
                reflect_groupnode.select = False
            else:
                reflect_groupnode = reflectiveNodeGroups[0]
            #end if
            for texImageNode in texImageNodes:
                bmat.node_tree.links.new(texImageNode.inputs[0], reflect_groupnode.outputs[0])
            #end for
        else:
            for node in reflectiveNodeGroups:
                bmat.node_tree.nodes.remove(node)
            #end for
            if len(texImageMirroredNodes) > 0:
                if len(mirroredNodeGroups) == 0:
                    mirror_uv_groupnode = bmat.node_tree.nodes.new("ShaderNodeGroup")
                    mirror_uv_groupnode.node_tree = create_mirror_uv_nodegroup()
                    mirror_uv_groupnode.location = (-480, 100)
                    mirror_uv_groupnode.select = False
                else:
                    mirror_uv_groupnode = mirroredNodeGroups[0]
                #end if
                for texImageNode in texImageMirroredNodes:
                    bmat.node_tree.links.new(texImageNode.inputs[0], mirror_uv_groupnode.outputs[0])
                #end for
            else:
                for node in mirroredNodeGroups:
                    bmat.node_tree.nodes.remove(node)
                #end for
            #end if
        #end if
        
        #Forces rendering to refresh (There appears to be a bug when removing links. It doesn't update the shading rendering)
        bmat.use_backface_culling = bmat.use_backface_culling
    #end if
#end def
# -----------------------------------------------------------------------------
# Returns the Reflective node group. Creates it if it doesn't exist
def create_reflective_nodegroup():
    if TUROK2_NODEGROUP_REFLECTIVE not in bpy.data.node_groups:
        nodeTree = bpy.data.node_groups.new(TUROK2_NODEGROUP_REFLECTIVE, 'ShaderNodeTree')
        
        # create group inputs
        group_inputs = nodeTree.nodes.new('NodeGroupInput')
        group_inputs.location = (-1600, 0)

        # create group outputs
        group_outputs = nodeTree.nodes.new('NodeGroupOutput')
        group_outputs.location = (400, 0)
        nodeTree.outputs.new('NodeSocketVector','UV Output')

        # create content nodes in a group
        node_tex_coord = nodeTree.nodes.new('ShaderNodeTexCoord')
        node_tex_coord.location = (-1400, 0)

        node_reflect = nodeTree.nodes.new('ShaderNodeVectorMath')
        node_reflect.operation = 'REFLECT'
        node_reflect.location = (-1200, 0)
        
        node_sep_xyz = nodeTree.nodes.new('ShaderNodeSeparateXYZ')
        node_sep_xyz.location = (-1000, 0)
        
        node_math_add = nodeTree.nodes.new('ShaderNodeMath')
        node_math_add.operation = 'ADD'
        node_math_add.inputs[1].default_value = 1.0
        node_math_add.location = (-800, -200)

        node_com_xyz = nodeTree.nodes.new('ShaderNodeCombineXYZ')
        node_com_xyz.location = (-600, 0)
        
        node_normalize = nodeTree.nodes.new('ShaderNodeVectorMath')
        node_normalize.operation = 'NORMALIZE'
        node_normalize.location = (-400, 0)
        
        node_sep_xyz2 = nodeTree.nodes.new('ShaderNodeSeparateXYZ')
        node_sep_xyz2.location = (-200, 0)
        
        node_muladd = nodeTree.nodes.new('ShaderNodeMath')
        node_muladd.operation = 'MULTIPLY_ADD'
        node_muladd.inputs[1].default_value = 0.5
        node_muladd.inputs[2].default_value = 0.5
        node_muladd.location = (0, 0)

        node_muladd2 = nodeTree.nodes.new('ShaderNodeMath')
        node_muladd2.operation = 'MULTIPLY_ADD'
        node_muladd2.inputs[1].default_value = 0.5
        node_muladd2.inputs[2].default_value = 0.5
        node_muladd2.location = (0, -200)
        
        node_com_xyz2 = nodeTree.nodes.new('ShaderNodeCombineXYZ')
        node_com_xyz2.location = (200, 0)
        
        # link nodes together
        nodeTree.links.new(node_reflect.inputs[0], node_tex_coord.outputs[4])
        nodeTree.links.new(node_reflect.inputs[1], node_tex_coord.outputs[1])
        nodeTree.links.new(node_sep_xyz.inputs[0], node_reflect.outputs[0])
        nodeTree.links.new(node_com_xyz.inputs[0], node_sep_xyz.outputs[0])
        nodeTree.links.new(node_com_xyz.inputs[1], node_sep_xyz.outputs[1])
        nodeTree.links.new(node_math_add.inputs[0], node_sep_xyz.outputs[2])
        nodeTree.links.new(node_com_xyz.inputs[2], node_math_add.outputs[0])
        nodeTree.links.new(node_normalize.inputs[0], node_com_xyz.outputs[0])
        nodeTree.links.new(node_sep_xyz2.inputs[0], node_normalize.outputs[0])
        nodeTree.links.new(node_muladd.inputs[0], node_sep_xyz2.outputs[0])
        nodeTree.links.new(node_muladd2.inputs[0], node_sep_xyz2.outputs[1])
        nodeTree.links.new(node_com_xyz2.inputs[0], node_muladd.outputs[0])
        nodeTree.links.new(node_com_xyz2.inputs[1], node_muladd2.outputs[0])
        nodeTree.links.new(group_outputs.inputs[0], node_com_xyz2.outputs[0])
        
        for shaderNode in nodeTree.nodes:
            shaderNode.select = False
        #end for
    #end if
    
    return bpy.data.node_groups[TUROK2_NODEGROUP_REFLECTIVE]
#end def

# -----------------------------------------------------------------------------
# Returns the Mirror node group. Creates it if it doesn't exist
def create_mirror_uv_nodegroup():
    if TUROK2_NODEGROUP_MIRRORUV not in bpy.data.node_groups:
        nodeTree = bpy.data.node_groups.new(TUROK2_NODEGROUP_MIRRORUV, 'ShaderNodeTree')
        
        # create group inputs
        group_inputs = nodeTree.nodes.new('NodeGroupInput')
        group_inputs.location = (-800, 0)

        # create group outputs
        group_outputs = nodeTree.nodes.new('NodeGroupOutput')
        group_outputs.location = (300, 0)
        nodeTree.outputs.new('NodeSocketVector','UV Output')

        # create content nodes in a group
        node_tex_coord = nodeTree.nodes.new('ShaderNodeTexCoord')
        node_tex_coord.location = (-550, 0)
        
        node_sep_xyz = nodeTree.nodes.new('ShaderNodeSeparateXYZ')
        node_sep_xyz.location = (-350, 0)
        
        node_pp_x = nodeTree.nodes.new('ShaderNodeMath')
        node_pp_x.operation = 'PINGPONG'
        node_pp_x.inputs[1].default_value = 0.987
        node_pp_x.location = (-150, 100)

        node_pp_y = nodeTree.nodes.new('ShaderNodeMath')
        node_pp_y.operation = 'PINGPONG'
        node_pp_y.inputs[1].default_value = 0.987
        node_pp_y.location = (-150, -100)

        node_com_xyz = nodeTree.nodes.new('ShaderNodeCombineXYZ')
        node_com_xyz.location = (50, 0)
        
        # link nodes together
        nodeTree.links.new(node_sep_xyz.inputs[0], node_tex_coord.outputs[2])
        nodeTree.links.new(node_pp_x.inputs[0], node_sep_xyz.outputs[0])
        nodeTree.links.new(node_pp_y.inputs[0], node_sep_xyz.outputs[1])
        nodeTree.links.new(node_com_xyz.inputs[0], node_pp_x.outputs[0])
        nodeTree.links.new(node_com_xyz.inputs[1], node_pp_y.outputs[0])
        nodeTree.links.new(group_outputs.inputs[0], node_com_xyz.outputs[0])
        
        for shaderNode in nodeTree.nodes:
            shaderNode.select = False
        #end for
    #end if
    
    return bpy.data.node_groups[TUROK2_NODEGROUP_MIRRORUV]
#end def

# -----------------------------------------------------------------------------
# Returns True if the bone index is valid
def bone_index_is_valid(armObj, bone):
    if bone.t2.boneIndex >= len(armObj.data.bones):
        return False
    #end if
    for b in armObj.data.bones:
        if b != bone and b.t2.boneIndex == bone.t2.boneIndex:
            return False
        #end if
    #end for
    
    return True
#end def

# -----------------------------------------------------------------------------
# Returns the objs bounding box obj or None
def get_bounding_box(obj, boneName=''):
    bbObj = None
    if boneName:
        bbObj = get_bone_boundingbox(obj, boneName)
    elif obj.t2.objType == "BOUNDING BOX" and obj.type == 'MESH':
        bbObj = obj
    elif obj.t2.objType == "SKINNEDMESH" and obj.type == 'ARMATURE':
        meshObj = get_skinnedmesh_mesh(obj)
        if meshObj is not None:
            bbObj = get_skinnedmesh_boundingbox(meshObj)
        #end if
    elif obj.t2.objType == "SKINNEDMESH" and obj.type == 'MESH':
        bbObj = get_skinnedmesh_boundingbox(obj)
    #end if
    
    return bbObj
#end def

# -----------------------------------------------------------------------------
# Returns turok2PointObj
def create_kfa_point(context, kfa, armObj):
    turok2PointObj = get_object(TUROK2_ACTIONPOINT_NAME)
    if turok2PointObj is None:
        turok2PointObj = bpy.data.objects.new(TUROK2_ACTIONPOINT_NAME, None)
        context.scene.collection.objects.link(turok2PointObj)
    #end if
    
    argNames = kfa.get_arg_names()
    radiusArgIndex = index_of(argNames, "Radius")
    if radiusArgIndex != -1:
        turok2PointObj.empty_display_type = 'SPHERE'
        turok2PointObj.empty_display_size = 1.0
        s = (kfa.get_args(radiusArgIndex) / 256.0) * TUROK2_KFA_POINT_RADIUS_SCALE
        turok2PointObj.scale = (s, s, s)
    else:
        turok2PointObj.empty_display_type = 'PLAIN_AXES'
        turok2PointObj.empty_display_size = TUROK2_KFA_POINT_DISPLAY_SIZE
        s = 1.0
        turok2PointObj.scale = (s, s, s)
    #end if
    turok2PointObj.show_in_front = True
    turok2PointObj.location = (-kfa.offset[0], -kfa.offset[1], kfa.offset[2])
    
    clConstraint = None
    for constraint in turok2PointObj.constraints:
        if constraint.type == 'COPY_LOCATION':
            clConstraint = constraint
            break
        #end if
    #end for
    if clConstraint is None:
        clConstraint = turok2PointObj.constraints.new('COPY_LOCATION')
    #end if
    clConstraint.influence = 1.0
    clConstraint.use_offset = True
    clConstraint.use_x = True
    clConstraint.use_y = True
    clConstraint.use_z = False
    
    clConstraint.target = armObj
    if armObj.t2.kfaPointIsRoot:
        #always target first root bone with lowest bone index value
        rootBones = [bone for bone in armObj.data.bones if bone.parent is None]
        rootBones.sort(key=lambda bone: bone.t2.boneIndex)
        clConstraint.subtarget = rootBones[0].name
    else:
        clConstraint.subtarget = ''
    #end if
    
    return turok2PointObj
#end def

# -----------------------------------------------------------------------------
# Returns index of item in array or -1 if not found
def index_of(ary, item):
    return next((i for i, x in enumerate(ary) if x == item), -1)
#end def

# -----------------------------------------------------------------------------
# Returns a list of the used frame Indexes. Always includes the first and last frame.
# aryValues: An array of frame indexes(int) for all frames used
def get_used_frames(aryValues):
    usedFrames = []
    frameLength = len(aryValues)
    if frameLength > 0:
        usedFrames.append(0)
        if frameLength > 1:
            for frameIndex in range(1, frameLength - 1):
                if aryValues[frameIndex] != aryValues[frameIndex-1] or aryValues[frameIndex] != aryValues[frameIndex+1]:
                    usedFrames.append(frameIndex)
                #end if
            #end for
            usedFrames.append(frameLength - 1)
        #end if
    #end if
    
    return usedFrames
#end def

# -----------------------------------------------------------------------------
# Returns a flags value from a set of enum identifiers
def get_enumflags_value(enumItems, enumIDSet):
    result = 0
    for e in enumItems:
        if e[0] in enumIDSet:
            result |= e[3]
        #end if
    #end for
            
    return result
#end def

# -----------------------------------------------------------------------------
# Returns a set of enum identifiers if the values bits are on. Else returns empty set
def get_enumflags_id(enumItems, enumValue):
    ids = set()
    for e in enumItems:
        if enumValue & e[3] != 0:
            ids.add(e[0])
        #end if
    #end for
            
    return ids
#end def

# -----------------------------------------------------------------------------
# Returns the value of the least significant set bit
def get_lowest_setbit(value):
    for i in range(32):
        if (value & (1 << i)) != 0:
            return (1 << i)
        #end if
    #end for

    return 0
#end def

# -----------------------------------------------------------------------------
# Returns the bit index of the least significant set bit
def get_lowest_setbit_index(value):
    for i in range(32):
        if (value & (1 << i)) != 0:
            return i
        #end if
    #end for

    return 0
#end def

# -----------------------------------------------------------------------------
# Returns the objects bounding box min and max values
def get_boundingbox_minmax(obj):
    objBB = obj.bound_box
    bbMin = Vector((objBB[0][0], objBB[0][1], objBB[0][2]))
    bbMax = Vector((objBB[0][0], objBB[0][1], objBB[0][2]))
    for i in range(1, 8):
        for j in range(3):
            if bbMin[j] > objBB[i][j]:
                bbMin[j] = objBB[i][j]
            #end if
            if bbMax[j] < objBB[i][j]:
                bbMax[j] = objBB[i][j]
            #end if
        #end for
    #end for
    
    return bbMin, bbMax
#end def

# -----------------------------------------------------------------------------
# Creates a bounding box mesh that gets attached to parentObj. Returns the BoundingBox Object.
def create_bounding_box(parentObj, bbMin, bbMax):
    bbObj = create_cube("Bounding Box", bbMin, bbMax)
    bbObj.t2.objType = "BOUNDING BOX"
    get_collection(TUROK2_COLLECTION_NAME).objects.link(bbObj)
    bbObj.lock_location = (True, True, True)
    bbObj.lock_rotation = (True, True, True)
    bbObj.lock_rotation_w = True
    bbObj.lock_rotations_4d = True
    bbObj.lock_scale = (True, True, True)
    bbObj.show_in_front = True
    bbObj.display_type = "BOUNDS"
    bbObj.display.show_shadows = False
    bbObj.parent = parentObj
    ctConstraint = bbObj.constraints.new('COPY_TRANSFORMS')
    ctConstraint.target = parentObj
    ctConstraint.head_tail = 0.0
    ctConstraint.mix_mode = 'REPLACE'
    ctConstraint.owner_space = 'WORLD'
    ctConstraint.target_space = 'WORLD'
    ctConstraint.influence = 1.0
    
    return bbObj
#end def

# -----------------------------------------------------------------------------
# Creates a bounding box mesh that gets attached to the armature objects bone. Returns the BoundingBox Object.
def create_bone_bounding_box(bbName, armObj, boneIndex, bbMin, bbMax):
    bbObj = create_cube(bbName, bbMin, bbMax)
    bbObj.t2.objType = "BOUNDING BOX"
    get_collection(TUROK2_COLLECTION_NAME).objects.link(bbObj)
    bbObj.lock_location = (True, True, True)
    bbObj.lock_rotation = (True, True, True)
    bbObj.lock_rotation_w = True
    bbObj.lock_rotations_4d = True
    bbObj.lock_scale = (True, True, True)
    bbObj.show_in_front = True
    bbObj.display_type = "BOUNDS"
    bbObj.display.show_shadows = False
    bbObj.parent = armObj
    bbObj.parent_type = 'BONE'
    boneName = find_bone_name(armObj, boneIndex)
    bbObj.parent_bone = boneName
    ctConstraint = bbObj.constraints.new('COPY_TRANSFORMS')
    ctConstraint.target = armObj
    ctConstraint.head_tail = 0.0
    ctConstraint.mix_mode = 'REPLACE'
    ctConstraint.owner_space = 'WORLD'
    ctConstraint.target_space = 'WORLD'
    ctConstraint.influence = 1.0
    ctConstraint.subtarget = boneName
    
    return bbObj
#end def

# -----------------------------------------------------------------------------
# Creates a point object that gets attached to the armature objects bone. Returns the point Object.
def create_bone_hotpoint(armObj, boneIndex):
    pointObj = create_turok2_object("Hot Point", None, "HOT POINT")
    pointObj.lock_rotation = (True, True, True)
    pointObj.lock_rotation_w = True
    pointObj.lock_rotations_4d = True
    pointObj.lock_scale = (True, True, True)
    pointObj.show_in_front = True
    pointObj.empty_display_type = 'PLAIN_AXES'
    pointObj.empty_display_size = TUROK2_KFA_POINT_DISPLAY_SIZE
    pointObj.parent = armObj
    pointObj.parent_type = 'BONE'
    boneName = find_bone_name(armObj, boneIndex)
    pointObj.parent_bone = boneName
    
    return pointObj
#end def

# -----------------------------------------------------------------------------
# Returns an int8 of the symbols from the bone
def get_bone_symbols(bone):
    typeIndex = get_enum_type(panel_bone.EnumSymbolTypes, bone.t2.symbolType)
    sideIndex = get_enum_type(panel_bone.EnumSymbolSides, bone.t2.symbolSide)
    partIndex = 0
    if bone.t2.symbolType == "BODY":
        partIndex = get_enum_type(panel_bone.EnumSymbolPartsBody, bone.t2.symbolPartsBody)
    elif bone.t2.symbolType == "HEAD":
        partIndex = get_enum_type(panel_bone.EnumSymbolPartsHead, bone.t2.symbolPartsHead)
    elif bone.t2.symbolType == "ARM":
        partIndex = get_enum_type(panel_bone.EnumSymbolPartsArm, bone.t2.symbolPartsArm)
    else: #"LEG"
        partIndex = get_enum_type(panel_bone.EnumSymbolPartsLeg, bone.t2.symbolPartsLeg)
    #end if
    
    return (typeIndex << BONE_TYPE_OFFSET) | (sideIndex << BONE_SIDE_OFFSET) | (partIndex << BONE_PART_OFFSET);
#end def

# -----------------------------------------------------------------------------
# Sets the bones symbol enums from int8 symbol flags
def set_bone_symbols(bone, flags):
    symbolType = (flags & BONE_TYPE_MASK) >> BONE_TYPE_OFFSET;
    symbolSide = (flags & BONE_SIDE_MASK) >> BONE_SIDE_OFFSET;
    symbolPart = (flags & BONE_PART_MASK) >> BONE_PART_OFFSET;
    bone.t2.symbolType = get_enum_id(panel_bone.EnumSymbolTypes, symbolType)
    bone.t2.symbolSide = get_enum_id(panel_bone.EnumSymbolSides, symbolSide)
    bone.t2.symbolPartsBody = get_enum_id(panel_bone.EnumSymbolPartsBody, 0)
    bone.t2.symbolPartsHead = get_enum_id(panel_bone.EnumSymbolPartsHead, 0)
    bone.t2.symbolPartsArm = get_enum_id(panel_bone.EnumSymbolPartsArm, 0)
    bone.t2.symbolPartsLeg = get_enum_id(panel_bone.EnumSymbolPartsLeg, 0)
    if bone.t2.symbolType == "BODY":
        bone.t2.symbolPartsBody = get_enum_id(panel_bone.EnumSymbolPartsBody, symbolPart)
    elif bone.t2.symbolType == "HEAD":
        bone.t2.symbolPartsHead = get_enum_id(panel_bone.EnumSymbolPartsHead, symbolPart)
    elif bone.t2.symbolType == "ARM":
        bone.t2.symbolPartsArm = get_enum_id(panel_bone.EnumSymbolPartsArm, symbolPart)
    else: #"LEG"
        bone.t2.symbolPartsLeg = get_enum_id(panel_bone.EnumSymbolPartsLeg, symbolPart)
    #end if
#end def

# -----------------------------------------------------------------------------
# Returns the name of the bone using the bones symbol and index values
def get_bonename_symbols(flags, index, useIndex=False):
    symbolType = (flags & BONE_TYPE_MASK) >> BONE_TYPE_OFFSET;
    symbolSide = (flags & BONE_SIDE_MASK) >> BONE_SIDE_OFFSET;
    symbolPart = (flags & BONE_PART_MASK) >> BONE_PART_OFFSET;
    typeName = get_enum_id(panel_bone.EnumSymbolTypes, symbolType)
    sideName = "L" if get_enum_id(panel_bone.EnumSymbolSides, symbolSide) == "LEFT" else "R"
    partName = ''
    if typeName == "BODY":
        partName = get_enum_id(panel_bone.EnumSymbolPartsBody, symbolPart)
    elif typeName == "HEAD":
        partName = get_enum_id(panel_bone.EnumSymbolPartsHead, symbolPart)
    elif typeName == "ARM":
        partName = get_enum_id(panel_bone.EnumSymbolPartsArm, symbolPart)
    else: #"LEG"
        partName = get_enum_id(panel_bone.EnumSymbolPartsLeg, symbolPart)
    #end if
    if useIndex:
        return "%03d_%s.%s" % (index, partName, sideName)
    else:
        return "%s.%s" % (partName, sideName)
    #end if
#end def

# -----------------------------------------------------------------------------
# Finds the bone with the boneIndex and returns it's name else returns an empty string
def find_bone_name(armObj, boneIndex):
    for bone in armObj.data.bones:
        if bone.t2.boneIndex == boneIndex:
            return bone.name
        #end if
    #end for
    
    return ''
#end def

# -----------------------------------------------------------------------------
# Returns a list of materials that are actually used in the obj. Assumes obj is type 'MESH'
def obj_used_materials(obj):
    materialIndexsUsed = set()
    matCount = len(obj.data.materials)
    for poly in obj.data.polygons:
        if poly.material_index not in materialIndexsUsed and poly.material_index < matCount:
            materialIndexsUsed.add(poly.material_index)
            if len(materialIndexsUsed) == matCount: #used all materials
                break
            #end if
        #end if
    #end for
    
    return [material for matIndex, material in enumerate(obj.data.materials) if material is not None and matIndex in materialIndexsUsed]
#end def

# -----------------------------------------------------------------------------
#
def lerp(a, b, t):
    return (1 - t) * a + t * b
#end def

# -----------------------------------------------------------------------------
#
def clamp(value, a, b):
    if value < a:
        return a
    if value > b:
        return b
    return value
#end def

# -----------------------------------------------------------------------------
# Returns current time in float
def time_from(timeOffset=0.0):
    global curTime
    return curTime + timeOffset
#end def

# -----------------------------------------------------------------------------
# Returns true if current time has reached the target time
def time_reached(time):
    global curTime
    return curTime >= time
#end def

# -----------------------------------------------------------------------------
# Returns the amount of time from curTime to time
def time_diff(time):
    global curTime
    return math.fabs(curTime - time)
#end def

# -----------------------------------------------------------------------------
# Returns the normalized time from 0..1 using the params and curTime
def time_normalize(timeFrom, timeTo):
    length = timeTo - timeFrom;
    if length <= 0.0:
        return 1.0
    
    return clamp(time_diff(timeFrom) / length, 0.0, 1.0)
#end def

# -----------------------------------------------------------------------------
# Returns the normalized time from 0..1 using the params
def time_normalize(timeCurrent, timeFrom, timeTo):
    length = timeTo - timeFrom;
    if length <= 0.0:
        return 1.0;
    
    return clamp((timeCurrent - timeFrom) / length, 0.0, 1.0);
#end def

# -----------------------------------------------------------------------------
# Removes unused data blocks (does not have fake user and has at least 1 user)
def clean_data_block(collectionName, block, silent):
    global removeDataBlockCount
    
    for data in block:
        if data.use_fake_user or data.users:
            continue
        #end if
        
        if not silent:
            print("%s removed block: %s" % (collectionName, data.name))
        #end if
        removeDataBlockCount += 1
        block.remove(data)
    #end for
#end def

# -----------------------------------------------------------------------------
# Removes unused data blocks (does not have fake user and has at least 1 user)
# that a model would use.
def clean_all_unused_model_data(silent=True):
    global removeDataBlockCount
    
    removeCount = 0
    while True:
        removeDataBlockCount = 0
        clean_data_block("actions", bpy.data.actions, silent)
        clean_data_block("armatures", bpy.data.armatures, silent)
        clean_data_block("images", bpy.data.images, silent)
        clean_data_block("materials", bpy.data.materials, silent)
        clean_data_block("meshes", bpy.data.meshes, silent)
        if removeDataBlockCount == 0:
            break
        #end if
        removeCount += removeDataBlockCount
    #end while
    
    return removeCount
#end def

# -----------------------------------------------------------------------------
# Removes unused data blocks (does not have fake user and has at least 1 user)
def clean_all_unused_data(silent=False):
    global removeDataBlockCount
    
    removeCount = 0
    while True:
        removeDataBlockCount = 0
        clean_data_block("actions", bpy.data.actions, silent)
        clean_data_block("armatures", bpy.data.armatures, silent)
        #clean_data_block("brushes", bpy.data.brushes, silent)
        clean_data_block("cache_files", bpy.data.cache_files, silent)
        clean_data_block("cameras", bpy.data.cameras, silent)
        clean_data_block("collections", bpy.data.collections, silent)
        clean_data_block("curves", bpy.data.curves, silent)
        clean_data_block("fonts", bpy.data.fonts, silent)
        clean_data_block("grease_pencils", bpy.data.grease_pencils, silent)
        clean_data_block("images", bpy.data.images, silent)
        clean_data_block("lattices", bpy.data.lattices, silent)
        clean_data_block("libraries", bpy.data.libraries, silent)
        clean_data_block("lightprobes", bpy.data.lightprobes, silent)
        clean_data_block("lights", bpy.data.lights, silent)
        clean_data_block("linestyles", bpy.data.linestyles, silent)
        clean_data_block("masks", bpy.data.masks, silent)
        clean_data_block("materials", bpy.data.materials, silent)
        clean_data_block("meshes", bpy.data.meshes, silent)
        clean_data_block("metaballs", bpy.data.metaballs, silent)
        clean_data_block("movieclips", bpy.data.movieclips, silent)
        clean_data_block("node_groups", bpy.data.node_groups, silent)
        clean_data_block("objects", bpy.data.objects, silent)
        clean_data_block("paint_curves", bpy.data.paint_curves, silent)
        clean_data_block("palettes", bpy.data.palettes, silent)
        clean_data_block("particles", bpy.data.particles, silent)
        clean_data_block("scenes", bpy.data.scenes, silent)
        clean_data_block("shape_keys", bpy.data.shape_keys, silent)
        clean_data_block("sounds", bpy.data.sounds, silent)
        clean_data_block("speakers", bpy.data.speakers, silent)
        clean_data_block("texts", bpy.data.texts, silent)
        clean_data_block("textures", bpy.data.textures, silent)
        clean_data_block("window_managers", bpy.data.window_managers, silent)
        clean_data_block("workspaces", bpy.data.workspaces, silent)
        clean_data_block("worlds", bpy.data.worlds, silent)
        if removeDataBlockCount == 0:
            break
        #end if
        removeCount += removeDataBlockCount
    #end while
    
    return removeCount
#end def

# -----------------------------------------------------------------------------
# Gets an existing material or creates a new one if no material with matName and contains the same kmat data doesn't exist
def get_or_create_material(kmatName, kmat):
    #get existing blender material
    bmat = bpy.data.materials.get(kmatName)
    if bmat is None:
        bmat = bpy.data.materials.get(get_material_texset_name(kmatName))
    #end if

    if bmat is not None:
        if not kmaterial.bmat_has_kmat(bmat, kmat):
            bmat = bpy.data.materials.new(kmatName)
        #end if
    else:
        bmat = bpy.data.materials.new(kmatName)
    #end if
    
    return bmat
#end def

# -----------------------------------------------------------------------------
# Returns the mesh object of the armature object or None if not found.
def get_skinnedmesh_mesh(armObj):
    for obj in bpy.data.objects:
        if obj.type == 'MESH' and obj.parent == armObj and obj.t2.objType == "SKINNEDMESH" and obj.parent_type == "OBJECT":
            return obj
        #end if
    #end for
    
    return None
#end def

# -----------------------------------------------------------------------------
# Returns the bounding box for this meshObj or None if not found
def get_skinnedmesh_boundingbox(meshObj):
    for obj in bpy.data.objects:
        if obj.type == 'MESH' and obj.parent == meshObj and obj.t2.objType == "BOUNDING BOX" and obj.parent_type == "OBJECT":
            return obj
        #end if
    #end for
    
    return None
#end def

# -----------------------------------------------------------------------------
# Returns the bounding box for this Armatures Bone or None if not found
def get_bone_boundingbox(armObj, boneName):
    objs = [o for o in bpy.data.objects if o.type == 'MESH' and o.t2.objType == "BOUNDING BOX" and o.parent == armObj and o.parent_type == "BONE"]
    for obj in objs:
        if obj.parent_bone == boneName:
            return obj
        #end if
    #end for
    
    return None
#end def

# -----------------------------------------------------------------------------
# Returns a list of all the children of obj and their children
def get_all_mesh_children(obj, includeSelf=False):
    return [o for o in bpy.data.objects if o.type == 'MESH' and obj_child_of(obj, o, includeSelf)]
#end def

# -----------------------------------------------------------------------------
# Returns true if childObj is a child or sub child and so on of obj
def obj_child_of(obj, childObj, includeSelf=False):
    if obj == childObj:
        if includeSelf:
            return True
        else:
            return False
        #end if
    #end if
    parentObj = childObj.parent
    while parentObj is not None:
        if parentObj == obj:
            return True
        #end if
        parentObj = parentObj.parent
    #end while
    
    return False
#end def

# -----------------------------------------------------------------------------
# Sets the object and it's children materials to use the material with the targetTextureSetIndex
def set_obj_material_tex_index(obj, targetTextureSetIndex):
    #for all child objects in object
    meshChildren = get_all_mesh_children(obj, True)
    for meshChild in meshChildren:
        for matIndex, mat in enumerate(meshChild.data.materials):
            #skip empty materials
            if mat is None:
                continue
            
            firstMat = get_material_texset_index(mat, 0, False)
            newMat = get_material_texset_index(mat, targetTextureSetIndex, True)
            #if is already the right material don't do anything
            if newMat == mat:
                continue
            #end if
            
            #create a new material for the texturesetIndex if it doesn't exist
            if newMat is None:
                newMat = create_material_texset(firstMat, targetTextureSetIndex)
                #texture sets doesn't have a index of targetTextureSetIndex, then use the firstMat
                if newMat is None:
                    newMat = firstMat
                    #if is already the right material don't do anything
                    if newMat == mat:
                        continue
                    #end if
                #end if
            #end if
            
            meshChild.data.materials[matIndex] = newMat
        #end for
    #end for
#end def

# -----------------------------------------------------------------------------
# Returns created material with texture set index. Returns None if texSetIndex doesn't exist in defs
def create_material_texset(firstMat, texSetIndex):
    # Mat has no nodes
    if firstMat.node_tree is None:
        return None
    #end if
    
    #check if any texture nodes in firstMat have a def for the texSetIndex
    isValidTexSetIndex = False
    for node in firstMat.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeTexImage' and node.image is not None:
            texSetKey = paths.texture_filepath_to_local_path(node.image.filepath)
            texSet = kdefs.defs["texsets"].get(texSetKey)
            if texSet is not None and texSetIndex >= 0 and texSetIndex < len(texSet["textures"]):
                isValidTexSetIndex = True
                break
            #end if
        #end if
    #end for
    
    if not isValidTexSetIndex:
        return None
    #end if

    realMatName, matTexIndex = get_real_material_name(firstMat.name)
    newName = get_material_texset_name(realMatName, texSetIndex)
    newMat = firstMat.copy()
    newMat.name = newName

    #for every texNode
    for node in newMat.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeTexImage' and node.image is not None:
            texSetKey = paths.texture_filepath_to_local_path(node.image.filepath)
            texLocalPath = get_textureset_texturepath(texSetKey, texSetIndex)
            if texLocalPath is not None:
                texPath = paths.find_t2_file_path(texLocalPath)
                if texPath is not None:
                    node.image = load_image(texPath, check_existing=True)
                #end if
            #end if
        #end if
    #end for
    
    return newMat
#end def

# -----------------------------------------------------------------------------
# 
def get_textureset_texturepath(key, index):
    texSet = kdefs.defs["texsets"].get(key, 0)
    if index >= 0 and index < len(texSet["textures"]):
        return texSet["textures"][index]
    #end if
    
    return texSet["textures"][0]
#end def

# -----------------------------------------------------------------------------
# Returns material index for this mesh of the mat name or -1 if not found
def find_material_index(mesh, mat):
    for i in range(len(mesh.materials)):
        if mat == mesh.materials[i]:
            return i
        #end if
    #end for
    return -1
#end def

# -----------------------------------------------------------------------------
# Returns the name of the material with the texture set suffix and number
def get_material_texset_name(matName, textureSetIndex=0):
    return "%s%s%i" % (matName, MAT_TEXSET_SUFFIX, textureSetIndex)
#end def

# -----------------------------------------------------------------------------
# Returns the material with the texture set index in it's name based on the mat name
def get_material_texset_index(mat, targetTexSetIndex, useNoneForMat=False):
    matName, textureSetIndex = get_real_material_name(mat.name)
    if targetTexSetIndex == 0:
        baseMat = bpy.data.materials.get(matName)
        if baseMat is not None:
            return baseMat
        #end if
    #end if
    baseMat = bpy.data.materials.get(get_material_texset_name(matName, targetTexSetIndex))
    if baseMat is not None:
        return baseMat
    #end if
    if useNoneForMat:
        return None
    else:
        return mat
    #end if
#end def

# -----------------------------------------------------------------------------
# Returns the name of the material without the texture set suffix and then
# returns the texture set index if any else returns -1
def get_real_material_name(matName):
    match1 = re.search("%s([0-9]+)$" % (MAT_TEXSET_SUFFIX), matName)
    if match1:
        textureSetIndex = int(match1.group(1))
        realMatName = matName[:match1.span()[0]]
        return realMatName, textureSetIndex
    #end if
    
    return matName, -1
#end def

# -----------------------------------------------------------------------------
# Returns the enum identifier if the value if matches the enums value. Else returns ""
def get_enum_id(enumItems, enumValue):
    for e in enumItems:
        if e[3] == enumValue:
            return e[0]
        #end if
    #end for
            
    return ''
#end def

# -----------------------------------------------------------------------------
# Returns the enum value if matches the enums stringID. Else returns 0.
def get_enum_type(enumItems, enumStr):
    for e in enumItems:
        if e[0] == enumStr:
            return e[3]
        #end if
    #end for
    return 0
#end def

# -----------------------------------------------------------------------------
# Returns true if string is in the enumItems
def has_enum_type(enumItems, enumStr):
    for e in enumItems:
        if e[0] == enumStr:
            return True
        #end if
    #end for
            
    return False
#end def

# -----------------------------------------------------------------------------
# Create common text Preprocessor
def create_preprocessor():
    p = Preprocessor()
    p.compress = True
    p.line_directive = None
    p.add_path(prefs().devDirectory)
    p.add_path(prefs().gameDirectory)
    
    return p
#end def

# -----------------------------------------------------------------------------
# Returns a list of the Preprocessor Tokens without CPP_WS types
def get_preprocessor_tokens(p):
    tokens = []
    while True:
        tok = p.token()
        if not tok:
            break
        if tok.type == 'CPP_WS':
            continue
        tokens.append(tok)
    #end while
    
    return tokens
#end def

# -----------------------------------------------------------------------------
# Returns the Vector (as a tuple of 3 so it can be pickled) from the string
def get_vector_from_string(s):
    text = string_remove_quotes(s)
    p = create_preprocessor()
    p.parse(text)
    tokens = get_preprocessor_tokens(p)
    if len(tokens) == 3:
        return (tokens[0].value, tokens[1].value, tokens[2].value)
    #end if
    
    return (0.0, 0.0, 0.0)
#end def

# -----------------------------------------------------------------------------
# Returns a 4 float tuple from the string
def get_vector4_from_string(s):
    text = string_remove_quotes(s)
    p = create_preprocessor()
    p.parse(text)
    tokens = get_preprocessor_tokens(p)
    if len(tokens) == 4:
        return (tokens[0].value, tokens[1].value, tokens[2].value, tokens[3].value)
    #end if
    
    return (0.0, 0.0, 0.0, 0.0)
#end def

# -----------------------------------------------------------------------------
# Removes the first and last character as quotes if the string has them
def string_remove_quotes(s):
    if s.startswith('"') and s.endswith('"'):
        return s[1:-1]
    #end if
    
    return s
#end def

# -----------------------------------------------------------------------------
# Mutiplies each vector component in a by b and returns the new Vector
def vector_multiply(a, b):
    return Vector((a.x * b.x, a.y * b.y, a.z * b.z))
#end def

# -----------------------------------------------------------------------------
# Creates a cube mesh object in the scene and returns the Object
def create_cube(objName='Cube', width=1.0, height=1.0, depth=1.0, triangulate=False):
    return create_cube(objName, Vector((-width, -depth, -height)), Vector((width, depth, height)), triangulate)
#end def

# -----------------------------------------------------------------------------
# Creates a cube mesh object in the scene and returns the Object
def create_cube(objName, min, max, triangulate=False):
    verts = []
    faces = []
    if triangulate:
        verts = [(min.x, min.y, min.z), (min.x, min.y, max.z),
                (min.x, max.y, min.z), (min.x, max.y, max.z),
                (max.x, min.y, min.z), (max.x, min.y, max.z),
                (max.x, max.y, min.z), (max.x, max.y, max.z)]
        faces = [(1, 2, 0), (3, 6, 2), (7, 4, 6), (5, 0, 4), (6, 0, 2), (3, 5, 7), (1, 3, 2),
                (3, 7, 6), (7, 5, 4), (5, 1, 0), (6, 4, 0), (3, 1, 5)]
    else:
        verts = [
            (max.x, max.y, min.z),
            (max.x, min.y, min.z),
            (min.x, min.y, min.z),
            (min.x, max.y, min.z),
            (max.x, max.y, max.z),
            (max.x, min.y, max.z),
            (min.x, min.y, max.z),
            (min.x, max.y, max.z)
        ]
        faces = [
            (0, 1, 2, 3),
            (4, 7, 6, 5),
            (0, 4, 5, 1),
            (1, 5, 6, 2),
            (2, 6, 7, 3),
            (4, 0, 3, 7)
        ]
    #end if
        
    mesh = bpy.data.meshes.new("Box")
    mesh.from_pydata(verts, [], faces)
    mesh.validate(clean_customdata=False)
    mesh.update(calc_edges=True)

    obj = bpy.data.objects.new(objName, mesh)
    bpy.context.scene.collection.objects.link(obj)
    return obj
#end def

# -----------------------------------------------------------------------------
# Returns a string of the chars from set that are in str
def string_contains_any(str, set):
    result = ""
    for c in set:
        if c in str:
            result += c
        #end if
    #end for
    return result
#end def

# -----------------------------------------------------------------------------
# Returns the name of the material without the texture set suffix and then
# returns the texture set index if any else returns -1
def get_real_material_name(matName):
    match1 = re.search("%s[0-9]+$" % (MAT_TEXSET_SUFFIX), matName)
    if match1:
        match2 = re.search("[0-9]+", match1.group())
        textureSetIndex = int(match2.group())
        realMatName = matName[:match1.span()[0]]
        return realMatName, textureSetIndex
    #end if
    
    return matName, -1
#end def

# -----------------------------------------------------------------------------
# Returns Object else returns None
def get_object(objName):
    if objName in bpy.data.objects:
        return bpy.data.objects[objName]
    #end if
    
    return None
#end def

# -----------------------------------------------------------------------------
# Returns the root turok 2 object from current obj
def get_root_turok_object2(obj, turokObjType, objType):
    rootObj = obj
    parentObj = obj.parent
    while True:
        if parentObj is None:
            return rootObj
        #end if
        if parentObj.t2.objType == turokObjType and parentObj.type == objType:
            rootObj = parentObj
        #end if
        parentObj = parentObj.parent
    #end while
#end def

# -----------------------------------------------------------------------------
# Returns the root turok 2 object from current obj
def get_root_turok_object(obj, turokObjType):
    rootObj = obj
    parentObj = obj.parent
    while True:
        if parentObj is None:
            return rootObj
        #end if
        if parentObj.t2.objType == turokObjType:
            rootObj = parentObj
        #end if
        parentObj = parentObj.parent
    #end while
#end def

# -----------------------------------------------------------------------------
# Returns the root object from current obj
def get_root_object(obj, objType):
    rootObj = obj
    parentObj = obj.parent
    while True:
        if parentObj is None:
            return rootObj
        #end if
        if parentObj.type == objType:
            rootObj = parentObj
        #end if
        parentObj = parentObj.parent
    #end while
#end def

# -----------------------------------------------------------------------------
# Returns Object
def create_turok2_object(objName, objData, turokObjType, rotMode="QUATERNION"):
    obj = bpy.data.objects.new(name=objName, object_data=objData)
    obj.rotation_mode = rotMode
    obj.t2.objType = turokObjType
    bpy.context.scene.collection.objects.link(obj)
    get_collection(TUROK2_COLLECTION_NAME).objects.link(obj)
    
    return obj
#end def

# -----------------------------------------------------------------------------
# Returns collection object
def get_collection(collectionName):
    if collectionName not in bpy.data.collections:
        collection = bpy.data.collections.new(collectionName)
        bpy.context.scene.collection.children.link(collection)
        return collection
    #end if
    
    return bpy.data.collections[collectionName]
#end for

# -----------------------------------------------------------------------------
# Sets the blender mode. If context object is None then does nothing. (instead of throwing an error)
def mode_set(newMode):
    if bpy.context.object is None:
        return
    #end if
        
    bpy.ops.object.mode_set(mode=newMode)
#end def

# -----------------------------------------------------------------------------
# returns the packages preferences object
def prefs():
    return bpy.context.preferences.addons[__package__].preferences
#end def

# -----------------------------------------------------------------------------
# 
def message_box_draw(self, context):
    global last_popup_message
    
    self.layout.label(text=last_popup_message)
#end def
    
# -----------------------------------------------------------------------------
# 
def show_message(message, title = "Info", icon = 'INFO', toConsole = True):
    global last_popup_message
    
    last_popup_message = message
    bpy.context.window_manager.popup_menu(message_box_draw, title = title, icon = icon)
    if toConsole:
        print("[%s] %s" % (title, message))
    #end if
#end def

# -----------------------------------------------------------------------------
# 
def show_error(message, title = "Error", icon = 'ERROR', toConsole = True):
    global last_popup_message
    
    last_popup_message = message
    bpy.context.window_manager.popup_menu(message_box_draw, title = title, icon = icon)
    if toConsole:
        print("[%s] %s" % (title, message))
    #end if
#end def
